                         (* EIGHT.PAS *)


unit eight;

(* This unit contains data and procedures for access to
   signed digraphs (sparse shortint arrays defined in DIGRAPH)
   yn, yn[.,.,h] for h = 0 to ?, ymis.
   In earlier versions (up to SIENA 1.57), this was done
   in a different way in order to save memory,
   using bytes to store 8 booleans; hence the name of this unit.
   In addition, it contains data types and variables for covariates:
   z, zz, zx, zc.
   Array variables that are at least of order nmax
   are defined as dynamic variables.
   arrays of which parallel versions are available for networks and for behavior
   are denoted by yn* for networks and ya* (a = action) for behavior.
   yn is current network, ya is current behavior
    *)

interface

uses SysUtils, dialogs, digraph, classes, Math, s_constants, s_matrices;

const
      nzmax = 30;
(* Maximum number of individual covariates: changing plus non-changing *)
      nzzmax = 20;
(* Maximum number of dyadic covariates; must not be less than nzmax. *)

Type IntVec = array of Integer; // typically to be declared with dimension n+1
     Realnvec = array of Real;  // typically to be declared with dimension n+1
     array_0_20 = array[0..20] of real;
 Snzarray = array[1..nzmax] of string[12];
 Snzzarray = array[1..nzzmax] of string[12];

var  z   : array of array of Real;           (* matrix of constant actor covariates                  *)
     zc  : array of array of array of Real;  (* matrices of changing actor covariates,
                                                INCLUDING the dependent ones                         *)
     zz  : array of array of array of Real;  (* matrix of constant dyadic covariates                 *)
     zzc : array of array of array           (* CS: matrices of changing dyadic covariates           *)
                          of array of real;
// local : array[1..2,1..130,1..3] of integer;
     zx  : array of array of real;      (* MH: matrix of exogenous events; begin and end times  *)

     zamean : array[1..nzmax] of real;   (* overall average of dependent actor variables *)
     zcsglobalmean: array of real;           (* means of similarity variables for zc                 *)

//     SaveMat: array of array of ShortInt;     (* used in procedure S_Base.Statistics, defined globally  [s]
//                                                in order to have to define array bounds only once    *)
     ExistStruct: array of boolean;              (* existence of structurally determined values    *)
     ExistNewStruct: array of boolean;                 (* added 3.17s: there is change in structurally determined values *)


     SimDegrees: array of array of Integer;  (* Simulated ordered degrees: SimDegrees[i,j] is the
                                                number of i'th ordered degrees equal to j. This is
                                                used in S_Sim and S_Est for analysis of goodness-
                                                of-fit, only for descriptive purposes.               *)
    CumulDegrees: array of integer;
    (* cumulative degrees, used only for observations = 1 for more efficient
    drawing of random outties. *)
   mi, gc: array of array of real;
   (* observed moran and geary's network autocorrelations *)
   zsmean,           (* means of similarity scores                      *)
   range             (* range of the resp. actor variable in the sample *)
               : array[1..nzmax] of real;
    ZNames   : Snzarray;      (* names of individual covariates                          *)

//    xxx: string;

   (* functions for saving typing effort, to read-access data arrays: *)
function yn(i,j:  Integer): Boolean;              (* current network: binary           *)
function ynv(i,j:  Integer): ShortInt;    (*s*)       (* current network: valued           *)
(* yn == (ynv >= 1)  *)
function ya(i,j:  Integer): Byte;                 (* current behaviourals              *)
function ynm(i,j,m: Integer): Boolean;            (* network at beginning of period: binary *)
function ynvm(i,j,m: Integer): ShortInt;  (*s*)       (* network at beginning of period: valued *)
(* ynm == (ynvm >= 1)  *)
function mis(i,j,m: Integer): Boolean;            (* network missing at    --"--       *)
function struct(i,j,m: Integer): Boolean;         (* network structurally determined in --"-- *)
function misz(i,h: Integer): Boolean;             (* constant covariate missing        *)
function miszc(i,h,m: Integer): Boolean;          (* changing covariate missing --"--  *)
function miszz(i,j,h: Integer): Boolean;          (* constant dyadic covariate missing *)
(* AA start *)
function anymisz(h,nn,nnz,nobs: Integer): Boolean;          (* any missings in constant or changing covariate *)
function nstruct(m: Integer): integer;            (* number of structurally determined values in network wave m *)
(* AA end *)

(* functions for flexibility in data storage, to write-access data arrays: *)
procedure setyn(i,j: Integer; val: Boolean);      (* current network: binary           *)
procedure setynv(i,j: Integer; val: ShortInt); (*s*)  (* current network: valued           *)
procedure setya(i,j: Integer; val: Byte);         (* current behaviourals              *)
procedure minya(i,j: Integer);                    (* subtracts 1 from ya(i,h)          *)
procedure plusya(i,j: Integer);                   (* adds 1 to ya(i,h)                 *)
procedure setym(i,j,m: Integer; val: Boolean);    (* Sets element(i,j) of current network to val: binary *)
procedure setyvm(i,j,m: Integer; val: ShortInt); (*s*)(* Sets element(i,j) of current network to val: valued *)
procedure SetDataRowTo(const i,m: integer);       (* Sets row i    of current network to its value at observation m *)
procedure SetDataColumnTo(const i,m: integer);    (* Sets column i of current network to its value at observation m *)
procedure setmis(i,j,m: Integer; val: Boolean);   (* network missing at    --"--       *)
procedure setstruct(i,j,m: Integer; val: Boolean);   (* network missing at    --"--       *)
procedure setmisz(i,h: Integer; val: Boolean);    (* constant covariate missing        *)
procedure setmiszc(i,h,m: Integer; val: Boolean); (* changing covariate missing --"--  *)
procedure setmiszz(i,j,h: Integer; val: Boolean); (* constant dyadic covariate missing *)

function similarity(zi,zj: real; k: integer): real;
  (* calculates similarity value of actors i and j on actor covariate k *)

procedure PermitPlus(const nh,i,k: integer; const vp: shortint;
                     var pp: array of boolean);
(* for 1 <= j <= nh: pp[j] := ((not Struct(i,j,k)) and
                               ((ynv(i,j) < vp) or (ynv(i,j) = 0))) *)
procedure PermitMin(const nh,i,k: integer; const vm: shortint;
                     var pm: array of boolean);
(* for 1 <= j <= nh: pm[j] := ((not Struct(i,j,k)) and
                               ((ynv(i,j) > vm) or (ynv(i,j) = 0))) *)
procedure PermitChange(const nh,i,k: integer; const vp: integer;
                     var pc: array of boolean);
(* for 1 <= j <= nh: pc[j] := ((not Struct(i,j,k)) and
                               ((outd_yn(i) < v) or yn(i,j))) *)
procedure ForbidMissings(const i,k: integer; var pc: array of boolean);
(* changes pc[j] to false for all j with mis(i,j,k) or mis(i,j,k+1)
   or Struct(i,j,k+1) *)

function Outd_yn(const i: integer): cardinal;
(* outdegree *)
function Ind_yn(const i: integer): cardinal;
(* indegree *)
function PosOutd_yn(const i: integer): cardinal;
(* outdegree, only positive ties *)
function PosInd_yn(const i: integer): cardinal;
(* indegree, only positive ties *)
function NegOutd_yn(const i: integer): cardinal;
(* outdegree, only negative ties *)
function NegInd_yn(const i: integer): cardinal;
(* indegree, only negative ties *)
function ReciDegree_yn(const i: integer): cardinal;
(* reciprocated degree *)
function StructRowElements(const i,m: integer): cardinal;
(* number of elements in row i of the adjacency matrix for observation m,
   that are structurally predetermined *)
function StructElements(const m: integer): cardinal;
(* number of elements of the adjacency matrix for observation m,
   that are structurally predetermined *)
(* AA change *) procedure StructuralDataValues(const k: integer; var a: array of integer);
(* gives for all h the number of structurally determined values equal to h
   in a[h-mini], where mini is the minimum value (usually 0). *)
function ConnectedCheck(var f: TFileStream): boolean;
function ConnectedNetwork: boolean;
function PathNetwork(i,j: integer): boolean;
function OtherPathNetwork(i,j: integer): boolean;

   function Dist2(i,j,k: integer): boolean;
(* indicates that the directed distance from i to j in observation k is 2 *)

function fbalsim(const dg: TValuedDigraph;
                 const i,j,k: integer; const bm: real): real;
(* Balance Similarity Component *)

function CorrectionNDist2(const dg: TValuedDigraph; const i,k: integer): real;
(* missing data correction for of number of distances 2:
   subtracts distances two between nodes with missing tie variables *)

function CorrectionNTwoDist2(const dg: TValuedDigraph; const i,k: integer): real;
(* missing data correction for of number of two-distances 2:
   subtracts distances two between nodes with missing tie variables *)

   (* AA start *)
function CorrectionVarDist2(const dg: TValuedDigraph; const i,k,par: integer): real;
(* AA  end *)

function CorrectionBetween(const dg: TValuedDigraph; const i,k: integer): real;
(* missing data correction for betweenness statistic;
   subtracts number of twopaths between nodes with missing tie variables *)

function AllSymmetric(m: integer): Boolean;
(* Indicates that all outcome networks and all missing structures
   are symmetric *)

function MissingOut(i,m: integer): cardinal;
(* Number of missings in ynm(i,h,m) counted over h *)
function MissingIn(i,m: integer): cardinal;
(* Number of missings in ynm(h,i,m) counted over h *)
function ConsecutiveMissings(m: integer): cardinal;
(* Number of (i,j) for which there is a missing at time m or m+1 *)
function TotalMissings(m: integer): cardinal;
(* Total missings at time point m *)

function MissingInTwoStars(i,j,m: integer): cardinal;
(* Number of missings in ynm(i,h,m) and simultaneously ynm(j,h,m) counted over h *)

function IndMis(i,m: integer): cardinal;
(* Indegree of i at observation m, only counting missings;
   this can be nonzero because of the carry forward option. *)

function Outdm(const i,m: integer): cardinal;
function Indm(const i,m: integer): cardinal;
function PosOutdm(const i,m: integer): cardinal;
(* outdegree, only positive ties *)
function PosIndm(const i,m: integer): cardinal;
(* indegree, only positive ties *)
function NegOutdm(const i,m: integer): cardinal;
(* outdegree, only negative ties *)
function NegIndm(const i,m: integer): cardinal;
(* indegree, only negative ties *)
function Recidm(const i,m: integer): cardinal;
(* outdegree, indegree, reciprocated degree at beginning of period m;
   for valued graphs, counts -1 as well as +1 values *)

function StructureExists(const m: integer): boolean;
(* There is at least one structurally determined value *)

function StructSym(const m: integer): boolean;
(* The structurally determined entries are symmetric *)

// added 3.17s:
procedure SetChangeStruct(obs: integer);
(* Sets matrices for changes in structuraslly determined values *)

function RandomCurrentOutTie(const version: byte; const i: integer): integer;
(* Returns a random outtie of i; returns i if outdegree(i) = 0. *)

procedure RandomCurrentTie(const version: byte; var i,j: integer);
(* Returns the coordinates (i,j) of a random tie in the current network;
   0,0 if the network is empty;
   requires that CumulDegrees is updated properly;
   this is now (version 3.11) done only in ErgmRunStep, depending on Gibbs. *)

//procedure bal(const n,m: integer); // @@@


procedure AddNoTies_yn(const i,k: integer;
         var p: array of real; perm: array of boolean;
                                Neffects:TEffects); overload;
procedure AddNoTies_yn(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const calcch: boolean; var chas: rrrarray; Neffects:TEffects); overload;
(* Adds a[j]*f(i,j) over all j having no yn tie from i to j and satisfying perm[j] *)
procedure SubtractTies_yn(const i,k: integer;
         var p: array of real; perm: array of boolean;
                                Neffects: TEffects); overload;
procedure SubtractTies_yn(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const calcch: boolean; var chas: rrrarray; Neffects: TEffects); overload;
(* subtracts a[j]*f(i,j) over all j having a yn tie from i to j and satisfying perm[j] *)

procedure ERGMChangeStatistics(const i,j: integer;
         var s: pvec;
         const t,tr: shortint; Neffects:TEffects);
(* Adds the change statistics (all effects h) for toggling tie variable (i,j)
   to array element s[h];
   for symmetric, the effects for toggling (i,j) as well as (j,i);
   t is the current value of matyn(i,j); tr is the current value of matyn(j,i).*)

procedure ERGMStatistics(var s: pvec; const symm: boolean;
                         ordeg: array of integer; const generate: boolean; 
                         Neffects:TEffects);
(* Calculates the statistics from scratch and stores them in array s. *)

procedure UtilityComponents_yn(const i,k,h,n: integer;
         var s: rrarray; perm: array of boolean;
         Neffects: TEffects);
(* Stores the utility function contributions (number h1)
   for changing yn tie variable (i,j)
   in array s[j,h+h1] , for all j with perm[j]; k is the period;
   first evaluation function, then endowment function *)

function utilitygain(const i,j: integer; const k: byte; const sym: boolean;
                     const RecMod: boolean; Neffects:TEffects): real; overload;

function utilitygain(const yy: boolean;
                     const i,j: integer; const k: byte; const sym: boolean;
                     const RecMod: boolean; Neffects:TEffects): real; overload;

procedure CalcComponents_fa(i,nf,af,nuz: integer; kk,per: byte; dif: shortint;
                            initial: boolean;
                            var eff: BBarray; var fa_comps: array_0_20;
                            var outf: TFileStream);
  (* Calculates ingredients for contr_fa for different Action effects h;
     must be called before calling contr_fa. *)



//procedure CreateDatMat(const k: integer; var dg: TValuedDigraph);
//(* Creates a TValueDigraph dg defined as the current network,
//   with all ties skipped for which information was missing at time k or k+1.*)

procedure ImputePrevious(const k: integer);
(* Replaces matym[k](i,j) for which matym[k-1](i,j) <> 0 and matmis[k](i,j)=1,
   by matym[k-1](i,j). *)

procedure ForceStructurals(const k: integer);     
(* Replaces matyn(i,j) for which matstruct[k](i,j) <> 0
   by matym[k](i,j). *)

procedure ForceNewStructurals(const k: integer);     // added 3.17s
(* Replaces matyn(i,j) for which matnewstruct[k](i,j) <> 0
   by matym[k](i,j). *)

(*
procedure CreateDatMat1(const k: integer; var dg: TValuedDigraph);
*)
(* Creates a TValueDigraph dg defined as the current network,
   with all ties skipped for which information was missing at time k .*)

{procedure ReduceMat(const k: integer; var dg: TValuedDigraph);
(* Changes the existing data matrix dg so that all ties are skipped
  for which information was missing at time k+1 .*)
}

procedure MissingsToSaveMat(const k: integer);
(* Calculates SaveMat,
   with SaveMat(i,j) = max(ynv(i,j), min(mis(i,j,k), mis(i,j,k+1))),
   and subtracts SaveMat from ynv; note that ynv = current network. *)

procedure RestoreSaveMat;
(* calculates ynv = max(ynv, SaveMat) *)


procedure SumDifMat(const k: integer; var s, sup, sdown, stot: integer);

procedure CreateDifMat(const k: integer; var dg: TValuedDigraph);
(* Creates a TValueDigraph dg defined as the network
   with ties wherever the current network differs from the network at time k,
   with all ties skipped for which information was missing at times k or k+1 .*)


procedure CreateLostMat(const k: integer; var dg: TValuedDigraph);
(* Creates a TValueDigraph dg defined as the network
   with ties equal to the ties at time k,
   having values 1 if they still exist in the current network
   and -1 if they disappeared in the current network,
   with all ties skipped for which information was missing at times k or k+1 .*)

procedure CalcFunctions_yn(const k: integer; var fun: array of real; var ordeg: array of integer;
            Effs:TEffects; const CalcOD: Boolean; var outfile: TFileStream);
(* Calculates statistics for current network, for non-missing data
   at observations k and k+1. *)

function SaveMat(i,j: Integer): ShortInt;
procedure DefineSaveMat(i,j: Integer; v: ShortInt);

procedure ReadData(const fn: shortstring; const nh, k: integer;
                   var vmin, vmax: shortint);
(* Reads valued digraph data file for observation k from file with name fn *)
procedure SetDigraph(const k: integer);
(* Sets current valued digraph to its value for observation k. *)
procedure SetStructOne(const k: integer);
(* Sets current valued digraph to only the structural ones for observation k. *)
procedure ReadPajekDataFile(const fn: shortstring; var ert: shortstring;
                             var ext: boolean;
                             const k: integer; var vmin,vmax: shortint);
(* Reads valued digraph data file for observation k from Pajek file with name fn *)
procedure ReadPajekDataFileToMatrix(const fn: shortstring;
                             const nh: integer; var ert: shortstring;
                             var ext: boolean;
                             var zout: iiarray; var vmin,vmax: shortint);
(* Reads valued digraph data file for observation k from Pajek file with name fn *)
procedure ReadDigraph(const fn: shortstring; const nh, k: integer);
(* Reads valued digraph data file for observation k from file with name fn
   and copies this to the current digraph. *)
procedure WriteData(const fn: shortstring; const k: integer);
(* Writes valued digraph data file for observation k to file with name fn *)
procedure WriteTheseData(const fn: shortstring);
(* Writes current valued digraph data to file with name fn in internal Siena format *)
procedure WriteSimulatedData(const uppertr: boolean; var f: textfile);
(* Writes current valued digraph data file to file f *)
procedure WriteSubsetSimulatedData(
       const z1: iarray; const z2: boarray; var f: textfile);
(* Writes current valued digraph data file for subset of nodes to file f *)
procedure ReadMisData(const fn: shortstring; const nh, k: integer);
(* Reads missing data file for observation k from file with name fn *)
procedure WriteMisData(const fn: shortstring; const k: integer);
(* Writes missing data file for observation k to file with name fn *)
procedure ReadStructData(const fn: shortstring; const nh, k: integer);
(* Reads structurally determined data file for observation k from file with name fn *)
procedure WriteStructData(const fn: shortstring; const k: integer);
procedure DescribeCliques(var outf: TFileStream; const red, smin, smax: integer);
procedure MakeDigraph(const k: integer);
(* Sets the digraph for observation k to the current digraph. *)
procedure CopyCurrentDigraph(var dg: TValuedDigraph; const k: integer);
(* Sets dg to current valued digraph. *)
procedure ChangeDataValues(const a,b: array of shortint; const maxab: byte;
                           const k: integer; kind: byte);
(* Changes values in observation k equal to a[h] into b[h],
   and notes this in structural zeros if kind = 1, in missings if kind = 2. *)

procedure DescribeNetworksCovariate(const nn, nz, nk: integer;
          var outf: TFileStream);
procedure DescribeNetworksCCovariate(const nn, nz, nzc, nk: integer;
          var outf: TFileStream);

// for checking only procedure CompareDist2(n: integer; var d1, d2: integer);


procedure DistancesFromHighDegree(var d,z: iarray);

procedure SetPeriods(nh,nhzc,nhzz,nhzzc,per: integer); (* initializes length of some dynamic arrays *)
procedure FinishEight(var outf: tFileStream);

implementation

uses slib;

var
    matyn     : TValuedDigraph;
    matym     : array of TValuedDigraph;
    matmis,
    matstruct,
    matnewstruct : array of TDigraph;  // added 3.17s
    matya     : array of array of Byte;
    matmisz   : array of array of Boolean;
    matmiszc,
    matmiszz  : array of array of array of Boolean;
    matSaveMat: TValuedDigraph;
    MatricesCreated : Boolean;
function ident(const x: real): real;
begin
  ident := x;
end; (* ident *)


function yn(i,j:  Integer): Boolean;
begin
  yn := (matyn.TieValue(i,j) >= 1);
end;

function ynv(i,j:  Integer): ShortInt;
begin
  ynv := matyn.TieValue(i,j);
end;

function ya(i,j:  Integer): Byte;
begin
  ya:= matya[i,j];
end;

function ynm(i,j,m: Integer): Boolean;
begin
//  ynm:= (matym[i,j,m] >= 1);
  ynm := (matym[m].TieValue(i,j) >= 1);
end;

function ynvm(i,j,m: Integer): ShortInt;
begin
  ynvm := matym[m].TieValue(i,j);
end;

function mis(i,j,m: Integer): Boolean;
begin
  mis := (matmis[m].TieValue(i,j) >= 1);
end;

function struct(i,j,m: Integer): Boolean;
begin
  if (i=j) then struct := true else
  struct := (matstruct[m].TieValue(i,j) >= 1);
end;

function misz(i,h: Integer): Boolean;
begin
  misz:= matmisz[i,h];
end;

(* AA start *)
function anymisz(h,nn,nnz,nobs: Integer): Boolean;
var i,m: integer;
    am: boolean;
begin
  am := false;
  i :=0;
  if (h <= nnz) then
(* constant covariate *)
    begin
    repeat
      inc(i);
      if matmisz[i,h] then am := true;
     until (i >= nn) or am;
    end
   else
(* changing covariate *)
    begin
    m := 1;
    repeat
      inc(i);
      if matmiszc[i,h-nnz,m] then am := true;
      if (i >= nn) then (* go to next period *)
        begin
        i := 0;
        inc(m);
        end;
     until (m >= nobs) or am;
    end;
  anymisz := am;
end; (* anymisz *)

function nstruct(m: Integer): integer;
begin
  nstruct := matstruct[m].TotalTies;
end;

(* AA end *)
                                           
function miszc(i,h,m: Integer): Boolean;
begin
  miszc:= matmiszc[i,h,m];
end;

function miszz(i,j,h: Integer): Boolean;
begin
  miszz:= matmiszz[i,j,h];
end;

procedure setyn(i,j: Integer; val: Boolean);   (* sets yn(i,j) to value *)  (*[s]*)
begin
 if val then matyn.ChangeTieValue(i,j,1) else matyn.ChangeTieValue(i,j,0);
end;

procedure setynv(i,j: Integer; val: ShortInt);   (* sets ynv(i,j) to value *)  (*[s]*)
begin
 matyn.ChangeTieValue(i,j,val);
end;

procedure setya(i,j: Integer; val: Byte);   (* sets ya(i,j) to value *)
begin
 matya[i,j] := val;
end;

procedure minya(i,j: Integer);   (* subtracts 1 from ya(i,j) *)
begin
  dec(matya[i,j]);
end;

procedure plusya(i,j: Integer);   (* adds 1 to ya(i,j) *)
begin
  inc(matya[i,j]);
end;

procedure setym(i,j,m: Integer; val: Boolean);   (* sets ynm(i,j,m) to value *)
begin
 if val then matym[m].ChangeTieValue(i,j,1) else matym[m].ChangeTieValue(i,j,0);
end;

procedure setyvm(i,j,m: Integer; val: ShortInt);   (* sets ynm(i,j,m) to value *)
begin
 matym[m].ChangeTieValue(i,j,val);
end;

procedure SetDataRowTo(const i,m: integer);
begin
 matyn.SetRowTo(i,matym[m].OutTies[i]);
end;

procedure SetDataColumnTo(const i,m: integer);
begin
 matyn.SetColumnTo(i,matym[m].InTies[i]);
end;

procedure setmis(i,j,m: Integer; val: Boolean);   (* sets mis(i,j) to value *)
begin
// matmis[i,j,m] := val;
 if ((i <> j) and val) then matmis[m].ChangeTieValue(i,j,1) else matmis[m].ChangeTieValue(i,j,0);
end;

procedure setstruct(i,j,m: Integer; val: Boolean);   (* sets struct(i,j) to value *)
begin
// matstruct[i,j,m] := val;
  if (i = j) then exit;
 if val then matstruct[m].ChangeTieValue(i,j,1) else matstruct[m].ChangeTieValue(i,j,0);
end;

procedure setmisz(i,h: Integer; val: Boolean);   (* sets misz(i) to value *)
begin
 matmisz[i,h] := val;
end;

procedure setmiszc(i,h,m: Integer; val: Boolean);   (* sets miszc(i,h) to value *)
begin
  matmiszc[i,h,m] := val; 
end;

procedure setmiszz(i,j,h: Integer; val: Boolean);   (* sets miszz(i,j) to value *)
begin
 matmiszz[i,j,h] := val;
end;

procedure SetRow(const i, k: integer);
(* sets row i of the network to the corresponding network row of observation k *)
begin
end;

function similarity(zi,zj: real; k: integer): real;
(* similarity values of actors i and j on actor covariate k; values
   range between zero (maximally dissimilar) and one (maximally similar),
   centered.
   NOTE when calling the function that k refers to the WHOLE list of actor
   covariates 1...(nz+nzc)                                                 *)
begin
 similarity := 1.0 - (abs(zi-zj)/range[k]) - zsmean[k];
end;

procedure PermitPlus(const nh,i,k: integer; const vp: shortint;
                     var pp: array of boolean);
(* for 1 <= j <= nh: pp[j] := ((not Struct(i,j,k)) and
                               ((ynv(i,j) < vp) or (ynv(i,j) = 0))) *)
var j: integer;
   it: TDoublyLinkedDoubleInt;
begin
  for j:=nh downto 1 do pp[j] := true;
  it := matstruct[k].OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    pp[it.Value1] := false;
    end;
  it := matyn.OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    if (it.Value2 >= vp) then pp[it.Value1] := false;
    end;
end; (* PermitPlus *)

procedure PermitMin(const nh,i,k: integer; const vm: shortint;
                     var pm: array of boolean);
(* for 1 <= j <= nh: pm[j] := ((not Struct(i,j,k)) and
                               ((ynv(i,j) > vm) or (ynv(i,j) = 0))) *)
var j: integer;
   it: TDoublyLinkedDoubleInt;
begin
  for j:=nh downto 1 do pm[j] := true;
  it := matstruct[k].OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    pm[it.Value1] := false;
    end;
  it := matyn.OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    if (it.Value2 <= vm) then pm[it.Value1] := false;
    end;
end; (* PermitMin *)


procedure PermitChange(const nh,i,k: integer; const vp: integer;
                       var pc: array of boolean);
(* for 1 <= j <= nh: pc[j] := ((not Struct(i,j,k)) and
                               ((outd_yn(i) < v) or yn(i,j))) *)
var j: integer;
   it: TDoublyLinkedDoubleInt;
begin
  if (outd_yn(i) >= vp) then
    begin
(* Only changes are allowed for j to which there currently is a tie *)
    for j:=nh downto 1 do pc[j] := false;
    it := matyn.OutTies[i].First;
    while it.HasNext do
      begin
      it := it.Next;
      pc[it.Value1] := true;
      end;
    end
   else for j:=nh downto 1 do pc[j] := true;
(* Forbid changes for structurally determined values. *)
  it := matstruct[k].OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    pc[it.Value1] := false;
    end;
end; (* PermitChange *)

procedure ForbidMissings(const i,k: integer; var pc: array of boolean);
var it: TDoublyLinkedDoubleInt;
begin
  it := matmis[k].OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    pc[it.Value1] := false;
    end;                  
  it := matmis[k+1].OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    pc[it.Value1] := false;
    end;
  it := matstruct[k+1].OutTies[i].First;
  while it.HasNext do
    begin
    it := it.Next;
    pc[it.Value1] := false;
    end;
end; (* ForbidMissings *)

function Outd_yn(const i: integer): cardinal;
(* outdegree *)
begin
  Result := matyn.OutDegree(i);
end;

function Ind_yn(const i: integer): cardinal;
(* indegree *)
begin
  Result := matyn.InDegree(i);
end;

function PosOutd_yn(const i: integer): cardinal;
(* outdegree, only positive ties *)
begin
  Result := matyn.PosOutdegree(i);
end;

function PosInd_yn(const i: integer): cardinal;
(* indegree, only positive ties *)
begin
  Result := matyn.PosIndegree(i);
end;

function NegOutd_yn(const i: integer): cardinal;
(* outdegree, only negative ties *)
begin
  Result := matyn.NegOutdegree(i);
end;

function NegInd_yn(const i: integer): cardinal;
(* indegree, only negative ties *)
begin
  Result := matyn.NegIndegree(i);
end;





function ReciDegree_yn(const i: integer): cardinal;
(* indegree *)
begin
  Result := matyn.ReciDegree(i);
end;

function StructRowElements(const i,m: integer): cardinal;
(* number of nondiagonal elements in row i of the adjacency matrix for observation m,
   that are structurally predetermined *)
begin
  Result := matstruct[m].OutDegree(i);
end;

function StructElements(const m: integer): cardinal;
(* number of nondiagonal elements in row i of the adjacency matrix for observation m,
   that are structurally predetermined *)
begin
  Result := matstruct[m].TotalTies;
end;

function ConnectedCheck(var f: TFileStream): boolean;
var i,n: integer;
    ch: boolean;
begin
  n := matyn.Order;
  ch := matyn.Connected(n);
  Result := ch;
  for i:= n-1 downto 1 do
    if (matyn.Connected(i) <> ch) then
      begin
      wwriteln(f,'Misconnected: '+inttostr(i)+', '+inttostr(n)+'.');
      result := false;
      end;
 if result then wwriteln(f,'connected') else wwriteln(f,'disconnected');
end;


function ConnectedNetwork: boolean;
var i: integer;
begin
  i := 1;
  Result := matyn.Connected(i);
end;

function PathNetwork(i,j: integer): boolean;
begin
  Result := matyn.Path(i,j);
end;

function OtherPathNetwork(i,j: integer): boolean;
begin
  Result := matyn.OtherPath(i,j);
end;

function fbalsim(const dg: TValuedDigraph;
                 const i,j,k: integer; const bm: real): real;
(* Balance Similarity Component *)
(* To reproduce the results obtained in Siena versions 2 and earlier,
   delete the four lines indicated below as being extra conditions in Siena 3 *)
var it1, it2: TDoublyLinkedDoubleInt;
begin
(* programmed according to IndicatorInnerProduct and InTwoStars *)
  Result := (dg.Order + 0.0 - matmis[succ(k)].OutDegree(i) - matmis[succ(k)].OutDegree(j)
            + matmis[succ(k)].InTwoStars(i,j) - 2 + matmis[succ(k)].TieValue(i,j) + matmis[succ(k)].TieValue(j,i)   );
//the same:  @@@  kk := 0;
//for h:=order downto 1 do if (matmis[succ(k)].Tievalue(i,h) <= 0) and (matmis[succ(k)].TieValue(j,h) <= 0)
//             and (i <> h) and (j <> h) then inc(kk);
//Result := kk;

  Result := bm*Result;

  it1 := dg.OutTies[i].First;
  it2 := dg.OutTies[j].First;
  while (it1.hasnext and it2.hasnext) do
   case sign(it1.next.value1 - it2.next.value1) of
    -1: begin
         it1 := it1.next;
(* for h = it1.value1, there is a tie (i,h) and no tie (j,h) *)
         if (matmis[succ(k)].TieValue(i,it1.value1) <= 0) and
            (matmis[succ(k)].TieValue(j,it1.value1) <= 0) and
            (matmis[k].TieValue(j,it1.value1) <= 0) and
            // the preceding line is an extra condition in Siena 3
            (it1.value1 <> j)
          then Result := Result - 1.0;
        end;
    0: begin
         it1 := it1.next;
         it2 := it2.next;
       end; (* case 0 *)
    1: begin
         it2 := it2.next;
(* for h = it2.value1, there is a tie (j,h) and no tie (i,h) *)
         if (matmis[succ(k)].TieValue(i,it2.value1) <= 0) and
            (matmis[succ(k)].TieValue(j,it2.value1) <= 0)
            and (matmis[k].TieValue(i,it2.value1) <= 0)
            // the preceding line is an extra condition in Siena 3
 and (it2.value1 <> i)
           then Result := Result - 1.0;
         end;
   end; (* case *)

(* Now at least one of it1 or it2 has no next. *)
  while it1.HasNext do begin
    it1 := it1.next;
    if (matmis[succ(k)].TieValue(i,it1.value1) <= 0) and
       (matmis[succ(k)].TieValue(j,it1.value1) <= 0) and
       (matmis[k].TieValue(j,it1.value1) <= 0) and
       // the preceding line is an extra condition in Siena 3
       (it1.value1 <> j)
      then Result := Result - 1.0;
   end;
  while it2.HasNext do begin
    it2 := it2.next;
         if (matmis[succ(k)].TieValue(i,it2.value1) <= 0) and
            (matmis[succ(k)].TieValue(j,it2.value1) <= 0) and
            (matmis[k].TieValue(i,it2.value1) <= 0) and
         // the preceding line is an extra condition in Siena 3
       (it2.value1 <> i)
      then Result := Result - 1.0;
     end;

//   Result := Result/(dg.Order - 2.0);  dropped 3.17w
end; (* fbalsim *)

function CorrectionNDist2(const dg: TValuedDigraph; const i,k: integer): real;
(* missing data correction for number of distances 2:
   subtracts distances two between nodes with missing tie variables *)
var it1, it2: TDoublyLinkedDoubleInt;
  res, h1, h2: integer;
begin
(* h1 and h2 are the successive points with d1 and d2 ties from i;
   defined as high(integer) if there is no more such points *)
  Res := 0;
  it1 := matmis[k].OutTies[i].First;
  it2 := matmis[succ(k)].OutTies[i].First;
  if it1.hasnext then begin
    it1 := it1.next;
    h1 := it1.value1;
   end else h1 := high(integer);
  if it2.hasnext then begin
    it2 := it2.next;
    h2 := it2.value1;
   end else h2 := high(integer);
(* carry out action for the smallest of h1 and h2 *)
  while (h1 <= dg.order) or (h2 <= dg.order) do case sign(h1 - h2) of
   -1: begin
         if (not dg.NoTwoPath(i,h1,high(integer))) then dec(res);
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
        end;
    0: begin
         if (not dg.NoTwoPath(i,h1,high(integer))) then dec(res);
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
    1: begin
         if (not dg.NoTwoPath(i,h2,high(integer))) then dec(res);
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
     end; (* case *)
   result := res;
end; (* CorrectionNDist2 *)

function CorrectionNTwoDist2(const dg: TValuedDigraph; const i,k: integer): real;
(* missing data correction for number of two-distances 2:
   subtracts two-distances two between nodes with missing tie variables *)
var it1, it2: TDoublyLinkedDoubleInt;
  res, h1, h2: integer;
begin
(* h1 and h2 are the successive points with d1 and d2 ties from i;
   defined as high(integer) if there is no more such points *)
  Res := 0;
  it1 := matmis[k].OutTies[i].First;
  it2 := matmis[succ(k)].OutTies[i].First;
  if it1.hasnext then begin
    it1 := it1.next;
    h1 := it1.value1;
   end else h1 := high(integer);
  if it2.hasnext then begin
    it2 := it2.next;
    h2 := it2.value1;
   end else h2 := high(integer);
(* carry out action for the smallest of h1 and h2 *)
  while (h1 <= dg.order) or (h2 <= dg.order) do case sign(h1 - h2) of
   -1: begin
         if (not dg.NoOrOneTwoPathMax(i,h1,high(integer))) then dec(res);
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
        end;
    0: begin
         if (not dg.NoOrOneTwoPathMax(i,h1,high(integer))) then dec(res);
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
    1: begin
         if (not dg.NoOrOneTwoPathMax(i,h2,high(integer))) then dec(res);
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
     end; (* case *)
   result := res;
end; (* CorrectionNDist2 *)

(* AA start *)
function CorrectionVarDist2(const dg: TValuedDigraph; const i,k,par: integer): real;
(* missing data correction for values of variables of distances 2:
   subtracts distances two between nodes with missing tie variables *)
var it1, it2: TDoublyLinkedDoubleInt;
  h1, h2: integer;
begin
(* h1 and h2 are the successive points with d1 and d2 ties from i;
   defined as high(integer) if there is no more such points *)
  Result := 0;
  it1 := matmis[k].OutTies[i].First;
  it2 := matmis[succ(k)].OutTies[i].First;
  if it1.hasnext then begin
    it1 := it1.next;
    h1 := it1.value1;
   end else h1 := high(integer);
  if it2.hasnext then begin
    it2 := it2.next;
    h2 := it2.value1;
   end else h2 := high(integer);
(* carry out action for the smallest of h1 and h2 *)
  while (h1 <= dg.order) or (h2 <= dg.order) do case sign(h1 - h2) of
   -1: begin
         if (not dg.NoTwoPath(i,h1,high(integer)))
           then Result := Result - z[h1,par];
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
        end;
    0: begin
         if (not dg.NoTwoPath(i,h1,high(integer)))
           then Result := Result - z[h1,par];
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
    1: begin
         if (not dg.NoTwoPath(i,h2,high(integer)))
           then Result := Result - z[h2,par];
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
     end; (* case *)
end; (* CorrectionVarDist2 *)
(* AA end *)

function CorrectionBetween(const dg: TValuedDigraph; const i,k: integer): real;
(* missing data correction for betweenness statistic;
   subtracts number of twopaths between nodes with missing tie variables *)
var it1, it2: TDoublyLinkedDoubleInt;
  res, h1, h2: integer;
begin
(* h1 and h2 are the successive points with d1 and d2 ties from i;
   defined as high(integer) if there is no more such points *)
  Res := 0;
  it1 := matmis[k].OutTies[i].First;
  it2 := matmis[succ(k)].OutTies[i].First;
  if it1.hasnext then begin
    it1 := it1.next;
    h1 := it1.value1;
   end else h1 := high(integer);
  if it2.hasnext then begin
    it2 := it2.next;
    h2 := it2.value1;
   end else h2 := high(integer);
(* carry out action for the smallest of h1 and h2 *)
  while (h1 <= dg.order) or (h2 <= dg.order) do case sign(h1 - h2) of
   -1: begin
         dec(res, dg.TwoPaths(i,h1));
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
        end;
    0: begin
         dec(res, dg.TwoPaths(i,h1));
         if it1.hasnext then begin
          it1 := it1.next;
          h1 := it1.value1;
          end else h1 := high(integer);
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
    1: begin
         dec(res, dg.TwoPaths(i,h2));
         if it2.hasnext then begin
          it2 := it2.next;
          h2 := it2.value1;
          end else h2 := high(integer);
        end;
     end; (* case *)
   result := res;
end; (* CorrectionBetween *)

function Dist2(i,j,k: integer): boolean;
(* Note: mis(i,j,k) does not imply (not ynm(i,j,k))
         because of the carryforward option. *)
begin
  result := (i <> j) and (not mis(i,j,k)) and (not ynm(i,j,k)) and
    (not matym[k].notwopath(i,j,high(integer)));
end;

function Outdm(const i,m: integer): cardinal;
(* outdegree at beginning of period m *)
begin
  Result := matym[m].Outdegree(i);
end;

function Indm(const i,m: integer): cardinal;
(* indegree at beginning of period m*)
begin
  Result := matym[m].Indegree(i);
end;

function PosOutdm(const i,m: integer): cardinal;
(* outdegree, only positive ties *)
begin
  Result := matym[m].PosOutdegree(i);
end;

function PosIndm(const i,m: integer): cardinal;
(* indegree, only positive ties *)
begin
  Result := matym[m].PosIndegree(i);
end;

function NegOutdm(const i,m: integer): cardinal;
(* outdegree, only negative ties *)
begin
  Result := matym[m].NegOutdegree(i);
end;

function NegIndm(const i,m: integer): cardinal;
(* indegree, only negative ties *)
begin
  Result := matym[m].NegIndegree(i);
end;



function Recidm(const i,m: integer): cardinal;
(* reciprocated degree at beginning of period m*)
begin
  Result := matym[m].Recidegree(i);
end;

function StructSym(const m: integer): boolean;
(* The structurally determined entries are symmetric  *)
var i: integer;
begin
  StructSym := matstruct[m].SymmetricGraph;
end;

function StructureExists(const m: integer): boolean;
(* There is at least one structurally determined value *)
var i: integer;
    se: boolean;
begin
  se := false;
  i := matstruct[m].Order;
  repeat
    if (matstruct[m].OutDegree(i) > 0) then se := true;
    dec(i);
    until (se or (i <= 0));
   result := se;
end;

function AllSymmetric(m: integer): Boolean;
var sy: boolean;
    i: integer;
begin
  sy := true;
  i := m;
  while (sy and (i>= 1)) do
    begin
    sy := (matmis[i].SymmetricGraph and matym[i].SymmetricGraph
           and matstruct[i].SymmetricGraph);
    dec(i);
    end;
  AllSymmetric := sy;
end;

function MissingOut(i,m: integer): cardinal;
(* Number of missings in ynm(i,h,m) counted over h *)
begin
  Result := matmis[m].OutDegree(i);
end;

function MissingIn(i,m: integer): cardinal;
(* Number of missings in ynm(h,i,m) counted over h *)
begin
  Result := matmis[m].InDegree(i);
end;


function ConsecutiveMissings(m: integer): cardinal;
(* Number of (i,j) for which there is a missing at time m or m+1 *)
begin
  ConsecutiveMissings := OrCount(matmis[m],matmis[m+1]);
end;

function TotalMissings(m: integer): cardinal;
(* Total missings at time point m *)
begin
  TotalMissings := matmis[m].TotalTies;
end;

function MissingInTwoStars(i,j,m: integer): cardinal;
(* Number of missings in ynm(i,h,m) and simultaneously ynm(j,h,m) counted over h *)
begin
  Result := matmis[m].InTwoStars(i,j);
end;

function IndMis(i,m: integer): cardinal;
(* Indegree of i at observation m, only counting missings *)
begin
  Result := JointIn(matmis[m],matym[m],i);
end;


function RandomCurrentOutTie(const version: byte; const i: integer): integer;
(* Returns a random outtie of i; returns i if outdegree(i) = 0. *)
begin
  Result := matyn.RandomOutTie(version,i);
end;


procedure RandomCurrentTie(const version: byte; var i,j: integer);
(* Returns the coordinates (i,j) of a random tie in the current network;
   0,0 if the network is empty. *)
begin
  matyn.RandomTie(version,CumulDegrees,i,j);
end;


procedure AddNoTies_yn(const i,k: integer;
         var p: array of real; perm: array of boolean;
                                Neffects:TEffects);
(* Adds matyn.Eff1(Neffects,a,i,j) to p[j] over all j having no tie from i to j
                                             and satisfying perm[j] *)
begin
  Neffects.AddNoTies(i,k,p,perm,matyn);
end;

procedure AddNoTies_yn(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const calcch: boolean; var chas: rrrarray; Neffects:TEffects);
(* overloaded version *)         
begin
  Neffects.AddNoTies(i,k,p,a,perm,matyn,calcch,chas);
end;

procedure SubtractTies_yn(const i,k: integer;
         var p: array of real; perm: array of boolean;
                                Neffects:TEffects);
(* Subtracts matyn.Eff1(Neffects,a,i,j) from p[j] over all j
   having a tie from i to j and satisfying perm[j] *)
begin
  Neffects.SubtractTies(i,k,p,perm,matyn);
end;

procedure SubtractTies_yn(const i,k: integer;
         var p: array of real; const a: array of real; perm: array of boolean;
         const calcch: boolean; var chas: rrrarray; Neffects:TEffects);
(* Subtracts matyn.Eff1(Neffects,a,i,j) from p[j] over all j
   having a tie from i to j and satisfying perm[j] *)
begin
  Neffects.SubtractTies(i,k,p,a,perm,matyn,calcch,chas);
end;


procedure ERGMChangeStatistics(const i,j: integer;
         var s: pvec;
         const t,tr: shortint; Neffects:TEffects);
(* Adds the change statistics (all effects h) for toggling tie variable (i,j)
   to array element s[h];
   t is the current value of matyn(i,j); tr is the current value of matyn(j,i).*)
begin
  NEffects.ChangeStatistics(i,j,s,t,tr,true,matyn);
end;

procedure ERGMStatistics(var s: pvec; const symm: boolean;
                         ordeg: array of integer; const generate: boolean;
                         Neffects:TEffects);
(* Calculates the statistics from scratch and stores them in array s. *)
begin
  NEffects.CalcStatistics(s,symm,ordeg,generate,matyn);
end;



procedure UtilityComponents_yn(const i,k,h,n: integer;
         var s: rrarray; perm: array of boolean;
         Neffects: TEffects);
begin
  NEffects.UtilityComponents(i,k,h,n,s,perm,matyn);
end;

function utilitygain(const i,j: integer; const k: byte; const sym: boolean;
                     const RecMod: boolean; Neffects:TEffects): real;
begin
  if sym then
    begin
    if yn(i,j) then result := -NEffects.Contrib_n(matyn,i,j,k,true,RecMod)
     else  result := NEffects.Contrib_fn(matyn,i,j,k,false,RecMod);
    end
   else
    begin
    if yn(i,j) then result := -NEffects.Contrib_n(matyn,i,j,k,(matyn.Tievalue(j,i)>= 1),RecMod)
     else  result := NEffects.Contrib_fn(matyn,i,j,k,(matyn.Tievalue(j,i)>= 1),RecMod);
    end
end;

function utilitygain(const yy: boolean;
                     const i,j: integer; const k: byte; const sym: boolean;
                     const RecMod: boolean; Neffects:TEffects): real;
(* This overload is defined for use in cases where the value of yn(i,j)
   has to be utilized more than once, so it does not need to be
   looked up more than necessary. *)
(* Currently only in the symmetric case; for use in the non-directed case,
   check if yn(j,i) also can be determined externally *)   
begin
  if sym then
    begin
    if yy then result := -NEffects.Contrib_n(matyn,i,j,k,true,RecMod)
     else  result := NEffects.Contrib_fn(matyn,i,j,k,false,RecMod);
    end
   else
    begin
    if yy then result := -NEffects.Contrib_n(matyn,i,j,k,(matyn.Tievalue(j,i)>= 1),RecMod)
     else  result := NEffects.Contrib_fn(matyn,i,j,k,(matyn.Tievalue(j,i)>= 1),RecMod);
    end
end;


procedure CreateDifMat(const k: integer; var dg: TValuedDigraph);
begin
  DifNoNo(dg,matym[k],matyn,matmis[k],matmis[k+1],
       matstruct[k],matstruct[k+1]);
end;

procedure SumDifMat(const k: integer; var s,sup,sdown, stot: integer);
begin
   CountDifNoNo(s,sup,sdown, matym[k],matym[k+1],matmis[k],matmis[k+1],
       matstruct[k],matstruct[k+1]);
//CS: last two entries above added, 3.17i, to rule out counting of structurally precluded changes
   stot := matym[k].Order*(matym[k].Order-1) - OrCount(matmis[k],matmis[k+1]);
end;

procedure ForceStructurals(const k: integer);
(* Replaces matyn(i,j) for which matstruct[k](i,j) <> 0
   by matym[k](i,j). *)
begin
  if ExistStruct[k] then ReplaceTwo(matyn,matym[k],matstruct[k]);
end;

procedure ForceNewStructurals(const k: integer);     // added 3.17s
(* Replaces matyn(i,j) for which matnewstruct[k](i,j) <> 0
   by matym[k](i,j). *)
begin
  if ExistNewStruct[k] then ReplaceTwo(matyn,matym[k],matnewstruct[k]);
end;




procedure CreateLostMat(const k: integer; var dg: TValuedDigraph);
begin
  OldLost(dg,matym[k],matyn,matmis[k+1]);
end;

//procedure CreateDatMat(const k: integer; var dg: TValuedDigraph);
//begin
//  YesNoNo(dg,matyn,matmis[k],matmis[k+1]);
//end;

procedure ImputePrevious(const k: integer);
(* Replaces matym[k](i,j) for which matym[k-1](i,j) <> 0 and matmis[k](i,j)=1,
   by matym[k-1](i,j). *)
begin
  AddMinTwo(matym[k],matym[k-1],matmis[k]);
end;



(*procedure CreateDatMat1(const k: integer; var dg: TValuedDigraph);
begin
  YesNo(dg,matyn,matmis[k]);
end;
*)

{procedure ReduceMat(const k: integer; var dg: TValuedDigraph);
begin
  ReduceYesNo(dg,matmis[k+1]);
end;
}

procedure MissingsToSaveMat(const k: integer);
(* Calculates SaveMat,
   with SaveMat(i,j) = max(ynv(i,j), min(mis(i,j,k), mis(i,j,k+1))),
   and subtracts SaveMat from ynv; note that ynv = current network. *)
begin
  SplitTwo(matyn,matsavemat,matmis[k],matmis[k+1]);
end;

procedure RestoreSaveMat;
(* calculates ynv = max(ynv, SaveMat) *)
begin
  MaxTwo(matyn,matSaveMat);
end;

procedure CalcFunctions_yn(const k: integer; var fun: array of real; var ordeg: array of integer;
            Effs:TEffects; const CalcOD: Boolean; var outfile: TFileStream);
var dg: TValuedDigraph;
i: integer;
begin
  YesNoNo(dg,matyn,matmis[k],matmis[k+1]);
  Effs.CalcFunctions_f(fun,ordeg,k,dg,CalcOD);
  dg.Destroy(outfile,'dg 1 in CalcFunctions_yn');
  if (Effs.NumberStats_g >= 1) then begin
    OldLost(dg,matym[k],matyn,matmis[k+1]);
    Effs.CalcFunctions_g(fun,k,dg);
(* It will be more efficient to integrate CalcFunctions_g with CalcFunctions_f;
   if this can be handled well with the missings.
   But this function is not called extremely often. *)
    dg.Destroy(outfile,'dg 2 in CalcFunctions_yn');
   end;
end;

procedure CalcComponents_fa(i,nf,af,nuz: integer; kk,per: byte; dif: shortint;
                            initial: boolean;
                            var eff: BBarray; var fa_comps: array_0_20;
                            var outf: TFileStream);
  (* Calculates ingredients for contr_fa for different Action effects h
     for actor i in period per;
     for change dif in dependent action variable kk;
     if initial then for network of time per,
     else for current network;
     if dif = 0, for the given values (no difference);
     must be called before calling contr_fa.*)
var h,j,red,nmd: integer;
    pt,ptr,ptrn,n,start: integer;
    it,itr: TDoublyLinkedDoubleInt;
(* Changed 27/10/8: *)
    sr, sra, yaid, yam, zam: real;
(* End changed 27/10/8 *)
    yai, yaj: Byte;
    recip: boolean;
    mat: TValuedDigraph;
procedure UpdateComp;
var h: integer;
begin
  inc(nmd);
  if recip then inc(red);
    if (eff[start + 2] or eff[start + 3] or eff[start + 19]) then
        (* similarity to out-neigbors *)
           begin
           fa_comps[2] := fa_comps[2] + sr;
           end;
    if (eff[start + 7] or eff[start + 8]) then
           begin
        (* similarity to reciprocal neigbors *)
           if recip then fa_comps[7] := fa_comps[7] + sr;
           end;
    if (eff[start + 9] or eff[start + 10]) then
           begin
              (* similarity  popularity alter               *)
           fa_comps[9] := fa_comps[9] + sr*mat.InDegree(j);
           fa_comps[10] := fa_comps[10] + mat.InDegree(j);
           end;
    if (eff[start + 11] or eff[start + 12]) then
           begin
            (*   similarity  reciprocity  popularity alter  *)
           if recip then fa_comps[11] := fa_comps[11] + sr*mat.InDegree(j);
           end;
    if (eff[start + 13]) then
           begin
        (* average alter effect *)
           fa_comps[13] := fa_comps[13] + yaj;
           end;
    if (eff[start + 14]) then
           begin
         (* average reciprocal alter effect *)
           if recip then fa_comps[14] := fa_comps[14] + yaj;
           end;

(* Added 27/10/8: *)
        (* similarity to higher out-neigbors *)
           begin
           if (yaj > yai) or (yaj > yai+dif)
             then fa_comps[15] := fa_comps[15] + sr
            else
              begin
              if (dif = 0) then fa_comps[15] := fa_comps[15] + 1.0; (* 1 = similarity between equals *)
              end;
        (* one-sided average alter effect *)
           fa_comps[16] := fa_comps[16] + sra;
           end;
(* end Added 27/10/8 *)
end; (* UpdateComp *)
begin
  for h:=1 to 17 do fa_comps[h] := 0.0;
  red := 0;
  nmd := 0;
  yai := ya(i,kk);
  zam := zamean[kk];
  if initial then
    begin
    writeln;      // he never comes here!!!
    mat := matym[per];
//    yai := round(zc[i,kk,per]);
    if (dif <> 0) then writeln; // rough check, this should not occur.
    end
   else
    begin
    mat := matyn;
//    yai := ya(i,kk);
    end;
  n := mat.Order;
  it := mat.OutTies[i].First;  (* for going through the outties *)
  itr := mat.InTies[i].First; (* to check reciprocity: also go through the inties *)
  j := 0;
(* ptr is the smallest j' >= the previous j, for which there is a (j', i) tie;
   and n+1 if there is no such tie;
   ptrn is the next j' for which there is a (j',i) tie. *)
  if itr.HasNext then
    begin
    ptr := 0;
    itr := itr.next;
    ptrn := itr.Value1;
    end
   else
    begin
    ptr := n+1;
    ptrn := ptr;
    end;
  while it.HasNext do
    begin
    it := it.Next;
    j := it.value1;
(* there is a (i,j) tie *)
    while (ptr < j) do if itr.HasNext then
      begin
      ptr := ptrn;
      itr := itr.next;
      ptrn := itr.Value1;
      end
     else
      begin
      ptr := ptrn;
      ptrn := n+1;
      end;
(* there is a tie (i,j); and if (ptr <= j) there is also a tie (j,i) *)
    recip := (ptr <= j);    (* the tie is reciprocal *)
    start := nf + (kk-1)*af;

    if (dif = 0) then
(* For use in ActionStatistics: at the start or end of a run *)
      begin
      if (not (miszc(j,kk,per) or miszc(j,kk,per+1))) then
        begin
//        if initial then yaj := round(zc[j,kk,per]) else yaj := ya(j,kk);
        yaj := ya(j,kk);
        sr := similarity(yai,yaj,nuz+kk);
(* Start changed 27/10/8 *)
        if (yai >= yaj) then sra := sqr(yai-zam)
           else sra := (yai-zam)*(yaj-zam);
(* End changed 27/10/8 *)
        UpdateComp;
        end;
      end
     else  (* dif <> 0; this implies (not initial) *)
      begin
      yaj := ya(j,kk);
(* For use in RunStep, CalcRatelprob, and UtilityComponents_ya: during a run *)
      sr := similarity(yai+dif,yaj,nuz+kk)
              - similarity(yai,yaj,nuz+kk);
(* Start changed 27/10/8 *)
      yaid := yai + dif;
      if (yai > yaj) then yam := yai else yam := yaj;
      if (yaid > yaj) then
        sra := (yai-zam)*(yaid - yam) + dif*(yaid - zam)
       else
        sra := (yai-zam)*(yaj - yam) + dif*(yaj- zam);
(* End changed 27/10/8 *)
      UpdateComp;
      end;
(* if fa_comps[higher than 16] are needed, also initialise them to 0 at the start! *)
    end; (* while it.Hasnext *)
  fa_comps[0] := nmd;
  fa_comps[1] := red; (* reciprocated degree of i *)

//  if (red <> mat.ReciDegree(i)) then
 // writeln;
//   this is just for checking; only for the case dif <> 0
//      writeln(outf,'error: red ',i:4,' ',red:10,' ',mat.ReciDegree(i):10)
//      else writeln(outf,'OK   : red ',i:4,' ',red:10,' ',mat.ReciDegree(i):10);
end; (* CalcComponents_fa *)



function SaveMat(i,j: Integer): ShortInt;
begin
  SaveMat := matSaveMat.TieValue(i,j);
end;

procedure DefineSaveMat(i,j: Integer; v: ShortInt);
begin
  matSaveMat.ChangeTieValue(i,j,v);
end;

procedure ReadData(const fn: shortstring; const nh, k: integer;
                   var vmin, vmax: shortint);
(* Reads valued digraph data file for observation k from file with name fn *)
begin
  matym[k].ReadFromFile(nh,fn);
  vmin := matym[k].mini;
  vmax := matym[k].maxi;
end;

procedure SetDigraph(const k: integer);
(* Sets current valued digraph to its value for observation k. *)
begin
  CopyGraph(matym[k],matyn);
end;

procedure MakeDigraph(const k: integer);
(* Sets the digraph for observation k to the current digraph. *)
begin
  CopyGraph(matyn,matym[k]);
end;

procedure CopyCurrentDigraph(var dg: TValuedDigraph; const k: integer);
(* Sets dg to current valued digraph. *)
begin
  CopyGraph(matyn,dg);
end;

procedure ChangeDataValues(const a,b: array of shortint; const maxab: byte;
                           const k: integer; kind: byte);
(* Changes values in observation k equal to a[h] into b[h],
   and notes this in structural zeros if kind = 1, in missings if kind = 2,
   not at all if kind = 3. *)
var matdum: TDigraph;
    n: integer;
    ouf: TFileStream;
begin
  case kind of
    1: ChangeValues(a,b,maxab,matym[k],matstruct[k]);
    2: ChangeValues(a,b,maxab,matym[k],matmis[k]);
    3: begin
       n := matym[1].Order;
       matdum:= TDigraph.Create(n);
       ChangeValues(a,b,maxab,matym[k],matdum);
       matdum.Destroy(ouf,'');
       end;
  end;
end;

procedure SetStructOne(const k: integer);
(* Sets current valued digraph to only the structural ones for observation k. *)
begin
  CopyMinTwo(matyn,matym[k],matstruct[k]);
end;


// added 3.17s:
procedure SetChangeStruct(obs: integer);
var h: integer;
begin
  for h:=2 to obs do if ExistStruct[h] then
    begin
    ExistNewStruct[h] := NewArcs(MatStruct[h-1],MatStruct[h]);
    if ExistNewStruct[h] then
      begin
      YesNo(MatNewStruct[h],MatStruct[h],MatStruct[h-1]);
(* MatNewStruct[h] now is (MatStruct[h] and (not MatStruct[h-1])). *)
      end;
    end
   else ExistNewStruct[h] := false;
end; (* SetChangeStruct *)

procedure ReadPajekDataFile(const fn: shortstring; var ert: shortstring;
                             var ext: boolean;
                             const k: integer; var vmin,vmax: shortint);
(* Reads valued digraph data file for observation k from Pajek file with name fn *)
begin
  matym[k].ReadPajekFile(fn,ert,ext);
  vmin := matym[k].mini;
  vmax := matym[k].maxi;
end;

procedure ReadPajekDataFileToMatrix(const fn: shortstring;
                             const nh: integer; var ert: shortstring;
                             var ext: boolean;
                             var zout: iiarray; var vmin,vmax: shortint);
(* Reads valued digraph data file for observation k from Pajek file with name fn *)
var maty: TValuedDigraph;
begin
  vmin := 0;
  vmax := 0;
  maty := TValuedDigraph.Create(nh);
  maty.ReadPajekFile(fn,ert,ext);
  if (ert > '') then exit;
  vmin := maty.mini;
  vmax := maty.maxi;
  maty.WriteToArray(zout);
  maty.Destroy;
end;


procedure ReadDigraph(const fn: shortstring; const nh, k: integer);
(* Reads valued digraph data file for observation k
  from file with name fn
  and copies this to the current digraph. *)
begin
  matym[k].ReadFromFile(nh,fn);
  CopyGraph(matym[k],matyn);
end;

procedure WriteData(const fn: shortstring; const k: integer);
(* Writes valued digraph data for observation k to file with name fn in internal Siena format *)
begin
  matym[k].WriteToFile(fn);
end;

procedure WriteTheseData(const fn: shortstring);
(* Writes current valued digraph data to file with name fn in internal Siena format *)
begin
  matyn.WriteToFile(fn);
end;


procedure WriteSimulatedData(const uppertr: boolean; var f: textfile);
(* Writes current valued digraph data file to file f *)
begin
  matyn.WriteEdgeList(uppertr,f);
end;

procedure WriteSubsetSimulatedData(
       const z1: iarray; const z2: boarray; var f: textfile);
(* Writes current valued digraph data file for subset of nodes to file f *)
var i,d: integer;
  it:TDoublyLinkedDoubleInt;
begin
  if (matyn.Order < 100) then d := 2 else d := 3;
  if (matyn.Order > 999) then d := 4;
  for i:=1 to matyn.Order do if z2[i] then
    begin
    it := matyn.OutTies[i].First;
    while it.HasNext do begin
      it := it.Next;
      if z2[it.Value1] then
        writeln(f,z1[i]:d,' ',z1[it.Value1]:d,' 1');
     end;
   end; (* for i *)
end;

procedure ReadMisData(const fn: shortstring; const nh, k: integer);
(* Reads missing data file for observation k from file with name fn *)
begin
  matmis[k].ReadFromFile(nh,fn);
end;

procedure WriteMisData(const fn: shortstring; const k: integer);
(* Writes missing data file for observation k to file with name fn *)
begin
  matmis[k].WriteToFile(fn);
end;

procedure ReadStructData(const fn: shortstring; const nh, k: integer);
(* Reads structurally determined data file for observation k from file with name fn *)
begin
  matStruct[k].ReadFromFile(nh,fn);
end;

procedure WriteStructData(const fn: shortstring; const k: integer);
(* Writes structurally determined data file for observation k to file with name fn *)
begin
  matStruct[k].WriteToFile(fn);
end;


(* AA change *)
procedure StructuralDataValues(const k: integer; var a: array of integer);
(* gives for all h the number of structurally determined values equal to h
   in a[h-mini], where mini is the minimum value (usually 0). *)
var valmin, valmax, h: shortint;
    st: integer;
begin
 valmin := matym[k].Mini;
 valmax := matym[k].Maxi;
 ValuesCount(matym[k],matstruct[k],valmin,valmax,a);
 if (valmin <= 0) and (valmax >= 0) then
   begin
   st := 0;
   for h := valmin to valmax do if (h <> 0) then inc(st,a[h-valmin]);
   a[-valmin] := structelements(k)-st;
   end;
end;



{procedure bal(const n,m: integer); // @@@
var i,j,h,yih,yjh: integer;
begin
  bbal := 0;
  bbaltot := 0;
  bties := 0;
  for i:=n downto 1 do for j := n downto 1 do
   if
   (not mis(i,j,m)) and (not mis(i,j,m+1)) and
   yn(i,j) then begin
     inc(bties);
     for h:= n downto 1 do begin
    if (yn(i,h) and not mis(i,h,m) and (i <> h)) then yih := 1 else yih := 0;
    if (yn(j,h) and not mis(j,h,m) and (j <> h)) then yjh := 1 else yjh := 0;
        if ((not mis(i,h,m+1)) and (not mis(j,h,m+1)) and (j <> h) and (i <> h))
         then
         begin (* if potential ties to actor h are observed also at the end of the period.. *)
           if (yih <> yjh) then Inc(bbal); (* ..calculate imbalance of i and j with respect to h *)
              inc(bbaltot); (* ..and update number of triplets ijh on which this calculation is based. *)
            end; (* if not missings *)
         end; (* h *)
        end; (* if *)
end;
}


procedure DescribeCliques(var outf: TFileStream; const red, smin, smax: integer);
var i,j,h,smi,sma: integer;
  t: shortint;
  nc: array of integer;
  elt: boolean;
  ssmin,ssmax: integer;
begin
  matym[1].allcliques(false,outf,smin,smax);
//  for i:=1 to red do
//    matym[1].reducedcliques(false,outf,i,smin,smax);



  setlength(nc,smax+2);
{
  for i:=1 to matyn.Order do for j :=1 to matyn.Order do
    begin
    matyn.AllCliques_ij(false,outf,smin,smax,i,j,t,nc);
//    for h:=0 to smax do write(outf,nc[h]:3); writeln(outf);
    elt := false;
    for h:=smin to smax do if (nc[h] > 0) then elt := true;
    if elt then
      begin
      write(outf,i:3,j:3,'  ');
      if (t >= 1) then write(outf,' +  ') else write(outf,' -  ');
      for h:=smin to smax do write(outf,nc[h]:3); writeln(outf);
      end;
    end;
 }
{
 i:=1; j := 2;
//  for i:=1 to matyn.Order do for j :=1 to matyn.Order do
    begin
    t := ynv(i,j);
writeln(outf,i:3,j:3,' :::');
//for smi := 2 to 3 do for sma := 5 to 8 do
begin
    matyn.All_red1_Cliques_ij(true,outf,smin,smax,i,j,t,nc);
   for h:=0 to smax do write(outf,nc[h]:3,' '); writeln(outf);
  end;

end;
}
end;


procedure DescribeCovariate(const n: integer;
          var z: array of real; const zname: shortstring;
          var mat: TvaluedDigraph;
          var outf: TFileStream);
(* This procedure gives some plots of incoming and outgoing ties
   dependent on mean splits of actor covariates *)
var i,j,h,erc: integer;
  x1,x2: array of real;
  vn,vp: array of integer;
  error: boolean;
  x: shortstring;
  nsin,nsin2,nsout,nsout2,psin,psin2,psout,psout2: real;
  nn,np: integer;
begin
  if (high(z) < n) then exit;
  wwriteln(outf,zname);
  SetLength(x1,n+1);
  SetLength(x2,n+1);
  SetLength(vn,n+1);
  SetLength(vp,n+1);
  x1[0] := 0.0;
  x2[0] := 0.0;
  vn[0] := 0;
  vp[0] := 0;
  nn := 0; np := 0;
  for i:=n downto 1 do
    begin
    if (z[i] <= 0.0) then
      begin
      vn[i] := 1;
      vp[i] := 0;
      inc(nn);
      end
     else
      begin
      vn[i] := 0;
      vp[i] := 1;
      inc(np);
      end;
    end; (* i *)
    wwriteln(outf,inttostr(nn)+' actors below or equal to the mean; '+inttostr(np)+' above the mean.');

{    error := false;
    for i:=n downto 1 do
      begin
      x1[i] :=  ArrayInnerProduct(mat.OutTies[i],vn,erc);
      if (erc <> 0) then error := true;
      x2[i] :=  ArrayInnerProduct(mat.OutTies[i],vp,erc);
      if (erc <> 0) then error := true;
      end;
    if (not error) then
      plot(outf,x1,x2,nil,n,'outdegree to low '+zname,'outdegree to high '+zname);
    outf.write(pchar(eol)^,length(eol));//writeln(outf);

    error := false;
    for i:=n downto 1 do
      begin
      x1[i] :=  ArrayInnerProduct(mat.InTies[i],vn,erc);
      if (erc <> 0) then error := true;
      x2[i] :=  ArrayInnerProduct(mat.InTies[i],vp,erc);
      if (erc <> 0) then error := true;
      end;
    if (not error) then
      plot(outf,x1,x2,nil,n,'indegree to low '+zname,'indegree to high '+zname);
    outf.write(pchar(eol)^,length(eol));//writeln(outf);
}
    error := false;
    nsin := 0; nsin2 := 0;
    nsout := 0; nsout2 := 0;
    psin := 0; psin2 := 0;
    psout := 0; psout2 := 0;
    for i:=n downto 1 do
      begin
      if (vn[i] >= 1) then
        begin
        nsin := nsin+mat.InDegree(i);
        nsin2 := nsin2+sqr(mat.InDegree(i));
        nsout := nsout+mat.OutDegree(i);
        nsout2 := nsout2+sqr(mat.OutDegree(i));
        end
       else
        begin
        psin := psin+mat.InDegree(i);
        psin2 := psin2+sqr(mat.InDegree(i));
        psout := psout+mat.OutDegree(i);
        psout2 := psout2+sqr(mat.OutDegree(i));
        end;
      end;

     if (nn >= 2) and (np >= 2) then
       begin
       nsin := nsin/nn;
       nsin2 := (nn/(nn-1))*(nsin2/nn - sqr(nsin));
       nsout := nsout/nn;
       nsout2 := (nn/(nn-1))*(nsout2/nn - sqr(nsout));
       psin := psin/np;
       psin2 := (np/(np-1))*(psin2/np - sqr(psin));
       psout := psout/np;
       psout2 := (np/(np-1))*(psout2/np - sqr(psout));
       wwriteln(outf,'Indegrees :');
       wwriteln(outf,'Actors with '+zname+' smaller than or equal to the mean:');
       str(nsin:10:4,x);
       wwrite(outf,'Average '+x+', ');
       str(root(nsin2):10:4,x);
       wwriteln(outf,'st. dev. '+x+';');
       wwriteln(outf,'Actors with '+zname+' greater than the mean:');
       str(psin:10:4,x);
       wwrite(outf,'Average '+x+', ');
       str(root(psin2):10:4,x);
       wwriteln(outf,'st. dev. '+x+';');
       wwriteln(outf,'Outdegrees :');
       wwriteln(outf,'Actors with '+zname+' smaller than or equal to the mean:');
       str(nsout:10:4,x);
       wwrite(outf,'Average '+x+', ');
       str(root(nsout2):10:4,x);
       wwriteln(outf,'st. dev. '+x+'; ');
       wwriteln(outf,'Actors with '+zname+' greater than the mean:');
       str(psout:10:4,x);
       wwrite(outf,'Average '+x+', ');
       str(root(psout2):10:4,x);
       wwriteln(outf,'st. dev. '+x+'.');
       end;

     wwriteln(outf,'');
end; (* DescribeCovariate *)

procedure DescribeNetworksCovariate(const nn, nz, nk: integer;
          var outf: TFileStream);
var h,k,i: integer;
   zy: array of real;
begin
  SetLength(zy,nn+1);
  zy[0] := 0;
  for h:=1 to nz do begin
    for i:=1 to nn do zy[i] := z[i,h];
    for k:=1 to nk do DescribeCovariate(nn,zy,
          ZNames[h]+', period '+inttostr(k),matym[k],outf);
    end;
end;

procedure DescribeNetworksCCovariate(const nn, nz, nzc, nk: integer;
          var outf: TFileStream);
var h,k,i: integer;
   zy: array of real;
begin
  SetLength(zy,nn+1);
  zy[0] := 0;
  for h:=1 to nzc do
    begin
    for k:=1 to nk-1 do
      begin
      for i:=1 to nn do zy[i] := zc[i,h,k];
      DescribeCovariate(nn,zy,znames[nz+h]+', period '+inttostr(k),matym[k],outf);
      end;
    end;
end;


procedure DistancesFromHighDegree(var d,z: iarray);
begin
  matyn.ComponentOne(d,z);
end;

procedure SetPeriods(nh,nhzc,nhzz,nhzzc,per: integer);
var i,j,h: Integer;
begin
 (* dependent network variables *)
 if (not matricescreated) then matyn := TValuedDigraph.Create(nh);
 h := High(matym);
 if (h < 0) then h := 0;
 SetLength(matym,per+2);
 for i:=h+1 to per+1 do matym[i]:= TValuedDigraph.Create(nh);
// for i:=1 to per+1 do matym[i]:= TValuedDigraph.Create(nh);  // abcdefg
 SetLength(matmis,per+2);
 for i:=h to per+1 do
   begin
   matmis[i]:= TDigraph.Create(nh);
   matmis[i].PermitDiagonal;
   end;
 SetLength(matstruct,per+2);
 SetLength(matnewstruct,per+2);  // added 3.17s
 for i:=h to per+1 do matstruct[i]:= TDigraph.Create(nh);
 for i:=h+1 to per+1 do matnewstruct[i]:= TDigraph.Create(nh);  // added 3.17s

 (* dependent behavioural variables *)
 SetLength(matya,nh+1,nhzc+1);  (* TS: this is a bit too much, as it now is
                                       measured to fit ALL changing covariates *)
 (* covariates *)
 SetLength(z,nh+1,nzmax+1); (* array is used also for changing actor covariates *)
 SetLength(matmisz,nh+1,nzmax+1); (* --"-- *)
 SetLength(zx,nh+1,3);

 SetLength(zc,nh+1,nhzc+1,per+2);
// SetLength(matmiszc,nh+1,nhzc+1,per+2);
 SetLength(matmiszc,nh+4,nhzc+4,per+4);

 SetLength(zz,nh+1,nh+1,nhzz+1);
 SetLength(matmiszz,nh+1,nh+1,nhzz+1);

 SetLength(zzc,nh+1,nh+1,nhzzc+1,per+2); 

 (* similarity means *)
 SetLength(zcsglobalmean,nhzc+1);

// SetLength(SaveMat,nh+1,nh+1);
 if (not MatricesCreated) then matSaveMat:= TValuedDigraph.Create(nh);

 SetLength(SimDegrees,nh+1,nh+1);
 SetLength(mi,nhzc+1,per+1);
 SetLength(gc,nhzc+1,per+1);

 if (per <= 1) then Setlength(cumuldegrees,nh+1);

 for i:=1 to nh do for j:=1 to nhzc do matya[i,j] := 0;

 for i:=1 to nh do begin
//   for j:=1 to nzmax do z[i,j] := 0.0;
   for j:=0 to nh do SimDegrees[i,j] := 0;
//   for j:=1 to nh do for h:=1 to nhzz do zz[i,j,h] := 0;
 end;
 MatricesCreated := true;
end; (* SetPeriods *)


procedure FinishEight(var outf: TFileStream);
var i,i0,n: integer;
var x : string;
begin
  if not MatricesCreated then exit;
  n := matyn.order;
  matyn.Destroy(outf,'matyn');
  finalize(matya);
  n := high(matym);
  i0 := low(matym);
  if (i0 <= 0) then i0 := 1;
  for i:=n downto i0 do
    begin
    if (not matym[i].IsCorrect) then
      begin
      matym[i].writelists(outf);
      x := format('%s%d%s',['Internal representation of network for observation ',i,' is incorrect.'])+eol;
      outf.Write(pchar(x)^,length(x));
      end;
    matym[i].Destroy(outf,'matym['+inttostr(i)+']');
    end;
  Setlength(matym,0); // abcdefg
  n := high(matmis);
  i0 := low(matmis);
  if (i0 <= 0) then i0 := 1;
  for i:=n downto i0 do matmis[i].Destroy(outf,'matmis['+inttostr(i)+']');
  Setlength(matmis,0); // abcdefg
  n := high(matstruct);
  i0 := low(matstruct);
  if (i0 <= 0) then i0 := 1;
  for i:=n downto i0 do matstruct[i].Destroy(outf,'matstruct['+inttostr(i)+']');
  Setlength(matstruct,0); // abcdefg
  for i:=n downto i0 do
    matnewstruct[i].Destroy(outf,'matnewstruct['+inttostr(i)+']');
  Setlength(matnewstruct,0);

  finalize(z);
  finalize(zc);
  finalize(zz);
  finalize(zzc);
  finalize(zcsglobalmean);
  if assigned(matSavemat) then matSaveMat.Destroy(outf,'MatSaveMat');
  finalize(SimDegrees);
//xxx := xxx + ' finisheight xxx ';  
end;

{
The following was used to check the first-mentioned two functions.
function DDist2Contr(const i,j: integer): integer;
(* Contribution of tie (i,j) to number of distances 2:
   sum_(i,j) yn(i,j) Dist2contr(i,j) is the number of directed distances 2.
   There are many functions Dist2contr(i,j) with this property.
   The function used here was derived as follows:
   Dist2contr(i,j) is the change in number of distances 2
   from node i to any other node,
   when the arc i => j is added,
   given that all arcs i => h for h > j are considered absent.
   In other words, the arcs in row i are added one by one,
   and for each j, Dist2contr(i,j) is the change in distances 2
   emanating from node i. *)
var it1,it2: TDoublyLinkedDoubleInt;
 h: integer;
begin
  if matyn.NoTwoPath(i,j,j) then Result := 0 else result := -1;
  it1 := matyn.outties[i].first;
  it2 := matyn.outties[j].first;
  h := 0;
  while (it1.HasNext and it2.HasNext) do // and (h < j)) do
    begin
    case sign(it1.Next.Value1 - it2.Next.Value1) of
     -1:  it1 := it1.Next;
      0:  begin
          it1 := it1.Next;
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h and from i to h *)
          if (j < h) and matyn.NoTwoPath(i,h,j) then Result := Result+1;
          end;
      1:  begin
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h but not from i to h *)
          if (i <> h) and matyn.NoTwoPath(i,h,j) then Result := Result+1;
          end;
      end; (* case *)
    end;
  while it2.HasNext do
    begin
    it2 := it2.Next;
    h := it2.Value1; (* there is a tie from j to h but not from i to h *)
    if (i <> h) and matyn.NoTwoPath(i,h,j) then Result := Result+1;
    end;
end; (* Dist2Contr *)

function DTwoDist2Contr(const i,j: integer): integer;
(* Contribution of tie (i,j) to number of TWO-distances 2:
   A two-distance 2 is (i,j) where there are at least two h with i >= h >= j.
   sum_(i,j) yn(i,j) DTwoDist2contr(i,j) is the number of directed two-distances 2.
   There are many functions with this property.
   The function used here was derived as follows:
   DTwoDist2contr(i,j) is the change in number of two-distances 2
   from node i to any other node,
   when the arc i => j is added,
   given that all arcs i => h for h > j are considered absent.
   In other words, the arcs in row i are added one by one,
   and for each j, DTwoDist2contr(i,j) is the change in two-distances 2
   emanating from node i. *)
var it1,it2: TDoublyLinkedDoubleInt;
 h: integer;
begin
  if matyn.NoOrOneTwoPath(i,j,j) then Result := 0 else result := -1;
  it1 := matyn.outties[i].first;
  it2 := matyn.outties[j].first;
  h := 0;
  while (it1.HasNext and it2.HasNext) and (h < j) do
    begin
    case sign(it1.Next.Value1 - it2.Next.Value1) of
     -1:  it1 := it1.Next;
      0:  begin
          it1 := it1.Next;
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h and from i to h *)
          if (j < h) and matyn.OneTwoPath(i,h,j) then Result := Result+1;
          end;
      1:  begin
          it2 := it2.Next;
          h := it2.Value1; (* there is a tie from j to h but not from i to h *)
          if (i <> h) and matyn.OneTwoPath(i,h,j) then Result := Result+1;
          end;
      end; (* case *)
    end;
  while it2.HasNext do
    begin
    it2 := it2.Next;
    h := it2.Value1; (* there is a tie from j to h but not from i to h *)
    if (i <> h) and matyn.OneTwoPath(i,h,j) then Result := Result+1;
    end;
end; (* DTwoDist2Contr *)

procedure CompareDist2(n: integer; var d1, d2: integer);
var i, j: integer;
begin
  for i:=1 to n do for j:=1 to n do
    if (i <> j) then
    begin
    if (random < 0.1) then setyn(i,j,true) else setyn(i,j,false);
    end
     else setyn(i,j,false);
}
  {d1 := 0;
  for i:=1 to n do for j :=1 to n do
    if yn(i,j) then d1 := d1 + DDist2contr(i,j);
  d2 := 0;
  for i:=1 to n do for j := 1 to n do
    if (i <> j) and (matyn.twopaths(i,j) >= 1) and (not yn(i,j)) then inc(d2);
 }
{
  d1 := 0;
  for i:=1 to n do for j :=1 to n do
    if yn(i,j) then d1 := d1 + DTwoDist2contr(i,j);
  d2 := 0;
  for i:=1 to n do for j := 1 to n do
    if (i <> j) and (matyn.twopaths(i,j) >= 2) and (not yn(i,j)) then inc(d2);

end;
 }

begin (* EIGHT *)
 MatricesCreated := false;
// xxx := 'xxx = ';
end.
