Unit rangen;

(* Uses random numbers from Urns;
   versions 0-2 are from Press, Teukolsky & Vetterling's book,
   versions 3-4 are from Marsaglia. *)

Interface

uses s_constants, slib;

const VersionMax = 4;
   (* Random number generators for version = 0..VersionMax *)
type matrixarray = array of array of real;

var version, version2: byte;
   gloldm,glsq,glalxm,glg: real; (* for POIDEV *)
   RanGenErrorMessage: string;

FUNCTION rando(version: Byte): Extended;
(* random number *)

PROCEDURE init(version: byte; var seed: integer; generate : boolean);
(* Initialises randomx, with x = version,
   and with a time-dependent random choice of seed (if 'generate' is true)
   or the given seed (when 'generate' is false ) *)

PROCEDURE SetFree(version: byte);
(* Frees random number generator *)

FUNCTION takeseed(version: byte): integer;

FUNCTION randomd(version: Byte; k: Integer): Integer;
(* random between 0 and k-1  *)

function ranp(version: Byte; const n: integer; var p: array of real): Integer;
(* draws at random between 0 and n according to probability vector p *)

function newranp(version: Byte; const n: integer; const p: array of real): Integer; overload;
function newranp(version: Byte; const n: integer; const p: array of integer): Integer; overload;
(* draws at random between 0 and n according to cumulative probability vector
   proportional to p *)
(* P(result <= i) = p(i)/p(n) *)

procedure matrixranp(version: Byte; const n: integer;
                    const p: matrixarray; var i,j: integer);
(* draws at random between 0 and n according to cumulative probability vector
   proportional to p *)
(* P((i,j) lexicographically <= (h,k)) = p(h,k)/p(n,n) *)

{
function rannp(version: Byte; var p: realvec; psum: real; n: integer;
               var pr: real): Integer;
(* draws at random between 1 and n according to
   nonnormalized probability vector p  with sum psum \;
   probability of realized ouctome is pr.   *)
}

(* For the following, the dependence on the version is implicit *)

FUNCTION gamma(w: Extended):Extended;
(* Gamma distribution according to Ahrens & Dieter *)

FUNCTION beta1(b1,b2: Extended):Extended;
(* Beta distribution according to Ahrens & Dieter *)

FUNCTION beta2(b1,b2: Extended):Extended;
(* Beta distribution according to Narayanan & Johnk *)

FUNCTION normrand: Extended;
(* Normal distribution according to Box-Muller *)

FUNCTION poidev(xm: real): integer;
(* Poisson random deviate from Numerical Recipes *)

procedure d7r8 (var xm: real; var f: textfile);
(* driver for routine POIDEV from Numerical Recipes  *)

Implementation

uses SysUtils,random; (* from URNS = Ultimate Random Numbers Suite *)

VAR
   i: integer;
   ridum: array[0..VersionMax] of Integer; 
(* ridum just checks whether the random number generator
   has been initialised or not *)
   nrb: boolean; (* nrb, nru used in NormRand *)
   nru: Real;
   ran0: TeoRandomGood;
   ran1: TeoRandomBetter;
   ran2: TeoRandomBest;
   ran3: TeoRandomMarsagliaFast;
   ran4: TeoRandomMarsaglia;

PROCEDURE init(version: byte; var seed: integer; generate : boolean);
var Present: TDateTime;
 Hour, Min, Sec, MSec: Word;
begin
  if ((ridum[version] < 0) and generate)
  then begin
    Present:= Now;
    DecodeTime(Present, Hour, Min, Sec, MSec);
    seed := msec + 1000*sec + 60000*Min;
    if (seed = 0) then seed := 1;
  end;
  case version of
    0: begin
        if (ridum[0] < 0) then ran0 := TeoRandomGood.Create(seed)
         else ran0.Reset(seed);
        ridum[0] := 1;
       end;
    1: begin
        if (ridum[1] < 0) then ran1 := TeoRandomBetter.Create(seed)
         else ran1.Reset(seed);
        ridum[1] := 1;
       end;
    2: begin
        if (ridum[2] < 0) then ran2 := TeoRandomBest.Create(seed)
         else ran2.Reset(seed);
        ridum[2] := 1;
       end;
    3: begin
        if (ridum[3] < 0) then ran3 := TeoRandomMarsagliafast.Create(seed)
         else ran3.Reset(seed);
        ridum[3] := 1;
       end;
    4: begin
        if (ridum[4] < 0) then ran4 := TeoRandomMarsaglia.Create(seed)
         else ran4.Reset(seed);
        ridum[4] := 1;
       end;
   end; (* case *)
end;


PROCEDURE SetFree(version: byte);
(* Frees randomx, with x = version *)
begin
 case version of
  0: ran0.Free;      
  1: ran1.Free;
  2: ran2.Free;
  3: ran3.Free;
  4: ran4.Free;
 end; (* case *)
end;

FUNCTION rando(version: Byte): Extended;
begin
  case version of
     0: rando := ran0.Value;
     1: rando := ran1.Value;
     2: rando := ran2.Value;
     3: rando := ran3.Value;
     4: rando := ran4.Value;
   end; (* case *)
end; (* rando *)


FUNCTION takeseed(version: byte): integer;
var s: integer;
begin
  if (ridum[version] < 0) then init(version,s,true)
   else begin
    s := Trunc(MaxInt*rando(version));
    case version of
      0: ran0.reset(s);
      1: ran1.reset(s);
      2: ran2.reset(s);
      3: ran3.reset(s);
      4: ran4.reset(s);
     end; (* case *)
   end; (* if else *)
  takeseed := s;
end;




FUNCTION gamdev(ia: integer): Extended;
(* gamma variable for integer shape parameter *)
VAR
   am,e,s,v1,v2,x,y,h: Extended;
   j: integer;
BEGIN
   IF (ia < 1) THEN BEGIN
      RanGenErrorMessage := 'error: ia < 1 in routine GAMDEV';
      gamdev := 0.0;
      exit;
      END;
   IF (ia < 6) THEN BEGIN
      x := 1.0;
      FOR j := 1 to ia DO BEGIN h:=rando(version);
                          IF h>1.e-12 THEN x := x*h
                                 ELSE x :=x*1.e-12;
                          END;
      x := -ln(x);
   END ELSE BEGIN
      REPEAT
         REPEAT
            REPEAT
               v1 := 2.0*rando(version)-1.0; IF v1<1.e-12 THEN v1:=1.e-12;
               v2 := 2.0*rando(version)-1.0;
            UNTIL ((sqr(v1)+sqr(v2)) <= 1.0);
            y := v2/v1;
            am := ia-1;
            s := sqrt(2.0*am+1.0);
            x := s*y+am;
         UNTIL (x > 1.e-12);
         e := (1.0+sqr(y))*exp(am*ln(x/am)-s*y);
      UNTIL (rando(version) <= e)
   END;
   gamdev := x
END; (* gamdev *)

FUNCTION gammaran01(w: Extended):Extended;
LABEL 1;
VAR u1,u2,x0,x1,x2,k,x: Extended;
BEGIN
IF (w<=0) THEN
  BEGIN
    RanGenErrorMessage := 'error: w <= 0 in routine GAMMARAN01';
    gammaran01 := 0.0;
    exit;
  END;
1:    u1:=rando(version); u2:=rando(version);
      k:=1+w*exp(-1);
      x0:=k*u1;
      IF (x0>1) THEN BEGIN x2:=-ln((k-x0)/w);
                           IF (u2<=exp((w-1)*ln(x2))) THEN x:=x2
                                                      ELSE GOTO 1;
                     END
             ELSE BEGIN (* if u1 = 0 so that x0 = 0 then things go wrong *)
                           IF (x0<=1.e-12) THEN x0:=1.e-12;
                           x1:=exp(ln(x0)/w);
                           IF u2<=exp(-x1) THEN x:=x1
                                           ELSE GOTO 1;
                     END;
      gammaran01:=x;
END; (* gammaran01 *)

FUNCTION gammaran1(w: Extended):Extended;
VAR u,x,y: Extended;
BEGIN IF (w<=0) THEN
  BEGIN
    RanGenErrorMessage := 'error: w <= 0 in routine GAMMARAN1';
    gammaran1 := 0.0;
    exit;
  END;
      REPEAT BEGIN u:=rando(version);
                   y:=-ln(rando(version));
             END
      UNTIL (ln(u) <= (w-1)*(ln(y)+1-y));
      x:=w*y;
      gammaran1:=x;
END; (* gammaran1 *)

FUNCTION gammaran(w: Extended):Extended;
BEGIN IF (w<=0) THEN
  BEGIN
    RanGenErrorMessage := 'error: w <= 0 in routine GAMMARAN';
    gammaran := 0.0;
    exit;
  END;
      IF (w<1) THEN gammaran:=gammaran01(w)
          ELSE    gammaran:=gammaran1(w);
END; (* gammaran *)

FUNCTION beta1(b1,b2: Extended):Extended;
VAR y1,y2: Extended;
BEGIN y1:=gammaran(b1);
      y2:=gammaran(b2);
      if (y1 + y2 <> 0.0) then beta1:=y1/(y1+y2)
       else begin
       RanGenErrorMessage := 'error in beta1';
       beta1 := 0.0;
       end;
END;

FUNCTION beta2(b1,b2: Extended):Extended;
VAR u1,u2,y1,y2: Extended;
BEGIN REPEAT BEGIN u1:=rando(version); u2:=rando(version);
                   IF u1<=0 then u1:=0.0001;
                   IF u2<=0 then u2:=0.0001;
                   y1:=exp(ln(u1)/b1);
                   y2:=exp(ln(u2)/b2);
             END
      UNTIL (y1+y2 <=1);
      beta2:=y1/(y1+y2);
END;


FUNCTION NORMRAND: extended;
(* Normal distribution according to Box-Muller procedure *)
    VAR U1,U2,S: extended;
BEGIN
  if nrb then normrand := nru else begin
    REPEAT
      U1:=2*rando(version)-1;
      U2:=2*rando(version)-1;
      S:=SQR(U1)+SQR(U2);
     UNTIL S<1;
    if (S < 1e-40) then begin
      normrand := 0.0; nru := 0.0;
     end else begin
      NORMRAND:=U1*SQRT(-2*LN(S)/S);
      nru:=U2*SQRT(-2*LN(S)/S);
     end;
   end; (* if *)
  nrb := not nrb;
END; (* normrand *)

FUNCTION randomd(version: Byte; k: Integer): Integer;
var r: integer;
begin
  r := trunc(rando(version)*k);
  if (r < 0) then r := 0;
  if (r >= k) then randomd := k-1 else randomd := r;
end;

{
function rannp(version: Byte; var p: realvec; psum: real; n: integer;
               var pr: real): Integer;
(* draws at random between 1 and n according to
   nonnormalized probability vector p  with sum psum;
   probability of realized ouctome is pr.  *)
var i: integer; s,r: real;
begin
  r := psum*rando(version);
  s := 0.0;
  i:= 0;
  while ((s < r) and (i < n)) do begin
    inc(i);
    s := s + p[i];
   end;
  if (i <= 0) then rannp := 1 else rannp := i;
  pr := p[i]/psum;
end;
}

function ranp(version: Byte; const n: integer; var p: array of real): Integer;
(* draws at random between 0 and n according to probability vector p *)
var i: integer; s,r: real; OK: boolean;
begin
(* @@@ for checking, activate the following :
  OK := true; s := 0.0;
  i := -1;
  repeat
    inc(i);
    s := s+p[i]; if p[i] < 0.0 then OK := false;
   until (s >= 0.99999999) or (i >= nrg);
  if (abs(s - 1.0) > 1e-8) then OK := false;
  if not OK then begin
    ranp := -999;
   end;
    here the checks stop *)
  r := rando(version);
  s := 0;
  i:= -1;
  if (r >= 1.0) then r := 0.9999999999;
  repeat
    Inc(i);
    s := s+p[i];
   until (s >= r) or (i >= n);
  ranp := i;
end;


function newranp(version: Byte; const n: integer; const p: array of real): Integer;
(* draws at random between 0 and n according to cumulative probability vector
   proportional to p *)
(* P(result <= i) = p(i)/p(n) *)
var plo, lo, hi, mid: integer;
    r: real;
begin
  r := rando(version);
  plo := low(p);
(* find lo, hi: bracket r so that p(lo+plo) < r <= p(hi+plo) *)
  lo := -1;
  hi := n;
  while ((hi - lo) > 1) do begin
    mid := hi - (hi - lo) div 2;
    if (mid < 0) then result := 0 else // @@@ I think this never happens and can be dropped; also below
    if (p[mid+plo] < r*p[n+plo]) then lo := mid else hi := mid;
   end;
  newranp := hi;
end;

function newranp(version: Byte; const n: integer; const p: array of integer): Integer;
(* draws at random between 0 and n according to cumulative probability vector
   proportional to p *)
(* P(result <= i) = p(i)/p(n) *)
var plo, lo, hi, mid: integer;
    r: real;
begin
  r := rando(version);
  plo := low(p);
(* find lo, hi: bracket r so that p(lo+plo) < r <= p(hi+plo) *)
  lo := -1;
  hi := n;
  while ((hi - lo) > 1) do begin
    mid := hi - (hi - lo) div 2;
    if (mid < 0) then result := 0 else // @@@ I think this never happens and can be dropped; also below
    if (p[mid+plo] < r*p[n+plo]) then lo := mid else hi := mid;
   end;
  newranp := hi;
end;

procedure matrixranp(version: Byte; const n: integer;
                    const p: matrixarray; var i,j: integer);
(* draws at random between 0 and n according to cumulative probability vector
   proportional to p *)
(* P((i,j) lexicographically <= (h,k)) = p(h,k)/p(n,n) *)
var lo, hi, mid, ii: integer;
    r: real;
begin
  r := rando(version);
(* find lo, hi: bracket r so that p(lo,n) < r <= p(hi,n) *)
  lo := -1;
  hi := n;
  while ((hi - lo) > 1) do begin
    mid := hi - (hi - lo) div 2;
    if (mid < 0) then ii := 0 else
    if (p[mid,n] < r*p[n,n]) then lo := mid else hi := mid;
   end;
  ii := hi;
  i := ii;
(* find lo, hi: bracket r so that p(ii,lo) < r <= p(ii,hi) *)
  lo := -1;
  hi := n;
  while ((hi - lo) > 1) do begin
    mid := hi - (hi - lo) div 2;
    if (mid < 0) then j := 0 else
    if (p[ii,mid] < r*p[n,n]) then lo := mid else hi := mid;
   end;
  j := hi;
end; (* matrixranp *)


FUNCTION gamma(w: Extended):Extended;
begin
gamma := gammaran(w);
end;


FUNCTION gammln(xx: real): real;
CONST
   stp = 2.50662827465;
   half = 0.5;
   one = 1.0;
   fpf = 5.5;
VAR
   x,tmp,ser: double;
   j: integer;
   cof: ARRAY [1..6] OF double;
BEGIN
   cof[1] := 76.18009173;
   cof[2] := -86.50532033;
   cof[3] := 24.01409822;
   cof[4] := -1.231739516;
   cof[5] := 0.120858003e-2;
   cof[6] := -0.536382e-5;
   x := xx-one;
   tmp := x+fpf;
   tmp := (x+half)*ln(tmp)-tmp;
   ser := one;
   FOR j := 1 TO 6 DO BEGIN
      x := x+one;
      ser := ser+cof[j]/x
   END;
   gammln := (tmp+ln(stp*ser))
END;

FUNCTION poidev(xm: real): integer;
(* from numerical recipes *)
(* Programs using POIDEV must declare the variables
VAR
   gloldm,glsq,glalxm,glg: real;
in the main program and should intialize gloldm to
   gloldm := -1.0;   *)
CONST
   pi=3.141592654;
VAR
   em,t,y: real;
BEGIN
  poidev := trunc(xm);
   IF (xm < 12.0) THEN BEGIN
      IF (xm <> gloldm) THEN BEGIN
         gloldm := xm;
         glg := exp(-xm)
      END;
      em := -1;
      t := 1.0;
      REPEAT
         em := em+1.0;
         t := t*rando(version);
      UNTIL (t <= glg)
   END ELSE BEGIN
      IF (xm <> gloldm) THEN BEGIN
         gloldm := xm;
         glsq := sqrt(2.0*xm);
         glalxm := ln(xm);
         glg := xm*glalxm-gammln(xm+1.0)
      END;
      REPEAT
         REPEAT
            y := pi*rando(version);
            y := sin(y)/cos(y);
            em := glsq*y+xm;
         UNTIL (em >= 0.0);
         em := trunc(em);
         t := 0.9*(1.0+sqr(y))*exp(em*glalxm-gammln(em+1.0)-glg);
      UNTIL (rando(version) <= t)
   END;
   poidev := trunc(em)
END; (* poidev *)


procedure d7r8 (var xm: real; var f: textfile);
(* example/check for routine POIDEV *)
CONST
   n = 20;
   npts = 1000;
   iscal = 200;
   llen = 50;
VAR
   i,idum,j,k,klim : integer;
   dist : ARRAY [0..20] OF real;
   txt : PACKED ARRAY [1..50] OF char;
   gloldm,glsq,glalxm,glg : real;
   glinext,glinextp : integer;
   glma : ARRAY [1..55] OF real;


BEGIN
   idum := -13;
   FOR j := 0 to 20 DO BEGIN
      dist[j] := 0.0
   END;
   writeln(f,'Poisson-distributed deviate, mean ',xm:5:20);
   IF (xm < 0.0) THEN exit;
   IF (xm > 20.0) THEN exit;
   FOR i := 1 to npts DO BEGIN
      j := trunc(poidev(xm));
      IF ((j >= 0) AND (j <= 20)) THEN dist[j] := dist[j]+1
   END;
   writeln(f,npts:6,' points');
   FOR j := 0 to 19 DO dist[j] := dist[j]/npts;
   histogram(dist,0,20,50,f);
END; (* d7r8 *)



begin (* unit *)
version := 3;
version2 := 4;
(* version2 MUST be different from version;
   version defines the basic random number generator,
   while version2 defines the generator used only in S_Base-SimStats
   to align the common random number streams at the start
   of each observation period *)
for i:=0 to VersionMax do ridum[i] := -1;
nrb := false;
gloldm := -1.0;   (* initializes routine poidev *)
RanGenErrorMessage := '';
end.
