(*s_mcmc_I
Entire unit by MS:
File contains Markov chain Monte Carlo routines applied to conduct likelihood-based inference.
- estimate non-hierarchical Bayesian models
- estimate frequentist random effects models with actor-dependent random effects.
- estimate hierarchical Bayesian models with actor-dependent random effects.
Almost everything has been outsourced to units <S_mcmc>, <S_mcmc_II>, <S_re>.*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.InitialiseMS(var theta: pvec; var p: byte; var outf, lf: TFileStream);
(*MS: Initialises some arrays needed in maximum likelihood and Bayesian estimation of
models with and without actor-dependent random effects.*)
var
  a, b, count1, count2, d, dim_1, dim_2, sqr_nre, i, j: byte;
  c: integer;
  det: real;
  condition1, condition2: boolean;
  initial_theta: pvec;
  cholfactor, cov, ctheta: pmatrix;
  initial_re,
  e: rrarray; (*Commutation matrix: local variable, since it is not needed in its "pure form", but only in the form <- n E>.*)
  x: string;
  begin
(*The present procedure is nested in <MLInitialise>, which is called in case of likelihood-based inference.*)
hit2 := false; (*In case of random effects models, keeps track of whether certain procedures were already hit.*)
fine_tuning := true; (*Indicates whether MCMC algorithms should be fine-tuned.*)
OpenFiles; (*Prepare files for MCMC output.*)
ResetBlackBox(Bayesian,randomeffects,blackbox,scalebeta_i,p,scaleRE_i,nmax,nre); (*Reset blackbox to default values: must be placed before (!) <MCMC_intro>.*)
if Bayesian then
  begin
  p1 := ActiveNetworkEffects_l + ActiveActionEffects_l; (*Dimension of rate parameter.*)
  p2 := p - p1; (*Dimension of objective function parameter.*)
  p_n := ActiveNetworkEffects_l + ActiveNetworkEffects_f + ActiveNetworkEffects_g; (*Dimension of network rate and objective function parameter.*)
  InitialActionsPriorTheta(p1,gamma_p,p,ptheta,outf); (*Initial actions respecting the prior density of parameter vector.*)
  ReadCovTheta(p,cov,lf); (*Read educated guess of variance-covariance matrix of the parameter vector from file.*)
  (*Remark:
  <if (maxlike and randomeffects) = true>, then the (co)variance parameters are contained in <theta> and written to the file:
  however, in Bayesian estimation of fixed effects or random effects models,
  only the first <p x p> block of the inputted matrix is used, where <p> refers to the number of rate and objective function parameters,
  which means that using the inputted matrix is meaningful as long as the model to be estimated
  contains the same fixed effects as the model to which the inputted matrix refers; the random part of the model may be different.*)
  a := ActiveNetworkEffects_l + 1;
  b := p_n;
  c := p_n + ActiveActionEffects_l + 1;
  d := p;
  count1 := 0;
  for i := 1 to p do
    begin
    protheta[i] := theta[i]; (*Both the rate parameter and the objective function parameter estimates are used in the Bayesian MCMC algorithm.*)
    condition1 := (i in [a..b]) or (i in [c..d]); (*Objective function parameter.*)
    if condition1 then (*Objective function parameters.*)
      begin
      Inc(count1);
      reflectiontheta[count1] := theta[i];
      count2 := 0;
      for j := 1 to p do
        begin
        condition2 := (j in [a..b]) or (j in [c..d]); (*Objective function parameter.*)
        if condition2 then (*Objective function parameters.*)
          begin
          Inc(count2);
          ctheta[count1,count2] := cov[i,j];
          end;
        end;
      end;
    end;
  x := 'Reflection point used for objective function parameter:'+eol;
  lf.Write(pchar(x)^,length(x));
  lf.Write(pchar(eol)^,length(eol));//writeln(lf);
  for i := 1 to p2 do begin
    x := format('%10.6f',[reflectiontheta[i]])+eol;
    lf.Write(pchar(x)^,length(x));
  end;
  lf.Write(pchar(eol)^,length(eol));//writeln(lf);
  x := 'Variance-covariance matrix of candidate-generating (multivariate) Gaussian density of the objective function parameter vector:'+eol;
  lf.Write(pchar(x)^,length(x));
  MatrixWriteTriangle(p2,lf,ctheta);
  Cholesky(p2,choltheta,ctheta,det); (*Cholesky factor of variance-covariance matrix of candidate-generating density: used to sample candidates.*)
  MatrixI(p2,det,ictheta,ctheta,choltheta); (*Inverse variance-covariance matrix of candidate-generating density: used to evaluate the density of candidates.*)
  SetVectorValue(p,lastsample,0); (*Strictly speaking wrong, but its effect vanishes as soon as the first candidate is accepted.*)
  end;
if randomeffects then
  begin
  EvaluateObservedDistances(true,nmax,observations,lf); (*Evaluate row-wise distance between two adjacency matrices on some given actor set.*)
  SetProposalRE(true,nmax,nre,re,eb,lastmvn0,reflectionre,sumsigma,nno); (*Prepare some things to be exploited by the candidate-generating density of the actor-dependent random effects.*)
  SetArray(lastebre,nmax,nre);
  InitialSigma(nre,covRE,1,lf); (*Initial guess of the variance-covariance of random effects utilizing user input or <constant> times identity matrix.*)
  Cholesky(nre,cholfactor,covRE,det); (*Cholesky decomposition of positive definite variance-covariance-matrix.*)
  InitialRE(nmax,nre,cholfactor,re);
  {
  Important note: if <MakeChains> precedes <InitialiseMS>, then the following code must be used to ensure
  that the <lprob>'s (evaluated at the inital <re>) are adapted to the new random effects:
  EvaluateMultinomial(2,theta,theta,initial_re,re,mnlactor1,mnlactor2,outf,lf);
  }
  SetArray(sumre,nmax,nre);
  if maxlike then
    begin
    (*Elimination matrix and transposed elimination matrix of <Li>:*)
    EliminationMatrixII(nre,elimination_li); (*Elimination matrix satisfying <D v(Li) + vec(I) = vec(Li).*)
    dim_1 := sqr(nre); (*The dimension of <vec(Li)>.*)
    dim_2 := (nre * (nre - 1)) div 2; (*The total number of below-diagonal elements: the dimension of <v(Li)>.*)
    MatrixT(dim_1,dim_2,t_elimination_li,elimination_li); (*Transpose elimination matrix.*)
    (*Elimination matrix and transposed elimination matrix of <Ksi>:*)
    EliminationMatrixIII(nre,elimination_ksi); (*Elimination matrix satisfying <D v(Ksi) = vec(Ksi).*)
    dim_1 := sqr(nre); (*The dimension of <vec(Ksi)>.*)
    dim_2 := nre; (*The total number of diagonal elements: the dimension of <v(Ksi)>.*)
    MatrixT(dim_1,dim_2,t_elimination_ksi,elimination_ksi); (*Transpose elimination matrix.*)
    (*Obtain, from the variance-covariance matrix of random effects,
    - the Cholesky factor
    - the vectors <v(Li)> and <v(Ksi)>, and insert them into the parameter vector <theta>;
    both the parameter vector <theta> and its dimension <p> will be manipulated,
    therefore these entities must be declared as variables here and in <MLInitialise>.*)
    CommutationMatrix(nre,e); (*Construct the commutation matrix of order <square(nre) x square(nre)>.*)
    sqr_nre := sqr(nre); (*The dimension of the commutation matrix: <sqr_nre x sqr_nre>.*)
    c := - nmax;
    ScalarProd(sqr_nre,ce,c,e); (*The commutation matrix is needed only in the form <- n E>.*)
    Get_Ksi_Li_Cov(nre,c_covRE,covRE,li,ksi,p,theta);
    end (*Maximum likelihood.*)
  else InitialActionsPriorSigma(pdegrees,nre,psigma,outf); (*Bayesian: some initial actions for the prior probability density (Inverse-Wishart) of the variance-covariance matrix of the random effects.*)
  end;
MCMC_intro(maxlike,Bayesian,nmax,scaleRE_i,outf); (*Write some basic description of the Markov chain Monte Carlo sampling scheme to the output file.*)
initial_re := nil;
e := nil;
end; (*MS: InitialiseMS*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.SoNIA(var pp: byte; var theta: pvec);
(*MS: Generates SoNIA input file;
state of the art: network, two or more periods, but no covariates and behavior.*)
var
  i, j, count, num, h: integer;
  k: TMiniStep;
  u: real;
  condition: boolean;
  s, s1, s2: shortstring;
  f: textfile;
  timing: R_Ar;
  x: rrrarray;
begin
writeln;
s := pname + '.son'; (*SoNIA input file.*)
OpenFile(f,s);
(*Initial comment and header.*)
writeln(f,'//Comment: SoNIA input file as generated by Siena.');
writeln(f,'//Parameter values:');
for i := 1 to pp do writeln(f,'//',Alignstring(Description[i],50),theta[i]:9:4);
writeln(f,'// ',observations-1,' periods, having numbers of changes per period: ');
for h:=1 to observations-1 do if (not skipi[h]) then
  begin
  writeln(f,'// period ',h:2,': ',Chain[h-1].totnumber,'.');
  end;
write(f,'NodeId',#9,'Label',#9,'StartTime',#9,'EndTime');
condition := true;
if condition then
  begin
  write(f,#9,'ColorName',#9,'NodeShape');
  end;
writeln(f);
for i := 1 to nmax do
  begin
  {
  (*MS: must be extended when there are more than two observation points.*)
  if ((zx[i,1] > 1) and (zx[i,2] > observations)) then writeln(f,i,#9,i,#9,event_m[i,1]:9:6,#9,'10') (*Joining.*)
  else if ((zx[i,1] < 1)  and (zx[i,2] < observations)) then writeln(f,i,#9,i,#9,'0',#9,event_m[i,1]:9:6) (*Leaving.*)
  else
  }
  write(f,i,#9,i,#9,'0',#9,'10'); (*Active througout the period.*)
  u := random;
  if (u >= 0.5) then s1 := 'black'
  else s1 := 'red';
  u := random;
  if (u >= 0.5) then s2 := 'ellipse'
  else s2 := 'rect';
  if condition then write(f,#9,s1,#9,s2);
  writeln(f);
  end;
writeln(f,'FromId',#9,'ToId',#9,'StartTime',#9,'EndTime',#9,'ArcWeight');
x := nil;
SetLength(x,nmax+1,nmax+1,2);
for h:=1 to observations - 1 do if (not skipi[h]) then 
  begin
  Initialise_actors(h);
  Initialise_y(h); (*Initialise network observed at first observation point.*)
  for i := 1 to n do
    begin
    for j := 1 to n do
      begin
      if yn(i,j) then x[i,j,0] := 1 else x[i,j,0] := 0;
      end;
    end;
  (*Write sequence of changes to SoNIA input file in the form of an edge list.*)
  k := Chain[h-1].First;
  num := 0;
  count := 0;
  i := 10000;
  SetArray(timing,i);
  repeat
    inc(num);
    k := k.Such;
    if ((not(k.extreme)) and (k.ih <> 0)) then (*Omit the empty entries: the first and the last one.*)
      begin
      i := k.ih;
      j := k.jh;
      Inc(count);
      timing[k.kindex] := count / Chain[h-1].totnumber;
      if (x[i,j,0] = 1) then (*That is: (1) there is an arc, (2) but it is to be terminated.*)
        begin
        writeln(f,k.ih,#9, (*The actor that changes something.*)
                  k.jh,#9, (*The actor to which <i> terminates the arc.*)
                  x[i,j,1]:9:6,#9, (*The time point of creation.*)
                  timing[k.kindex]:9:6,#9, (*The time point of termination.*)
                  x[i,j,0]:2:0); (*Unity, which indicates that the arc was there in the indicated time period.*)
        end;
      x[i,j,0] := 1 - x[i,j,0]; (*Stores value to which the coordinate is changed.*)
      x[i,j,1] := timing[k.kindex]; (*Stores time point of last change.*)
      end;
  until (k.Extreme or (num >= 30000));
  u := 10; (*Time period from the last change up to time point <1>.*)
  condition := false;
  for i := 1 to n do
    begin
    for j := 1 to n do
      begin
      if (x[i,j,0] = 1) then (*That is: (1) there is an arc, (2) but it is to be terminated.*)
        begin
        if condition then writeln(f);
        condition := true;
        write(f,i,#9, (*The actor that changes something.*)
                j,#9, (*The actor to which <i> terminates the arc.*)
                x[i,j,1]:9:6,#9, (*The time point of creation.*)
                u:9:6,#9, (*Time point <1>.*)
                x[i,j,0]:2:0); (*The value <1>, which indicates that the arc was there in the indicated time period.*)
        end;
      end;
    end;
  end;  
closefile(f);
end; (*MS: SoNIA.*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.LastPath(var pp: byte; var theta: pvec; const boolean1: boolean);
(*MS: Writes last path sampled from the conditional distribution of the paths to ASCII file.*)
var
  interval: integer;
  s: shortstring;
  fI, fII: textfile;
begin
SoNIA(pp,theta);
if boolean1 then
  begin
  s := paramstr(1) + 'LSP.dat'; (*LSP = Last Sampled Path.*)
  OpenFile(fI,s);
  s := 'log.log';
  OpenFile(fII,s);
  for interval := 1 to (observations - 1) do
    if (not skipi[interval]) then WriteFullPeriod(interval,fI,fII);
  closefile(fI);
  closefile(fII);
  erase(fII);
  end;
end; (*MS: LastPath*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.GetInitialRE(var outf, lf: TFileStream; const p: byte; const theta: pvec);
(*MS: regarding the random effects as fixed, latent constants, the augmented-data likelihood function of the random effects is evaluated on a grid,
and initial guesses of the random effects are computed.*)
var
  i: integer;
  h: byte;
  average, sum, umin, umax: real;
  u: extended;
  f: textfile;
  s: shortstring;
  invsigma: pmatrix;
  initial_re, store_re: rrarray;
  x: string;
begin
(*WARNING: procedure messes up the chain for unknown reasons.*)
SetArray(initial_re,nmax,nre);
SetArray(store_re,nmax,nre);
for i := 1 to nmax do
  begin
  for h := 1 to nre do
    begin
    if (re[i,h] <> initial_re[i,h]) then
      begin
      x := 'GetInitialRE was called with non-null random effects matrix: no change, passing control to next-higher procedure...'+eol;
      lf.Write(pchar(x)^,length(x));
      lf.Write(pchar(x)^,length(eol));//writeln(lf);
      exit;
      end;
    end;
  end;
s := pname + 'InitialRE.m';
OpenFile(f,s);
SetIdentityMatrix(nre,invsigma);
umin := -4.05;
umax := 3.95;
for i := 1 to nmax do
  begin
  for h := 1 to nre do
    begin
    if (nre = 1) then write(f,'x',i,' = [ ')
    else write(f,'x',i,h,' = [ ');
    re[i,h] := umin;
    average := 0;
    sum := 0;
    while (re[i,h] <= umax) do
      begin
      re[i,h] := re[i,h] + 0.05;
      EvaluateMultinomial(1,theta,theta,re,re,mnlactor1,mnlactor2,outf,lf);  (*Remark: here: the first argument may be <1> or <2>.*)
      u := mnlactor1[0]; (*Multinomial logit mass.*)
      u := u + EvaluateMVNactor(nre,re,i,invsigma); (*Multivariate normal kernel.*)
      u := expo(u);
      sum := sum + u;
      average := average + (u * re[i,h]);
      WriteReal(f,u);
      end;
    writeln(f,']');
    store_re[i,h] := average / sum; (*Weighted average.*)
    writeln(f,'%weighted average: ',store_re[i,h]:4:4);
    re[i,h] := 0; (*Reset <re[i,h]> so that the random effects for all other actors and coordinates is null: order of evaluating the function is of no importance.*)
    end;
  end;
SetMatrix(nmax,nre,re,store_re); (*Set <re> to the weighted average <store_re>.*)
u := umin;
while (u <= umax) do
  begin
  u := u + 0.05;
  writeln(f,' ',u:10:10);
  end;
closefile(f);
initial_re := nil;
store_re := nil;
end; (*MS: GetInitialRE*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.ComputeTotal(var total_t: R_Ar; var total_n: iiarray);
(*MS: obtain and store the length of the time interval and the total number of changes in the time interval for each dependent variable.*)
var
  position, h: byte;
  number, period: integer;
  k: TMiniStep;
begin
total_t := nil;
total_n := nil;
number := 1 + nza; (*Number of dependent variables: network and behavior.*)
SetLength(total_t,observations); (*Elements <1..observations-1> are needed.*)
SetLength(total_n,number+1,observations);
for period := 1 to (observations - 1) do if (not skipi[period]) then
  begin
  total_t[period] := Chain[period-1].totaltime; (*Length of period.*)
  k := Chain[period-1].First.Such; (*Move ministep k from the first to the last in this period.*)
  repeat
  if (k.ash = Network) then position := 1
  else
    begin
    h := k.varh; (*Dependent action variable that changed.*)
    position := 1 + h;
    end;
  Inc(total_n[position,period]); (*Total number of changes in period for given dependent variable.*)
  k := k.Such;
  until k.Extreme;
  end;
end; (*MS: ComputeTotal*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.PartitionParameter(const dim1, dim2: byte; var vector1, vector2: pvec; const vector3: pvec);
(*MS: Partition the parameter vector into rate-function (vector valued) parameter and non-rate function (vector-valued) parameter.*)
var count1, count2, i: byte;
begin
(*Remark:
- input: parameter vector <vector3>.
- output: rate parameter <vector1> and objective function parameter <vector2>.*)
count1 := 0;
count2 := 0;
for i := 1 to ActiveNetworkEffects_l do (*Network rate parameters.*)
  begin
  Inc(count1);
  vector1[count1] := vector3[i];
  end;
for i := (ActiveNetworkEffects_l + 1) to dim1 do (*Network objective function parameters.*)
  begin
  Inc(count2);
  vector2[count2] := vector3[i];
  end;
for i := (dim1 + 1) to (dim1 + ActiveActionEffects_l) do (*Action rate parameters.*)
  begin
  Inc(count1);
  vector1[count1] := vector3[i];
  end;
for i := (dim1 + ActiveActionEffects_l + 1) to dim2 do (*Action objective function parameters.*)
  begin
  Inc(count2);
  vector2[count2] := vector3[i];
  end;
end; (*MS: PartitionParameter*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.InversePartitionParameter(const dim1, dim2: byte; const vector1, vector2: pvec; var vector3: pvec);
(*MS: Inverse the operation of partitioning the parameter vector: obtain the parameter vector <vector3> from rate parameter <vector1> and objective function parameter <vector2>.*)
var count1, count2, i: byte;
begin
(*Remark:
- input: rate parameter <vector1> and objective function parameter <vector2>.
- output: parameter vector <vector3>.*)
count1 := 0;
count2 := 0;
for i := 1 to ActiveNetworkEffects_l do (*Network rate parameters.*)
  begin
  Inc(count1);
  vector3[i] := vector1[count1];
  end;
for i := (ActiveNetworkEffects_l + 1) to dim1 do (*Network objective function parameters.*)
  begin
  Inc(count2);
  vector3[i] := vector2[count2];
  end;
for i := (dim1 + 1) to (dim1 + ActiveActionEffects_l) do (*Action rate parameters.*)
  begin
  Inc(count1);
  vector3[i] := vector1[count1];
  end;
for i := (dim1 + ActiveActionEffects_l + 1) to dim2 do (*Action objective function parameters.*)
  begin
  Inc(count2);
  vector3[i] := vector2[count2];
  end;
end; (*MS: InversePartitionParameter*)

(*--------------------------------------------------------------------------------------------------*)

function MLModel.Poisson(var theta: pvec): real;
(*MS: evaluate joint Poisson kernel on logarithmic scale.*)
var
  period, tot_n: integer;
  sum, tot_t: real;
  k: TChain;
begin
(*Assumption: network model; must be extended to network and behavior models.*)
sum := 0;
for period := 1 to (observations - 1) do if (not skipi[period]) then 
  begin
  k := Chain[period-1];
  tot_n := k.totnumber; (*Total number of changes in period.*)
  tot_t := k.totaltime; (*Length of period.*)
  sum := sum + EvaluatePoisson(n,theta[period],tot_t,tot_n); (*Evaluate Poisson kernel on logarithmic scale.*)
  end;
Poisson := sum;
end; (*MS: Poisson*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.EvaluateMultinomial(const sampling: byte; const theta1, theta2: pvec; const actorRE1, actorRE2: rrarray; (*Input.*)
var mnl_i_1, mnl_i_2: R_Ar; var outf, lf: TFileStream); (*Output.*)
(*MS: Evaluate joint multinomial choice probability on logarithmic scale twice:
- <sampling = 1>: once given <theta1> and once given <theta2>.
- <sampling = 2>: once given <actorRE1> and once given <actorRE2>.
Note: if <EvaluateMultinomial> is followed by statements manipulating either parameters or random effects or both,
then it must be followed by <UpdateLProbs>.
*)
const tiny = 1e-16;
var i, j, h, period: integer;
    denominator1, denominator2, numerator1, numerator2, mnl_1, mnl_2, u1, u2: real;
    Error, HError: Boolean;
    ErrorString: shortstring;
    k: TMiniSTep;
    statistic: rrarray;
    permitted: boarray;
    PermDir: B3array;
    x: string;
begin
(*Important remark:
<NumberParameters> refers to the number of fixed effects, excluding (co)variance parameters:
otherwise, some summations below would be wrong when random effects models are estimated by MLE.*)
(*Initialise:*)
Error := false;
ErrorString := '';
HError := false;
mnl_1 := 0;
mnl_2 := 0;
SetArray(permitted,n);
SetArray(mnl_i_1,n);
SetArray(mnl_i_2,n);
(*Period-dependent path:*)
for period := 1 to (observations - 1) do if (not skipi[period]) then 
  begin
  Initialise_actors(period);
  Initialise_y(period);
  AdaptProposalProbabilities;
  (*
  Important remark:
  Calling <RenewRateslprobs> from <Chain[period-1].First> to <Chain[period-1].Last> is not required:
  see extended remark at the end of <MCMC_cycle>.
  *)
  k := Chain[period-1].First.Such; (*Move ministep k from the first to the last in this period.*)
  SetData(Chain[period-1].First,period,outf,lf);
  repeat
  i := k.ih; (*The actor that changes something.*)
  denominator1 := 0;
  denominator2 := 0;
  if (k.ash = Network) then
    begin
    (*--------------------------------------*)
    (*Network: alternatives <j = 1, ..., n>.*)
    (*--------------------------------------*)
    if ((not(mis(i,k.jh,period) or mis(i,k.jh,period+1) or Struct(i,k.jh,period+1))) and StructActive[i]) then
      begin
      SetArray(statistic,n,NumberParameters);
      if checkchange then PermitChange(n,i,period,maxdegree,permitted)
      else
        begin
        for j := 1 to n do permitted[j] := true;
        end;
      if anynetworkmissings then ForbidMissings(i,period,permitted); (*Conditional probabilities among the nonmissings are calculated.*)
      permitted[i] := (StaySameType >= 2);
      UtilityComponents_yn(i,period,ActiveNetworkEffects_l,n,statistic,permitted,NewNetworkEffects);
      for j := 1 to n do if permitted[j] then
        begin
        u1 := Multinomial(ActiveNetworkEffects_l+1,var_n_tot,theta1,actorRE1,i,j,statistic); (*Multinomial choice probability up to the the normalizing constant on logarithmic scale.*)
        case sampling of
          1: (*Sampling objective function parameter*)
            begin
            u2 := Multinomial(ActiveNetworkEffects_l+1,var_n_tot,theta2,actorRE1,i,j,statistic);
            end;
          2: (*Sampling actor-dependent random effects*)
            begin
            u2 := Multinomial(ActiveNetworkEffects_l+1,var_n_tot,theta1,actorRE2,i,j,statistic);
            end;
          end; (* case sampling of *)
        if (j = k.jh) then (*The alternative that was in fact chosen.*)
          begin
          numerator1 := u1; (*On logarihmic scale!*)
          numerator2 := u2;
          end;
        denominator1 := denominator1 + expo(u1);
        denominator2 := denominator2 + expo(u2);
        end; (* if permitted *)
      (*Note: if not permitted, then nothing must be added to the denominators.*)
      end; (* if not missing *)
    ChangeTie(i,k.jh);
    end (* if k.ash = network *)
  else
    begin (* if k.ash = Behavior *)
    (*-----------------------------------*)
    (*Behavior: alternatives <-1, 0, +1>.*)
    (*-----------------------------------*)
    h := k.varh; (*The action variable that was changed.*)
    if ((not(miszc(i,h,period) or miszc(i,h,period+1))) and Structactive[i]) then
      begin
      SetArray(statistic,3,NumberParameters);
      PermDir[1] := (not Minimalya(i,h));  (*Change -1 permitted.*)
      PermDir[2] := true;  (*Change 0 permitted.*)
      PermDir[3] := (not Maximalya(i,h));  (*Change +1 permitted.*)
      UtilityComponents_ya(i,h,BeforeActionParameters_f[h],statistic,PermDir,HError,outf,lf);
      if HError then
        begin
        Error := true;
        ErrorString := 'Error in EvaluateMultinomial: Action Utility Components.';
        end;
      for j := 1 to 3 do if PermDir[j] then
        begin
        u1 := Multinomial(BeforeActionParameters_f[1]+1,NumberParameters,theta1,actorRE1,i,j,statistic); (*Multinomial choice probability up to the the normalizing constant on logarithmic scale.*)
        case sampling of
          1: (*Sampling objective function parameter*)
            begin
            u2 := Multinomial(BeforeActionParameters_f[1]+1,NumberParameters,theta2,actorRE1,i,j,statistic);
            end;
          2: (*Sampling actor-dependent random effects*)
            begin
            u2 := Multinomial(BeforeActionParameters_f[1]+1,NumberParameters,theta1,actorRE2,i,j,statistic);
            end;
          end; (* case sampling of *)
        if (j = k.difh) then (*The alternative that was in fact chosen.*)
          begin
          numerator1 := u1; (*On logarihmic scale!*)
          numerator2 := u2;
          end;
        denominator1 := denominator1 + expo(u1);
        denominator2 := denominator2 + expo(u2);
        end; (* if permDir *)
      (*Note: if not permDir, then nothing must be added to the denominators.*)
      end; (* not miszc *)
    ChangeBehavior(i,h,k.difh);
    end; (* if k.ash = Behavior *)
  if (denominator1 < tiny) then denominator1 := tiny; (*To avoid runtime errors when evaluating its logarithm.*)
  if (denominator2 < tiny) then denominator2 := tiny;
  u1 := numerator1 - ln(denominator1);
  u2 := numerator2 - ln(denominator2);
  k.prob1 := u1; (*Log multinomial choice probability: actor <i> choosing alternative <j>*)
  k.prob2 := u2;
  mnl_i_1[i] := mnl_i_1[i] + u1; (*Joint log multinomial choice probability: actor <i>.*)
  mnl_i_2[i] := mnl_i_2[i] + u2;
  mnl_1 := mnl_1 + u1; (*Joint log multinomial choice probability: path.*)
  mnl_2 := mnl_2 + u2;
  k := k.Such;
  until k.Extreme;
  (*Remark: the boolean function <k.Extreme> indicates whether <kindex <= 1>;
  the function is true for the last step in the embedded chain,
  corresponding to <kindex = 1>, which is an empty step in the sense that it does not correspond to a change; see <Chains.pas>.*)
  (*Checking:*)
  for i := 1 to n do
    begin
    for j := 1 to n do
      begin
      if ((i <> j) and (yn(i,j) <> ynm(i,j,period+1))) then Error := true;
      end;
    if (nza > 0) then
      begin
      if (ya(i,1) <> zc[i,1,period+1]) then Error := true;
      end;
    end;
  if Error then
    begin
    x := format('%s%d%s',['EvaluateMultinomial: Wrong data for this chain at end of period ',period,'.'])+eol;
    lf.Write(pchar(x)^,length(x));
    ErrorString := 'Chain does not lead to the correct end situation.';
    end;
  end; (* for period *)
(*Finalize:*)
mnl_i_1[0] := mnl_1; (*Joint multinomial choice probability on natural logarithmic scale.*)
mnl_i_2[0] := mnl_2;
end; (*MS: EvaluateMultinomial*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.ImproveMH(var outf, lf: TFileStream; var theta: pvec; var p: byte);
(*MS: Method to improve efficiency of Metropolis-Hastings algorithm by improving acceptance rate.*)
const tiny = 1e-8;
var
  actual, c1, c2, desired1, desired2, i, iteration, maximum, tolerance: integer;
  count1, count2, stage: byte;
  u: real;
  condition, hit_1, hit_2, success: boolean;
  x: string;
begin
(*First: check necessity of fine-tuning:*)
if not(Bayesian) then success1 := true;
if not(randomeffects) then success2 := true;
success := (success1 and success2);
if success then exit; (*Nothing to be done.*)
(*Otherwise initialise:*)
c1 := nrunMH_batches; (*Store value of <nrunMH_batches>.*)
nrunMH_batches := 100;
c2 := nrunMH; (*Store value of <nrunMH>.*)
nrunMH := 1;
stage := 0; (*Stage of the estimation process.*)
case cdensity1 of (*Desired acceptance rates depend on the particular Metropolis-Hastings algorithm.*)
  1: desired1 := trunc(0.40 * nrunMH_batches); (*Random walk.*)
  2: desired1 := trunc(0.40 * nrunMH_batches); (*Autoregressive density.*)
  3: desired1 := trunc(0.40 * nrunMH_batches); (*Independence sampler.*)
  end; (*<case cdensity1 of>.*)
case cdensity2 of (*Desired acceptance rates depend on the particular Metropolis-Hastings algorithm.*)
  1: desired2 := trunc(0.40 * nrunMH_batches); (*Random walk.*)
  2: desired2 := trunc(0.30 * nrunMH_batches); (*Autoregressive density.*)
  3: desired2 := trunc(0.40 * nrunMH_batches); (*Independence sampler.*)
  end; (*<case cdensity2 of>.*)
tolerance := round(0.15 * nrunMH_batches); 
count1 := 0;
count2 := 0;
condition := false;
hit_1 := false; (*Keep track of whether the scaling constant went to 0: parameter vector.*)
hit_2 := false; (*Keep track of whether the scaling constant went to 0: actor-dependent random effects.*)
maximum := 100; (*<maximum> is the maximum number of tuning runs of <MCMC_cycle>.*)
(*Iterations:*)
iteration := 0;
repeat
Inc(iteration);
(*Check whether it is sensible to continue:*)
if (not(success1) and (blackbox[12] < tiny)) then
  begin
  if not(hit_1) then (*First time.*)
    begin
    blackbox[12] := 2.54 / sqrt(p); (*Do it again...*)
    hit_1 := true;
    end
  else (*Second time: stop it, it seems to be hopeless...*)
    begin
    blackbox[12] := 1;
    success1 := true;
    end;
  end;
(*Check whether it is sensible to continue:*)
if not(success2) then
  begin
  for i := 1 to n do
    begin
    if (scaleRE_i[i] < tiny) then
      begin
      if not(hit_2) then (*First time.*)
        begin
        scaleRE_i[i] := 2.54 / sqrt(nre);
        hit_2 := true;
        end
      else condition := true; (*Second time: stop it, it seems to be hopeless...*)
      end;
    end;
  if condition then (*At least one of the actor-dependent random effects is in trouble.*)
    begin
    for i := 1 to n do
      begin
      if (scaleRE_i[i] < tiny) then scaleRE_i[i] := 1;
      end;
    success2 := true;
    end;
  end;
MCMC_cycle(true,outf,lf,p,theta,stage,1); (*Note: <aim = 1> and <setpar = false>.*)
(*MS: Cycle through
(1) M-H step to sample path: network and behavior (ML and Bayesian models)
(2) M-H step to sample actor-dependent random effects (ML and Bayesian random effects models)
(3) Gibbs and M-H steps to sample parameters (Bayesian models).
Note that the value of the first argument is not relevant here: it is used for MLE of fixed effects models,
while here, either <randomeffects = true> or <Bayesian = true>.*)
if not(success1) then (*Non-rate parameter vector: scale factor needs to be adapted.*)
  begin
  actual := trunc(blackbox[7]);
  u := RescaleCGD(actual,desired1,nrunMH_batches,tolerance,count1,success1); (*Rescale candidate-generating density to achieve desired acceptance rate.*)
  blackbox[12] := blackbox[12] * u;
  end;
if not(success2) then (*Actor-dependent random effects: scale factor needs to be adapted.*)
  begin
  actual := trunc(blackbox[1]);
  u := RescaleCGD(actual,desired2,nrunMH_batches,tolerance,count2,success2); (*Rescale candidate-generating density to achieve desired acceptance rate.*)
  blackbox[11] := blackbox[11] * u;
  for i := 1 to n do scaleRE_i[i] := blackbox[11];
  end;
success := (success1 and success2);
until (success or (iteration = maximum));
for i := 1 to 10 do blackbox[i] := 0;
x := format('%s%d%s',['ImproveMH: fine-tuning the MCMC algorithm for sampling random effects and / or objective function parameters required ',iteration,' iterations.'])+eol;
lf.Write(pchar(x)^,length(x));
lf.Write(pchar(eol)^,length(eol));//writeln(lf);
if Bayesian then WriteScaleCDtheta(lf,blackbox[12]); (*Writes the scale factor corresponding to the multivariate candidate-generating density for the parameter vector M-H steps to log-file.*)
if randomeffects then WriteScaleCDRE(lf,n,scaleRE_i); (*Writes the actor-dependent scale factors corresponding to the candidate-generating density for the random effects M-H steps to log-file.*)
nrunMH_batches := c1; (*Reset <nrunMH_batches> to its user-inputted value.*)
nrunMH := c2; (*Reset <nrunMH> to its user-inputted value.*)
end; (*MS: ImproveMH*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.UpdateLProbs(const proposal_accepted: boolean);
(*MS: update the change probabilities <lprob>;
required when
- random effects: the random effects have changed.
- Bayesian framework: the objective function parameter has changed.
Arguments of <EvaluateMultinomial> must be in accordance with the present procedure: see below.
*)
var
  period: integer;
  k: TMiniStep;
begin
if proposal_accepted then
  begin
  for period := 1 to (observations - 1) do  if (not skipi[period]) then 
    begin
    k := Chain[period-1].First.Such; (*Move ministep k from the first to the last in this period.*)
    repeat
    (*Note I: <lprob> is the probability of the change,
        while <prob1> and <prob2> are the multinomial probability of the change
        at the present and the future state (of the parameter / random effects) respectively.
      Note II: the arguments <theta1, theta2, actorRE1, actorRE2> of <EvaluateMultinomial> must be specified in accordance with the present procedure:
        e.g., <actorRE1> must be the present state, while <actorRE2> is the (possible) future state.*)
    k.lprob := k.lprob - k.prob1 + k.prob2;
    k := k.Such;
    until k.Extreme;
    (*Remark: the boolean function <k.Extreme> indicates whether <kindex <= 1>;
    the function is true for the last step in the embedded chain,
    corresponding to <kindex = 1>, which is an empty step in the sense that it does not correspond to a change; see <Chains.pas>.*)
    end; (* for period *)
  end;
end; (*MS: UpdateLProbs*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.UpdateLProbsActors;
(*MS: update the change probabilities <lprob>;
the present procedure must be called if multi-block M-H algorithms are used to sample actor-dependent random effects.
Arguments of <EvaluateMultinomial> must be in accordance with the present procedure: see below.
*)
var
  i, h, period: integer;
  k: TMiniStep;
begin
for period := 1 to (observations - 1) do if (not skipi[period]) then 
  begin
  k := Chain[period-1].First.Such; (*Move ministep k from the first to the last in this period.*)
  repeat
  (*Note I: <lprob> is the probability of the change,
      while <prob1> and <prob2> are the multinomial probability of the change
      at the present and the future state (of the parameter / random effects) respectively.
    Note II: the arguments <theta1, theta2, actorRE1, actorRE2> of <EvaluateMultinomial> must be specified in accordance with the present procedure:
      e.g., <actorRE1> must be the present state, while <actorRE2> is the (possible) future state.*)
  i := k.ih; (*Actor that initiated change indexed by <k>.*)
  if accepted[i] then 
    begin
    if (k.ash = Network) then h := 1
    else h := 1 + k.varh;
    (*Update the <lprobs> if, for the dependent variable <h> that changed, there are random coefficients in the objective function:*)
    if variable_rc[h] then k.lprob := k.lprob - k.prob1 + k.prob2;
    end;
  k := k.Such;
  until k.Extreme;
  (*Remark: the boolean function <k.Extreme> indicates whether <kindex <= 1>;
  the function is true for the last step in the embedded chain,
  corresponding to <kindex = 1>, which is an empty step in the sense that it does not correspond to a change; see <Chains.pas>.*)
  end; (* for period *)
end; (*MS: UpdateLProbsActors*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.EvaluateEfficiency(var estage: byte; var outf, lf: TFileStream; var p: byte; var theta: pvec);
(*MS: Evaluate efficiency of Markov chain Monte Carlo algorithm.*)
var
  i: byte;
  c1, c2, nres: integer;
begin
c1 := nrunMH_batches; (*Store value of <nrunMH_batches>.*)
nrunMH_batches := 5000;
c2 := nrunMH; (*Store value of <nrunMH>.*)
nrunMH := 1;
MCMC_cycle(true,outf,lf,p,theta,estage,2); (*Note: <aim = 2> and <setpar = false>.*)
(*MS: Cycle through
(1) M-H step to sample path: network and behavior (ML and Bayesian models)
(2) M-H step to sample actor-dependent random effects (ML and Bayesian random effects models)
(3) Gibbs and M-H steps to sample parameters (Bayesian models).
Note that the value of the first argument is not relevant here: it is used for MLE of fixed effects models,
while here, either <randomeffects = true> or <Bayesian = true>.*)
if Bayesian then IEFactor(lf,nrunMH_batches,p,autoTheta,ISE_theta); (*Evaluates inefficiency factors for parameter coordinates (Chib, 2004).*)
if randomeffects then
  begin
  IEFactor(lf,1,nrunMH_batches,n,nre,autoRE,ISE_RE); (*Evaluates inefficiency factors for random effects (Chib, 2004).*)
  LastPath(p,theta,randomeffects); (*Writes last path sampled from the conditional distribution of the paths to ASCII file.*)
  if Bayesian then
    begin
    nres := nre;
    IEFactor(lf,2,nrunMH_batches,nres,nre,autoCovRE,ISE_CovRE); (*Evaluates inefficiency factors for the variance-covariance matrix of the random effects (Chib, 2004).*)
    end;
  end;
nrunMH_batches := c1; (*Reset <nrunMH_batches> to its user-inputted value.*)
nrunMH := c2; (*Reset <nrunMH> to its user-inputted value.*)
end; (*MS: EvaluateEfficiency*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MH_P_Step(const p: byte; var theta: pvec; var outf, lf: TFileStream);
(*MS: Gibbs and Metropolis-Hastings steps to sample parameters: rate and objective function parameters.*)
var
  i, i_a, count, position: byte;
  j, n_a, number, tot_n, variable: integer;
  tot_t: real;
  vec1, vec2: pvec;
  no_rate_parameters: array of byte;
  total_t: R_Ar;
  n_active, total_n: iiarray;
begin
PartitionParameter(p_n,p,vec1,vec2,theta); (*Partition the parameter vector <theta> into rate parameter <vec1> and objective function parameter <vec2>.*)
ComputeTotal(total_t,total_n); (*Compute for each period: (1) its length, and (2) for each dependent variable: the total number of changes.*)
number := 1 + nza; (*Number of dependent variables.*)
no_rate_parameters := nil;
SetLength(no_rate_parameters,number+1);
for i := 1 to number do no_rate_parameters[i] := ActivePeriods; (*Constant rate functions; needs to be extended to non-constant rate functions.*)
n_active := nil;
SetLength(n_active,number+1,observations);
for i := 1 to (observations - 1) do if (not skipi[i]) then
  begin
  Initialise_actors(i); (*Initialise, for each dependent variable, the number of active actors at the beginning of period <i>.*)
  n_active[1,i] := n_act; (*Network.*)
  for j := 1 to nza do n_active[1+j,i] := n_actnonmis[j]; (*Action variable <j>.*)
  end;
variable := 1;
count := 0;
for i := 1 to p1 do (*Loop could be nested in <MH_PI_Step>, but that would make it impossible to sample e.g. for one period only.*)
  begin
  Inc(count);
  (*Important remark:
  since the rate functions are assumed to be constant, <count> represents the period to which the rate parameter corresponds.*)
  tot_t := total_t[count]; (*The length of the period.*)
  n_a := n_active[variable,count]; (*Dependent variable <variable>: number of active actors at the beginning of period <count>.*)
  tot_n := total_n[variable,count]; (*For <variable>: the total number of changes in the period.*)
  (*The position of rate parameter <i> in the parameter vector <theta>:*)
  if (i <= ActiveNetworkEffects_l) then position := i
  else
    begin
    i_a := i - ActiveNetworkEffects_l;
    position := p_n + i_a;
    end;
  MH_PI_Step(i,position,n_a,tot_n,tot_t,theta,outf,lf); (*MS: Gibbs or Metropolis-Hastings step to sample rate parameter <i>.*)
  (*The arguments of <MH_PI_Step> are:
    - <i>: the rate parameter.
    - <position>: the position of the rate parameter <i> in the parameter vector <theta>.
    - <tot_t>: the length of the period.
    - <n_a>: for <variable>: the number of active actors at the beginning of the period.
    - <tot_n>: for <variable>: the total number of changes in the period.*)
  if (count = no_rate_parameters[variable]) then
    begin
    count := 0;
    Inc(variable);
    end;
  end;
MH_PII_Step(p,vec1,vec2,theta,outf,lf); (*MS: Metropolis-Hastings step to sample objective function parameter.*)
no_rate_parameters := nil;
n_active := nil;
total_t := nil;
total_n := nil;
end; (*MS: MH_P_Step*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MH_PI_Step(const parameter, position: byte; const n_a, tot_n: integer; const tot_t: real; var theta: pvec; var outf, lf: TFIleStream);
(*MS: Gibbs sampler and Metropolis-Hastings step to sample rate parameters corresponding to
- two or more observations
- network and behavior models
assuming constant rate functions; to be extended to non-constant rate functions.*)
var
  a1: integer;
  b1: real;
  accept: boolean;
  candidate: pvec;
begin
(*Note: <candidate> needs not be reset in the current implementation.*)
if gibbs1 then (*Gibbs sampler.*)
  begin
  (*The hyperparameters <gamma_p[parameter,1]> should be chosen to be integer-valued, otherwise acceptance-rejection methods may be used to generate draws from the full conditional distribution.*)
  a1 := tot_n + round(gamma_p[parameter,1]); (*<tot_n> correponds to the total number of changes in the present period.*)
  b1 := (n_a * tot_t) + gamma_p[parameter,2]; (*<tot_t> corresponds to the length of the present time period; <n _a> corresponds to the number of actors which, for the dependent variable under consideration, are active at the beginning of period <period>.*)
  candidate[position] := SampleGamma(a1,b1); (*Sample from full conditional distribution.*)
  accept := true;
  end
else (*Metropolis-Hastings sampler: works fine (passed tests in <Siena 2>), but must be adapted to <Siena 3>.*)
  begin
  {
  CandidateTheta1(p1,protheta,candidate); (*Generates candidate for rate function parameter.*)
  proppoisson := Poisson(candidate); (*Evaluate Poisson probability mass of the number of transitions in the given period with the proposed rate parameter.*)
  prespoisson := Poisson(theta); (*Evaluate Poisson probability mass of the number of transitions in the given period with the present rate parameter.*)
  logpoissonratio := proppoisson - prespoisson; (*Ratio of Poisson masses on natural logarithmic scale.*)
  propgamma := EvaluateGamma(p1,vec1,gamma1,candidate,gamma2); (*Evaluate univariate Gamma density with parameters <gamma1> and <gamma2> for each coordinate of the proposed rate parameter.*)
  presgamma := EvaluateGamma(p1,vec2,gamma1,theta,gamma2); (*Evaluate univariate Gamma density with parameters <gamma1> and <gamma2> for each coordinate of the present rate parameter.*)
  logpratio := propgamma - presgamma; (*Ratio of prior Gamma densities on natural logarithmic scale.*)
  a := protheta[1]; (*First parameter of the candidate-generating Gamma distribution.*)
  proposal := EvaluateGamma(p1,vec3,a,candidate,gamma2); (*Evaluate univariate Gamma density with parameters <gamma1> and <gamma2> for each coordinate of some vector valued random variate.*);
  present := EvaluateGamma(p1,vec4,a,theta,gamma2); (*Evaluate univariate Gamma density with parameters <gamma1> and <gamma2> for each coordinate of some vector valued random variate.*);
  logcratio := proposal - present; (*Ratio of candidate-generating densities on natural logarithmic scale.*)
  logpratio := logpratio + logcratio;
  accept := DecisionMH(pr,logpoissonratio,logpratio,blackbox); (*Decide whether or not variance-covariance candidate is to be accepted.*)
  }
  end;
if accept then
  begin
  p1_accepted[parameter] := p1_accepted[parameter] + 1; (*Keep track of acceptance rate.*)
  theta[position] := candidate[position]; (*Change <position>-th coordinate of <theta>, while leaving the remaining coordinates unchanged.*)
  SetAlpha(theta); (*Set <alpa> to the new parameter vector.*)
  end;
end; (*MS: MH_PI_Step*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MH_PII_Step(const p: byte; const alpha_constant: pvec; var beta, theta: pvec; var outf, lf: TFileStream);
(*MS: Metropolis-Hastings step to sample network and behavior objective and gratification function parameter.
Candidate sampled from
(1) random walk density.
(2) autoregressive density.
(3) independence sampler.*)
var
  cratio, propmn, presmn, promvn, premvn, propmvn, presmvn, pratio, mnratio, pr, scalefactor: real;
  accept: boolean;
  candidate, beta_candidate, sample: pvec;
begin
(*Remarks:
- <alpha_constant>: vector of present network and behavior rate parameters.
- <beta>: vector of present network and behavior objective and gratification function parameters.
- <theta>: vector of parameters, consisting of <alpha_constant> and <beta>.
All probability densities are evaluated (1) up to a constant of proportionality and (2) on the log-scale.*)
{--------------------}
{-Generate candidate-}
{--------------------}
scalefactor := blackbox[12]; (*Scale factor by which Cholesky factor of the variance-covariance matrix of the candidate-generating density is rescaled.*)
CandidateTheta2(p2,choltheta,scalefactor,beta_candidate,sample,beta); (*Generate candidate <beta_candidate> for <beta>; note: <beta_candidate = mu + sample> for some <mu>.*)
InversePartitionParameter(p_n,p,alpha_constant,beta_candidate,candidate); (*Construct the candidate parameter vector <candidate>, consisting of <alpha_constant> and <beta_candidate>.*)
{-----------------------------------------}
{-Ratio of candidate-generating densities-}
{-----------------------------------------}
(*Remark: <scalefactor> may not be constant across iterations.*)
if (cdensity1 < 3) then cratio := 0 (*Random walk density, autoregressive density.*)
else (*Independence sampler.*)
  begin
  (*Remark: the variance-covariance matrix of the candidate-generating density is constant across iterations.*)
  propmvn := EvaluateFunctionMVN2(p2,sample,ictheta,scalefactor); (*Gaussian density of <sample>, corresponding to <beta_candidate>; note: <sample> was generated from <N(0, ctheta * sqr(scalefactor))> with <ictheta = inverse(ctheta)>.*)
  presmvn := EvaluateFunctionMVN2(p2,lastsample,ictheta,scalefactor); (*Gaussian density of <lastsample>, corresponding to <beta>.*)
  cratio := presmvn - propmvn; (*Log ratio of candidate-generating densities.*)
  end;
{--------------------------}
{-Ratio of prior densities-}
{--------------------------}
(*Remark: the prior is <N(0, vtheta)> with <ptheta = inverse(vtheta)>.*)
promvn := EvaluateMVN(p2,beta_candidate,ptheta); (*Gaussian prior density of <beta_candidate>.*)
premvn := EvaluateMVN(p2,beta,ptheta); (*Gaussian prior density of <beta>.*)
pratio := promvn - premvn; (*Log prior ratio.*)
{------------------------------------}
{-Ratio of multinomial probabilities-}
{------------------------------------}
EvaluateMultinomial(1,theta,candidate,re,re,mnlactor1,mnlactor2,outf,lf); (*(Joint) multinomial probabilities given (1) <theta> and (2) <candidate>.*)
presmn := mnlactor1[0]; (*Joint multinomial probability given <theta>.*)
propmn := mnlactor2[0]; (*Joint multinomial probability given <candidate>.*)
mnratio := propmn - presmn; (*Log likelihood ratio.*)
mnratio := mnratio + pratio; (*Log posterior ratio.*)
{----------}
{-Decision-}
{----------}
accept := DecisionMH(pr,mnratio,cratio,blackbox); (*Decide whether or not to accept <candidate>.*)
if accept then
  begin
  blackbox[7] := blackbox[7] + 1; (*Acceptance rate.*)
  SetVector(p2,lastsample,sample); (*Store Gaussian variate from <N(0, ctheta * sqr(scalefactor))> for use on the following iteration.*)
  SetVector(p,theta,candidate); (*Set <theta = candidate>.*)
  SetAlpha(theta); (*Set the coordinates of <alpa> to the corresponding coordinates of <candidate>.*)
  end;
end; (*MS: MH_PII_Step*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MH_RE_CovRE(const p: byte; const theta: pvec; var outf, lf: TFileStream);
(*MS: Metropolis-Hastings step to sample random effects and, in the Bayesian case, (inverse) variance-covariance matrices.*)
begin
(*Remark: <c_covRE> and <i_covRE> are updated by
- ML framework: procedure <Get_Cov_Theta>.
- Bayesian framework: procedure <MH_CovRE_Step>.*)
(*--------------*)
(*Random effects*)
(*--------------*)
CandidateRE(n,nre,c_covRE,scaleRE_i,mvn,re); (*Generate candidate <mvn>.*)
EvaluateMultinomial(2,theta,theta,re,mvn,mnlactor1,mnlactor2,outf,lf); (*Multinomial choice probabilities given <mvn> and <re>.*)
MH_RE_Step(n,nre,i_covRE,mproposal,re,mvn,mnlactor2,mnlactor1); (*Metropolis-Hastings step: decide whether to accept <mvn>.*)
(*------------------------------------*)
(*(Inverse) variance-covariance matrix*)
(*------------------------------------*)
if Bayesian then MH_CovRE_Step(nre,i_covRE,covRE,c_covRE,n,re,psigma,pdegrees,blackbox); (*Gibbs or Metropolis Hastings step to sample a (inverse) variance-covariance matrix.*)
end; (*MS: MH_RE_CovRE*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.Get_Cov(const p: byte; const theta: pvec);
(*MS: Compute the entities <i_covRE>, <c_covRE>, and <covRE> (in ML framework, not: Bayesian framework),
which are required for sampling actor-dependent random effects;
computation of these entities is not strictly necessary on each iteration,
but is safer while doing little harm in terms of efficiency.*)
var det: real;
begin
if maxlike then Get_Cov_Theta(nre,p,theta,li,ksi,c_icovRE,c_covRE,i_covRE,covRE)
else (*Bayesian.*)
  begin
  Cholesky(nre,c_covRE,covRE,det); (*Evaluate the determinant and Cholesky factor of the variance-covariance matrix: required by the following procedure (in the univariate case).*)
  MatrixI(nre,det,i_covRE,covRE,c_covRE);(*Invert the variance-covariance matrix: constant in phase 3.*)
  end;
end; (*MS: Get_Cov*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MCMC_initialise_n(const p: byte);
(*MS: Initialise entities on iteration <n>.*)
var
  h: byte;
  det: real;
begin
for h := 1 to 10 do blackbox[h] := 0;
lag := 1; (*Lag 1 autocorrelation.*)
if Bayesian then
  begin
  SetArray(autoTheta,p,nrunMH_batches); (*Store Sigma to evaluate the <lag> autocorrelation function.*)
  SetArray(p1_accepted,p1); (*Acceptance rates: the rate parameters.*)
  end;
if randomeffects then
  begin
  SetArray(autoRE,nmax,nre,nrunMH_batches); (*Store random effects to evaluate the <lag> autocorrelation function.*)
  SetArray(autoCovRE,nre,nre,nrunMH_batches); (*Store Sigma to evaluate the <lag> autocorrelation function.*)
  SetArray(eb_n,nmax,nre); (*Store Empirical Bayes estimates of actor-dependent random effects at present Stoch.App. iteration.*)
  end;
end; (*MS: MCMC_initialise_n*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MCMC_cycle_storage(const aim: byte; const iteration: integer; var estage: byte; var p: byte; var theta: pvec);
(*MS: At the end of each iteration of the MCMC cycle algorithm, some functions of the latent data
- random effects
- Bayesian framework: parameters
are stored.*)
var
  i: integer;
  MII: pmatrix;
begin
(*Remark: what is stored depends on the aim of the simulations:
(1) default: sampling latent data as required by the estimation algorithm.
(2) pre-estimation: burn-in.
(3) post-estimation: autocorrelations, sampling efficiency, and Monte Carlo standard errors.
*)
case aim of
(*-------*)
(*Default*)
(*-------*)
0:
  begin
  if randomeffects then
    begin
    MatrixAdd(n,nre,eb_n,re); (*To evaluate ergodic average with respect to random effects.*)
    MatrixStore(n,nre,autoRE,iteration,re); (*Store random effects to evaluate lag <lag> autocorrelation function.*)
    if Bayesian then MatrixStore(nre,autoCovRE,iteration,covRE); (*Store variance-covariance matrix of the random effects to evaluate lag <lag> autocorrelation function.*)
    MatrixAdd(nre,nre,sumsigma,covRE); (*Sum over the sampled variance-covariance matrices: in Bayesian algorithm needed to average over MCMC sample, in maximum likelihood algorithm needed to improve the candidate-generating density of the random effects.*)
    if (estage > 30) then
      begin
      MatrixAdd(n,nre,sumre,re); (*Sum over sampled actor-dependent random effects.*)
      RE_positive(n,nre,re,sumposre); (*Evaluate coordinate-wise the indicator function I(re > 0) and increment the count.*)
      end;
    end;
  if Bayesian then
    begin
    VectorStore(p,autotheta,iteration,theta); (*Store parameter vector to evaluate lag <lag> autocorrelation function.*)
    if (estage > 30) then
      begin
      VectorAdd(p,MI_posterior_theta,theta); (*Sum over the first moment of the sampled parameter vectors.*)
      VectorOProd(p,MII,theta); (*Evaluate second moment of the parameter vector, defined as the outer product of the parameter vector.*)
      MatrixAdd(p,p,MII_posterior_theta,MII); (*Sum over the second moment of the sampled parameter vectors.*)
      end;
    end;
  end;  
(*--------------*)
(*Pre-estimation*)
(*--------------*)
1:
if randomeffects then
  begin
  for i := 1 to n do
    begin
    if accepted[i] then actuals[i] := actuals[i] + 1;
    end;
  end;
(*---------------*)
(*Post-estimation*)
(*---------------*)
2:
  begin
  if randomeffects then
    begin
    MatrixStore(n,nre,autoRE,iteration,re); (*Store random effects to evaluate lag <lag> autocorrelation function.*)
    if Bayesian then MatrixStore(nre,autoCovRE,iteration,covRE); (*Store variance-covariance matrix of the random effects to evaluate lag <lag> autocorrelation function.*)
    end;
  if Bayesian then VectorStore(p,autotheta,iteration,theta); (*Store parameter vector to evaluate lag <lag> autocorrelation function.*)
  end;
end; (*<case aim of>.*)  
end; (*MS: MCMC_cycle_storage*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.MCMC_cycle(const setpar: boolean; var outf, lf: TFileStream; var p: byte; var theta: pvec; var estage: byte; const aim: byte);
(*MS: Cycle through
(1) M-H step to sample path: network and behavior (ML and Bayesian models)
(2) M-H step to sample actor-dependent random effects (ML and Bayesian random effects models)
(3) Gibbs and M-H steps to sample parameters (Bayesian models).*)
var
  i, period: byte;
  iteration, minimum, subiteration: integer;
  Error: boolean;
  Errstr: shortstring;
  count: array of integer;
  x: string;
begin
(*Remark:

(a) simple case: MLE without random effects:
--------------------------------------------
latent data: paths:
by the Markov assumption, (network and behavior) paths are conditionally independent given the panel data;
hence the computational burden is the primary decision criterion,
which leads to choice "iterations within periods" (requires less initializations).

(b) complicated case: MLE with random effects or Bayesian with or without random effects:
-----------------------------------------------------------------------------------------
latent data: paths, random effects, and in the Bayesian framework parameters:
dependence between the two or three sets of latent data and hence sampling efficiency
is an additional decision criterion, which leads to the choice "periods within iterations",
since otherwise one would condition on one of the periods for many iterations.
*)

MCMC_initialise_n(p); (*Initialise entities on iteration <n>.*)
if (maxlike and not(randomeffects)) then (*MLE: no random effects.*)
  begin
  minimum := nrunMH; (*The maximum number of accepted proposals in some given period.*)
  for period := 1 to (observations - 1) do if (not skipi[period]) then (*Period.*)
    begin
    blackbox[4] := 0;
    Initialise_actors(period);
    Initialise_y(period);
    AdaptProposalProbabilities;
    if setpar then RenewRateslprobs(Chain[period-1].First,Chain[period-1].Last,false,period,outf,lf); (*Update <lprobs> from first to last change in the embedded chain of <period>.*)
    (*Important note: <Initialise_y, Initialise_actors, AdaptProposalProbabilities>
    must be followed by <RenewRateslprobs> for <Chain[period-1].First> to <Chain[period-1].Last>.*)
    for iteration := 1 to nrunMH do (*Cycle.*)
      begin
      newpath := false; (*Keeps track of whether the proposed path is accepted or not: helps to cut computation time in estimating random effects models.*)
      MH_Step(period,outf,lf,0); (*Metropolis-Hastings step to sample a path (network and behavior) for the given period.*)
      if newpath then blackbox[4] := blackbox[4] + 1; (*Keeps track of acceptance rate of Metropolis-Hastings algorithm to sample paths.*)
      if (ErrorText > '') or (Chain[period-1].ErrorMes > '') then (*Error message concerns sampling of paths.*)
        begin
        x := 'Called from MCMC_cycle from Scores.'+eol;
        outf.Write(pchar(x)^,length(x));
        Error := true;
        Errstr := Errortext + Chain[period-1].ErrorMes;
        exit;
        end;
      end; (*Cycle.*)
    if (blackbox[4] < minimum) then minimum := trunc(blackbox[4]); (*Note: the number of iterations is constant across periods, hence acceptance frequencies are comparable.*)
    end; (*Period.*)
  end (*MLE: no random effects.*)
else (*MLE with random effects or Bayesian with or without random effects.*)
  begin
  count := nil;
  SetLength(count,observations); (*Elements <1..observations-1> are needed.*)
  if (randomeffects and (aim = 1)) then SetArray(actuals,n);
  for iteration := 1 to nrunMH_batches do (*Cycle.*)
    begin
    for period := 1 to (observations - 1) do if (not skipi[period]) then (*Period.*)
      begin
      Initialise_actors(period);
      Initialise_y(period);
      AdaptProposalProbabilities;
      RenewRateslprobs(Chain[period-1].First,Chain[period-1].Last,false,period,outf,lf); (*Update <lprobs> from first to last change in the embedded chain of <period>.*)
      (*Important note: <Initialise_y, Initialise_actors, AdaptProposalProbabilities>
      must be followed by <RenewRateslprobs> for <Chain[period-1].First> to <Chain[period-1].Last>.*)
      for subiteration := 1 to nrunMH do (*Sub-cycle: one subiteration would hardly change the network and behavior path: hence multiple subiterations.*)
        begin
        newpath := false; (*Keeps track of whether the proposed path is accepted or not: helps to cut computation time in estimating random effects models.*)
        MH_Step(period,outf,lf,0); (*Metropolis-Hastings step to sample a path (network and behavior) for the given period.*)
        if newpath then count[period] := count[period] + 1;  (*Keeps track of acceptance rate of Metropolis-Hastings algorithm to sample paths.*)
        if (ErrorText > '') or (Chain[period-1].ErrorMes > '') then (*Error message concerns sampling of paths.*)
          begin
          x := 'Called from MCMC_cycle from Scores.'+eol;
          outf.Write(pchar(x)^,length(x));
          Error := true;
          Errstr := Errortext + Chain[period-1].ErrorMes;
          exit;
          end;
        end; (*Sub-cycle.*) 
      end; (*Period.*)
    (*Some procedures follow that should not be nested in the <period> loop:*)
    if randomeffects then MH_RE_CovRE(p,theta,outf,lf); (*MS: Metropolis-Hastings step to sample random effects and, in the Bayesian case, (inverse) variance-covariance matrices.*)
    if Bayesian then MH_P_Step(p,theta,outf,lf); (*MS: Gibbs and Metropolis-Hastings step to sample parameters.*);
    MCMC_cycle_storage(aim,iteration,estage,p,theta); (*Some functions of the latent data are stored.*)
    end; (*Cycle.*)
  if skipi[1] then minimum := 0 else minimum := count[1];
  for i := 2 to (observations - 1) do if (not skipi[i]) then
    begin
    if (count[i] < minimum) then minimum := count[i]; (*Note: the number of iterations is constant across periods, hence acceptance frequencies are comparable.*)
    end;
  count := nil;
  end;
blackbox[4] := minimum; (*Keeps track of the minimum acceptance rate of Metropolis-Hastings algorithm to sample paths: minimum across periods.*)
(*Important remark:
The cycle algorithm samples, in the following order,
(1) paths (MLE and Bayesian: all models),
(2) actor-dependent random effects (MLE and Bayesian: random effects models),
(3) variance-covariance matrix of actor-dependent random effects (Bayesian: random effects models),
(4) rate parameters (Bayesian: all models),
(5) objective function parameters (Bayesian: all models).
The <lprobs> are required for step (1), but not for steps (2) - (5).
Thus it is sufficient to re-calculate <lprobs> before step (1).
More precisely, re-calculating <lprobs> is necessary if steps (2) - (5) changed parameters or random effects:
* random effects models: by default, multi-block M-H: hence there is a high probability that at least some random effects have been changed.
* Bayesian: by default, Gibbs sampler: hence with probability 1 the rate parameters have been changed.
More efficient algorithms to re-calculate the <lprob> are possible, for example, by using the procedure <UpdateLprobsActors>.
*)
end; (*MS: MCMC_cycle*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.Post_MCMC_cycle(var p: byte; var theta: pvec; var estage: byte; var outf, lf: TFIleStream);
(*MS: When the MCMC cycle algorithm has finished and control is <Siena> is to return to <S_est>,
some functions of the latent data
- random effects
- Bayesian framework: parameters
are stored.*)
var
  h: byte;
  nres: integer;
  minimum: real;
begin
(*First: <blackbox>.*)
if Bayesian then
  begin
  minimum := p1_accepted[1]; (*Rate parameters: minimum acceptance rate.*)
  for h := 2 to p1 do
    begin
    if (p1_accepted[h] < minimum) then minimum := p1_accepted[h];
    end;
  blackbox[6] := minimum;
  end;
if (Bayesian or randomeffects) then
  begin
  for h := 1 to 10 do blackbox[h] := blackbox[h] / nrunMH_batches;
  blackbox[4] := blackbox[4] / nrunMH; (*Acceptance rate of paths: based on <nrunMH_batches * nrunMH> proposals.*)
  end
else
  begin
  for h := 1 to 10 do blackbox[h] := blackbox[h] / nrunMH;
  end;
(*Second: all other entities.*)
if randomeffects then
  begin
  nres := nre;
  Autocorrelation(autocov2,nre,n,nrunMH_batches,autoRE,lag,aREfile); (*Evaluate the lag <lag> autocorrelation function: random effects.*)
  if Bayesian then Autocorrelation(autocov2,nre,nres,nrunMH_batches,autoCovRE,lag,aCovREfile); (*Evaluate the lag <lag> autocorrelation function: variance-covariance matrix of random effects.*)
  MatrixAdd(n,nre,eb,eb_n); (*Ergodic sum with respect to random effects. Must be placed before (!) averaging <eb_n>.*)
  ScalarDiv(n,nre,eb_n,nrunMH_batches); (*Empirical Bayes estimates of random effect on present <S_est> iteration.*)
  SetMatrix(n,nre,lastebre,eb_n);
  end;
if Bayesian then Autocorrelation(autocov1,p,nrunMH_batches,autoTheta,lag,athetafile); (*Evaluate the lag <lag> autocorrelation function: parameter vector.*)
MCMC_report(estage,maxlike,p,theta,randomeffects,nre,covRE,n,re,Bayesian,blackbox,thetafile,covrefile,refile,slrcovrefile,arrefile,adllfile,arpathfile,arcovrefile,arthetafile); (*Report information on how the Markov chain behaves to files.*)
if ((estage = 33) and (randomeffects or Bayesian)) then EvaluateEfficiency(estage,outf,lf,p,theta); (*Final iteration: evaluate efficiency of Markov chain Monte Carlo algorithm.*)
end; (*MS: Post_MCMC_cycle*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.Choose_MC_SE_Estimate(var ascii: TFileStream; const p: byte; const completed: boolean; var mcRE: rrarray; 
var cov_covRE: pmatrix; const cov_theta: pmatrix; var var_theta: R_Ar);
(*MS: There are two estimates of the Monte Carlo variance:
the present procedure selects between them and manipulates the relevant arrays.*)
var i, h: integer;
x: string;
begin
(*Remark:
The default method is the batch means method, which is always available,
but if phase 3 is completed, Initial Sequence Estimates (Geyer, 1992) are better alternatives;
aims:
- select method:
  * if phase 3 is completed, then replace the batch means estimates by the Initial Sequence Estimates.
  * otherwise use batch means estimates.
- report what method is selected.
Note that the estimates, if computed, have already been written to the log-file.
*)
if Bayesian then SetArray(var_theta,p); (*<var_theta> remains to be specified.*)
if completed then (*Phase 3 completed.*)
  begin
    x := 'Monte Carlo standard errors are estimated by Initial Monotone Sequence Estimates (Geyer, 1992).'+eol;
    ascii.Write(pchar(x)^,length(x));
  if Bayesian then
    begin
    for i := 1 to p do var_theta[i] := ISE_theta[i];
    end;
  if randomeffects then
    begin
    for i := 1 to n do
      begin
      for h := 1 to nre do mcRE[i,h] := ISE_RE[i,h];
      end;
    if Bayesian then
      begin
      for i := 1 to nre do
        begin
        for h := 1 to nre do cov_covRE[i,h] := ISE_CovRE[i,h];
        end;
      end;
    end;
  end
else (*By default: batch means estimates, which are used by default for <RE> and <CovRE>, while <var_theta> remains to be specified.*)
  begin
  x := 'Monte Carlo standard errors are estimated by batch means estimates.'+eol;
    ascii.Write(pchar(x)^,length(x));
  if Bayesian then
    begin
    for i := 1 to p do var_theta[i] := cov_theta[i,i]; (*Batch means estimates.*)
    end;
  (*For the remaining parameters and latent data, the batch means estimates are already in the relevant arrays.*)
  end;
   ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
end; (*MS: Choose_MC_SE_Estimate.*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.CD_Score_Hessian_RE(const p: byte; const evaluate_hessian: boolean; var d: pvec; var dd: pmatrix);
(*Evaluate the complete-data score and the complete-data Hessian with respect to
functions of the (co)variance parameters;
see paper on random effects models.*)
var
  i, j, count, dimension, dim_sqr, dim_K, dim_L, dim_S, row, column: byte;
  error: shortstring;
  gradient, gradient_L, gradient_K, vec, vecW: R_Ar;
  A, A_C, A_C_Psi, TC_A_C, nCovRE, t_C_icov, L, i_L, t_i_L, W, IW, IWI: pmatrix;
  BoB, cE_BoB, cE_BoB_PoW, Hessian, Hessian_LL, Hessian_LK, Hessian_KL, Hessian_KK, PoW, PoWL, PoLWL, Pre, Post, Psi, Psi_sqrt, IoIWI, WL, t_L, TLWL: rrarray;
begin
(*Some common computations.*)
EstimateCovIII(nre,n,re,W); (*Evaluate <W>.*)
ScalarProd(nre,nCovRE,CovRE,n); (*Compute <nCovRE = n * CovRE>.*)
MatrixMinus(nre,nre,A,nCovRE,W); (*Compute <A = nCovRE - W>.*)
dim_sqr := sqr(nre); (*Needed for all parametrizations.*)
dim_S := (nre * (nre + 1)) div 2; 
if (estage < 30) then
  begin
  (*----------------------------------------------------------------------------------*)
  (*Parametrization I: estimate convenient functions of the variance-covariance matrix*)
  (*----------------------------------------------------------------------------------*)
  dim_L := (nre * (nre - 1)) div 2; (*Number of non-redundant elements of <Li>: below-diagonal elements.*)
  dim_K := nre; (*Number of non-redundant elements of <Ksi>: diagonal elements.*)
  SetArray(Psi,nre); (*<Psi> is a diagonal, exponential matrix.*)
  SetArray(Psi_sqrt,nre); (*<Psi_sqrt> is the square root of <Psi>.*)
  for i := 1 to nre do
    begin
    Psi[i,i] := expo(Ksi[i,i]);
    Psi_sqrt[i,i] := expo(Ksi[i,i] / 2); (*Compute <sqrt(Psi)>, where <Psi = exp(Ksi)>.*)
    end;
  MatrixProd(nre,nre,A_C,A,C_icovRE); (*Compute <A_C = A * C_icovRE>, where <C_icovRE> is the Cholesky factor of the inverse variance-covariance matrix.*)
  if (nre > 1) then (*In case <nre = 1>, entering the following procedures would be unnecessary and thus inefficient.*)
    begin
    {----------------}
    (*Gradient: <Li>*)
    {----------------}
    Matrix_AdiagB_Prod(nre,A_C_Psi,A_C,Psi_sqrt); (*Compute <A_C_Psi = A_C * Psi_sqrt>, where <Psi_sqrt> is diagonal.*)
    VecOperator(nre,vec,A_C_Psi); (*Transform the matrix <A_C_Psi> into <vec> form.*)
    (*The transposed (!) elimination matrix of <Li> is of order <dim_L x dim_sqr>, not <dim_sqr x dim_L>:*)
    MatrixVecProd(dim_L,dim_sqr,gradient_L,t_Elimination_Li,vec); (*Premultiply by the transposed (!) elimination matrix of <Li>.*)
    end;
  {-----------------}
  (*Gradient: <Ksi>*)
  {-----------------}
  MatrixT(nre,t_C_icov,C_icovRE); (*Transpose the Cholesky factor of the inverse variance-covariance matrix.*)
  MatrixProd(nre,nre,TC_A_C,t_C_icov,A_C); (*Premultiply <A_C> by the transpose of the Cholesky factor of the inverse variance-covariance matrix.*)
  VecOperator(nre,vec,TC_A_C); (*Transform the matrix into <vec> form.*)
  (*The transposed (!) elimination matrix of <Ksi> is of order <dim_K x dim_sqr>, not <dim_sqr x dim_K>:*)
  MatrixVecProd(dim_K,dim_sqr,gradient_K,t_Elimination_Ksi,vec); (*Premultiply by the transposed (!) elimination matrix of <Ksi>.*)
  for i := 1 to dim_K do gradient_K[i] := gradient_K[i] / 2; 
  {----------}
  {*Gradient*}
  {----------}
  dimension := dim_K + dim_L; (*Dimension of the gradient.*)
  SetArray(gradient,dimension);
  count := 0;
  for i := 1 to dim_L do
    begin
    Inc(count);
    gradient[count] := gradient_L[i];
    end;
  for i := 1 to dim_K do
    begin
    Inc(count);
    gradient[count] := gradient_K[i];
    end;
  if evaluate_hessian then (*Evaluate Hessian; redundant in estimation phase 2.*)
    begin
    MatrixProd(nre,nre,WL,W,Li); (*Compute <WL = W * Li>, needed both in case <nre = 1> and <nre > 1>.*)
    if (nre > 1) then (*In case <nre = 1>, entering the following procedures would be unnecessary and thus inefficient.*)
      begin
      (*-------------------*)
      (*Hessian: <Li>, <Li>*)
      (*-------------------*)
      MakeStaticII(nre,L,Li); (*<L> is the static equivalent of <Li>.*)
      PInverse(nre,L,i_L,error); (*Invert <L> to obtain <i_L>.*)
      MatrixT(nre,t_i_L,i_L); (*Transpose <i_L> to obtain <t_i_L>.*)
      KroneckerProd(nre,nre,nre,nre,BoB,t_i_L,i_L); (*Compute <t_i_L o i_L>, where <o> is the Kronecker product.*)
      MatrixProd(cE_BoB,dim_sqr,cE,BoB); (*Premultiply the Kronecker product with <- n E>, the commutation matrix of order <dim_sqr x dim_sqr>, which was rescaled by the scalar <c = - n>.*)
      KroneckerProd(nre,nre,nre,nre,PoW,Psi,W); (*Compute <Psi o W>, where <o> is the Kronecker product.*)
      MatrixMinus(dim_sqr,dim_sqr,cE_BoB_PoW,cE_BoB,PoW); (*Compute <cE_BoB_PoW = cE_BoB - PoW>.*)
      MatrixProd(dim_L,dim_sqr,Pre,t_Elimination_Li,cE_BoB_PoW); (*Premultiply by transposed (!) elimination matrix of <Li>.*)
      (*Note: <Pre> is of order <dim_L x dim_sqr>, the elimination matrix of order <dim_sqr x dim_L>, and the resulting Hessian matrix of order <dim_L x dim_L>:*)
      MatrixP(dim_L,dim_sqr,dim_L,Hessian_LL,Pre,Elimination_Li); (*Postmultiply by elimination matrix of <Li>.*)
      (*--------------------*)
      (*Hessian: <Li>, <Ksi>*)
      (*--------------------*)
      (*Note: the resulting matrix needs not be square and symmetric!*)
      KroneckerProd(nre,nre,nre,nre,PoWL,Psi,WL); (*Compute <PoWL = Psi * WL>.*)
      MatrixProd(dim_L,dim_sqr,Pre,t_Elimination_Li,PoWL); (*Premultiply by transposed (!) elimination matrix of <Li>; the resulting matrix is of order <dim_L x dim_sqr>.*)
      (*Note: <Pre> is of order <dim_L x dim_sqr>, the elimination matrix of order <dim_sqr x dim_K>, and the resulting Hessian matrix of order <dim_L x dim_K>:*)
      MatrixP(dim_L,dim_sqr,dim_K,Hessian_LK,Pre,Elimination_Ksi); (*Postmultiply by elimination matrix of <Ksi>.*)
      ScalarProd(dim_L,dim_K,Hessian_LK,-2); (*Multiply by <-2>.*)
      MatrixT(dim_L,dim_K,Hessian_KL,Hessian_LK); (*Transpose Hessian matrix; note that the non-diagonal block of the Hessian is not symmetric.*)
      end;
    (*---------------------*)
    (*Hessian: <Ksi>, <Ksi>*)
    (*---------------------*)
    MatrixT(nre,t_L,Li); (*Transpose <Li> to obtain <t_L>.*)
    MatrixProd(nre,nre,TLWL,t_L,WL); (*Compute <TLWL = t_L * W * L>.*)
    KroneckerProd(nre,nre,nre,nre,PoLWL,Psi,TLWL); (*Compute <PoLWL = Psi * TLWL>.*)
    MatrixProd(dim_K,dim_sqr,Pre,t_Elimination_Ksi,PoLWL);  (*Premultiply by transposed (!) elimination matrix of <Ksi>.*)
    (*Note: <Pre> is of order <dim_K x dim_sqr>, the elimination matrix of order <dim_sqr x dim_K>, and the resulting Hessian matrix of order <dim_K x dim_K>:*)
    MatrixP(dim_K,dim_sqr,dim_K,Hessian_KK,Pre,Elimination_Ksi); (*Postmultiply by elimination matrix of <Ksi>.*)
    ScalarDiv(dim_K,dim_K,Hessian_KK,-2); (*Divide by <-2>.*)
    (*-------*)
    (*Hessian*)
    (*-------*)
    SetArray(Hessian,dimension); (*The Hessian matrix is of order <dimension x dimension>.*)
    row := 0;
    for i := 1 to dim_L do
      begin
      Inc(row);
      column := 0;
      for j := 1 to dim_L do (*LL-block.*)
        begin
        Inc(column);
        Hessian[i,j] := Hessian_LL[row,column];
        end;
      column := 0;
      for j := (dim_L + 1) to dimension do (*LK-block.*)
        begin
        Inc(column);
        Hessian[i,j] := Hessian_LK[row,column];
        end;
      end;
    row := 0;  
    for i := (dim_L + 1) to dimension do
      begin
      Inc(row);
      column := 0;
      for j := 1 to dim_L do (*KL-block.*)
        begin
        Inc(column);
        Hessian[i,j] := Hessian_KL[row,column];
        end;
      column := 0;  
      for j := (dim_L + 1) to dimension do (*KK-block.*)
        begin
        Inc(column);
        Hessian[i,j] := Hessian_KK[row,column];
        end;
      end;
    end;
  end
else
  begin
  (*---------------------------------------------------------------------------------------------------------------*)
  (*Parametrization II: used to estimate the information matrix of the (MCMC)MLE of the variance-covariance matrix.*)
  (*---------------------------------------------------------------------------------------------------------------*)
  (*Note: the constant terms are evaluated in the procedure <InitialiseMS_PhaseIII>.*)
  {----------}
  (*Gradient*)
  {----------}
  VecOperator(nre,vecW,W); (*Evaluate <vec(W)>.*)
  MatrixVecProd(dim_S,dim_sqr,vec,const_S_m,vecW); (*Compute <const_S_m * vec(W)>; the resulting vector is of dimension <dim_S>.*)
  SetArray(gradient,dim_S);
  for i := 1 to dim_S do gradient[i] := vec[i] - const_S_v[i];
  if evaluate_hessian then (*Evaluate Hessian; redundant in estimation phase 2.*)
    begin
    (*-------*)
    (*Hessian*)
    (*-------*)
    MatrixProd(nre,nre,IW,i_CovRE,W); (*Compute <IW = i_CovRE * W>.*)
    MatrixProd(nre,nre,IWI,IW,i_CovRE); (*Compute <IWI = i_CovRE * W * i_CovRE>.*)
    KroneckerProd(nre,nre,nre,nre,IoIWI,i_CovRE,IWI); (*Compute <IoIWI = i_CovRE * IWI>.*)
    MatrixProd(dim_S,dim_sqr,Pre,t_dupl,IoIWI); (*Compute <Pre = transposed duplication matrix * IoIWI>; note that the order of the transposed duplication matrix is <dim_S x dim_sqr>.*)
    (*Note: <Pre> is of order <dim_S x dim_sqr>, the duplication matrix of order <dim_sqr x dim_S>, and the resulting Hessian matrix of order <dim_S x dim_S>:*)
    MatrixP(dim_S,dim_sqr,dim_S,Post,Pre,dupl); (*Compute <Post = transposed duplication matrix * IoIWI * duplication matrix>.*)
    MatrixMinus(dim_S,dim_S,Hessian,const_H,Post); (*Compute <Hessian = const_H - Post>.*)
    end;
  end;
(*------------------------*)
(*Gradient, Hessian: theta*)
(*------------------------*)
dimension := p - dim_S; (*Note: <dim_K + dim_L = dim_S>, hence the statement is valid whether parametrization I or II is active.*)
(*Store gradient in gradient with respect to <theta>:*)
count := 0;
for i := (dimension + 1) to p do
  begin
  Inc(count);
  d[i] := gradient[count];
  end;
if evaluate_hessian then (*Evaluate Hessian; redundant in estimation phase 2.*)
  begin
  (*Store Hessian matrix in Hessian matrix with respect to <theta>:*)
  for i := 1 to dimension do
    begin
    for j := (dimension + 1) to p do dd[i,j] := 0; (*Above-diagonal null block.*)
    end;
  row := 0;
  for i := (dimension + 1) to p do
    begin
    Inc(row);
    column := 0;
    for j := 1 to dimension do dd[i,j] := 0; (*Below-diagonal null block.*)
    for j := (dimension + 1) to p do
      begin
      Inc(column);
      dd[i,j] := - Hessian[row,column]; (*Remark: negative sign of the Hessian matrix: see Louis (1982), ML and random effects paper, and <Scores>.*)  
      end;
    end;
  end;
(*Clean up...*)
gradient := nil;
gradient_L := nil;
gradient_K := nil;
vec := nil;
vecW := nil;
BoB := nil;
cE_BoB := nil;
cE_BoB_PoW := nil;
Hessian := nil;
Hessian_LL := nil;
Hessian_LK := nil;
Hessian_KL := nil;
Hessian_KK := nil;
PoW := nil;
PoWL := nil;
PoLWL := nil;
Pre := nil;
Psi := nil;
Psi_sqrt := nil;
IoIWI := nil;
Post := nil;
WL := nil;
t_L := nil;
TLWL := nil;
end; (*MS: CD_Score_Hessian_RE*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.InitialiseMS_PhaseIII;
(*MS: when estimating random effects models, phase 3 uses an alternative parametrization;
here, some constant vectors and matrices are evaluated, which, owing to their constant nature,
must be evaluated only once.*)
var
  dim_1, dim_2, i: byte;
  det: real;
  vecS: R_Ar;
  IoI: rrarray;
begin
(*First constant part of the score:*)
Cholesky(nre,C_CovRE,CovRE,det); (*Evaluate the determinant and Cholesky factor of the variance-covariance matrix: required by the following procedure (in the univariate case).*)
MatrixI(nre,det,i_CovRE,CovRE,C_CovRE);(*Invert the variance-covariance matrix: constant in phase 3.*)
DuplicationMatrix(nre,dupl); (*Duplication matrix.*)
dim_1 := sqr(nre);
dim_2 := (nre * (nre + 1)) div 2;
MatrixT(dim_1,dim_2,t_dupl,dupl); (*Transpose of duplication matrix.*)
KroneckerProd(nre,nre,nre,nre,IoI,i_CovRE,i_CovRE); (*Compute <inverse(Sigma) o inverse(Sigma), where <o> is the Kronecker product.*)
MatrixProd(dim_2,dim_1,const_S_m,t_dupl,IoI); (*Compute <transposed duplication matrix * IoI>; note that the order of the transposed duplication matrix is <dim_2 x dim_1>.*)
ScalarDiv(dim_2,dim_1,const_S_m,2); (*Scale down by factor <2>.*)
(*Second constant part of the score:*)
VecOperator(nre,vecS,CovRE); (*Evaluate <vec(Sigma)>.*)
MatrixVecProd(dim_2,dim_1,const_S_v,const_S_m,vecS); (*Compute <const_S_m * vecS>; the resulting vector is of dimension <dim_2>.*)
for i := 1 to dim_2 do const_S_v[i] := n * const_S_v[i]; (*Scale up by factor <n>: note that it was already scaled down by factor <2> so that it is effectively rescaled by factor <n/2>.*)
(*Constant term of the Hessian:*)
MatrixP(dim_2,dim_1,dim_2,const_H,const_S_m,dupl); (*Compute <const_S_m * dupl>, where <const_S_m> is of order <dim_2 x dim_1>, <dupl> of order <dim_1 x dim_2>, and the resulting matrix of order <dim_2 x dim_2>.*)
ScalarProd(dim_2,dim_2,const_H,n); (*Scale up by factor <n>: note that <const_S_m> was already scaled down by <2> so that it is effectively rescaled by factor <n/2>.*)
vecS := nil;
IoI := nil;
end; (*MS: InitialiseMS_PhaseIII*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.OpenFiles;
(*MS: Prepare files for MCMC output.*)
var a, s: shortstring;
begin
a := CurrentProjectName;
s := a + 'ARPATH.m';
OpenFile(arPATHfile,s); (*Acceptance rate: paths.*)
s := a + 'Theta.m';
OpenFile(thetafile,s); (*Parameter vector (rate and non-rate parameter vector).*)
if Bayesian then
  begin
  s := a + 'ARTheta.m';
  OpenFile(arThetafile,s); (*Acceptance rate: parameter vector (rate and non-rate parameter vector).*)
  s := a + 'ATheta.m';
  OpenFile(aThetafile,s); (*Autocorrelations: parameter vector (rate and non-rate parameter vector).*)
  end;
if randomeffects then
  begin
  s := a + 'RE.m';
  OpenFile(refile,s); (*Actor-dependent random effects.*)
  s := a + 'ARRE.m';
  OpenFile(arREfile,s); (*Acceptance rate: actor-dependent random effects.*)
  s := a + 'ARE.m';
  OpenFile(aREfile,s); (*Autocorrelations: actor-dependent random effects.*)
  s := a + 'ADLL.m';
  OpenFile(ADLLfile,s); (*Augmented-Data LogLikelihood (up to proportionality constant).*)
  s := a + 'CovRE.m';
  OpenFile(covREfile,s); (*Variance-covariance matrix of random effects.*)
  if Bayesian then
    begin
    s := a + 'ARCovRE.m';
    OpenFile(arCovREfile,s); (*Acceptance rate: variance-covariance matrix of the actor-dependent random effects.*)
    s := a + 'ACovRE.m';
    OpenFile(aCovREfile,s); (*Autocorrelations: variance-covariance matrix of the actor-dependent random effects.*)
    s := a + 'SLRCovRE.m';
    OpenFile(SLRCovREfile,s); (*Smallest latent root of the variance-covariance matrix of the actor-dependent random effects.*)
    end;
  end;
end; (*MS: OpenFiles*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.CloseFiles;
(*MS: Close MCMC output files.*)
begin
closefile(arPATHfile); (*Acceptance rate: paths.*)
closefile(thetafile); (*Parameter vector (rate and non-rate parameter vector).*)
if Bayesian then
  begin
  closefile(arThetafile); (*Acceptance rate: parameter vector (rate and non-rate parameter vector).*)
  closefile(aThetafile); (*Autocorrelations: parameter vector (rate and non-rate parameter vector).*)
  end;
if randomeffects then
  begin
  closefile(refile); (*Actor-dependent random effects.*)
  closefile(arREfile); (*Acceptance rate: actor-dependent random effects.*)
  closefile(aREfile); (*Autocorrelations: actor-dependent random effects.*)
  closefile(ADLLfile); (*Augmented-Data LogLikelihood (up to proportionality constant).*)
  closefile(covREfile); (*Variance-covariance matrix of random effects.*)
  if Bayesian then
    begin
    closefile(arCovREfile); (*Acceptance rate: variance-covariance matrix of the actor-dependent random effects.*)
    closefile(aCovREfile); (*Autocorrelations: variance-covariance matrix of the actor-dependent random effects.*)
    closefile(SLRCovREfile); (*Smallest latent root of the variance-covariance matrix of the actor-dependent random effects.*)
    end;
  end;
end; (*MS: CloseFiles*)

(*--------------------------------------------------------------------------------------------------*)

procedure MLModel.FinaliseMS;
(*Assign nil operator to arrays used in random effects models and Bayesian models.*)
begin
actuals := nil;
mnlactor1 := nil;
mnlactor2 := nil;
const_S_v := nil;
p1_accepted := nil;
ISE_theta := nil;
lastebre := nil;
li := nil;
ksi := nil;
elimination_li := nil;
elimination_ksi := nil;
t_elimination_li := nil;
t_elimination_ksi := nil;
ce := nil;
dupl := nil;
t_dupl := nil;
const_S_m := nil;
const_H := nil;
ISE_RE := nil;
ISE_CovRE := nil;
end; (*MS: FinaliseMS*)

(*--------------------------------------------------------------------------------------------------*)

{
procedure Quantile(const vector: R_Ar; const ldim: integer; const criterion: real);
var
  i, j, h, count: integer;
  a, b: real;
  success: boolean;
  o_vector: R_Ar;
begin
SetArray(o_vector,ldim);
if (vector[1] < vector[2]) then
  begin
  o_vector[1] := vector[1];
  o_vector[2] := vector[2];
  end
else
  begin
  o_vector[1] := vector[2];
  o_vector[2] := vector[1];
  end;
count := 1; (*Note: <count> is not the number of already ordered elements, but <count + 1> is the number of already ordered elements.*)
for i := 3 to ldim do
  begin
  h := 0;
  repeat
  Inc(h);
  a := o_vector[h];
  b := o_vector[h+1]
  success := vector[i] in [a, b];
  until (success or (h = count));
  for j := count downto (h + 1) do (*Move all elements exceeding <vector[i]> one position upwards.*)
    begin
    o_vector[j+1] := o_vector[j]; (*Move the value in position <j> to position <j + 1> so that <j> becomes unoccupied.*)
    end;
  o_vector[h+1] := vector[i]; (*The position <h+1> is free.*)
  Inc(count); (*Increment <number> since one element has been added.*)

  (*Cases: equal to lower bound, equal to upper bound, strictly in between.*)

  end;
if not(criterion in [0, 1]) then criterion := 0.05;
if criterion < 0.5) then criterion := 1 - criterion;
a := (criterion * ldim;
b := (1 - criterion) * ldim;
o_vector[trunc(a)];
o_vector[trunc(b)];
end;
}
