unit s_re;
(*Entire unit by MS:
- six Metropolis-Hastings algorithms to sample from p(v | x, w) indexed by parameter theta,
  where v corresponds to actor-dependent random effects in the network objective and gratification
  function, x to two networks observed on some given actors set at two subsequent observation points
  (plus covariates), and w to the latent sequence of tie changes between the two observation points.
  Three choices: (1) random-walk M-H, (2) autoregressive M-H, (3) independent M-H (independence sampler).
  Each can be used either as one-block M-H (actors altogether) or multi-block M-H (actors separately).
- maximum likelihood routines.
- Bayesian routines.
- other basic routines, in particular output routines (not limited to actor-dependent random effects).*)

(*--------------------------------------------------------------------------------------------------*)

interface

(*--------------------------------------------------------------------------------------------------*)

uses classes, Sysutils, s_constants, rangen, slib, s_dat, s_base, s_matrices, s_mcmc;

const zero = 0;

var
  (*********)
  (*Scalars*)
  (*********)
  estage: (*Indicates stage of estimation process.*)
    byte;
  nno, (*Number of iterations of the latest estimation phase in <S_est>.*)
  actors, (*Number of actors: formal parameter.*)
  degrees, (*Degrees of freedom.*)
  pdegrees: (*Degrees of freedom of the Wishart posterior probability density correponding to the precision matrix of the actor-dependent random effects.*)
    integer;
  acceptancepr, (*Acceptance probability: Metropolis-Hastings.*)
  threshold: real; (*Threshold value, used in monitoring divergence of actor-dependent random effects.*)
  hit2: boolean;
  refile: textfile; (*ASCII file containing actor-dependent random effects.*)
 (*********)
  (*Vectors*)
  (*********)
  autocov1, (*Instrumental array, used to store MCMC sample autocovariances.*)
  propmn, (*MultiNomial probabilty mass evaluated at proposed point.*)
  presmn, (*MultiNomial probabilty mass evaluated at present point.*)
  propMVN, (*MultiVariate Normal kernel evaluated at proposed point.*)
  presMVN, (*MultiVariate Normal kernel evaluated at present point.*)
  mvndensities, (*MultiVariate Normal densities, corresponding to one or more vectors.*)
  scaleRE_i, (*Actor-dependent scale factors to rescale variance-covariance matrices of candidate-generating densities.*)
  scalebeta_i: (*Coordinate-dependent scale factors to rescale variance-covariance matrices of candidate-generating densities.*)
    R_Ar;
  accepted: (*Multi-block Metropolis-Hastings, actor-dependent random effects: whether or not the last candidates were accepted.*)
    boarray;
  (**********)
  (*Matrices*)
  (**********)
  cholfac, (*Cholesky factor of symmmetric, positive-definite variance-covariance matrix.*)
  covRE, (*Variance-covariance matrix of actor-dependent random effects: in general actual parameter in the Delphi / Pascal sense.*)
  sigma, (*Variance-covariance matrix of actor-dependent random effects: in general formal parameter in the Delphi / Pascal sense.*)
  isigma, (*Inverse variance-covariance matrix of actor-dependent random effects: precision matrix.*)
  omega, (*Scale matrix of Wishart probability density.*)
  ssigma, sumsigma: (*Some intrumental variance-covariance matrices, used in computations.*)
    pmatrix;
  autocov2, autotheta, (*Autocorrelation-storing arrays.*)
  mvn, (*Matrix corresponding to actor-dependent random effects: after (!) linear transformations, candidate.*)
  lastmvn0, (*Matrix corresponding to actor-dependent random effects: before (!) linear transformations, present state.*)
  {
  re, (*Matrix corresponding to actor-dependent random effects: after (!) linear transformations, present state.*)
  }
  eb, (*Empirical Bayes estimates of actor-dependent random effects.*)
  eb_n, (*Empirical Bayes estimates of actor-dependent random effects, based on <S_ml> subphase.*)
  posre, (*(Conditional posterior) probability of actor-dependent random effect being positive.*)
  reflectionre, (*Matrix corresponding to actor-dependent random effects: exploited in candidate-generating density to (1) reflect candidates around some point and (2) other linear transformations of candidates.*)
  intermediate, (*Matrix corresponding to actor-dependent random effects: instrumental, used in computations.*)
  sumre, (*Used to evaluate batch means estimates of Monte Carlo s.e..*)
  sumposre, (*(Conditional posterior) probability of actor-dependent random effect being positive.*)
  mcre, (*Monte Carlo s.e. of ergodic averages corresponding to the actor-dependent random effects.*)
  mcposre: (*Monte Carlo s.e. of ergodic averages corresponding to the (conditional posterior) probability of actor-dependent random effect being positive.*)
    rrarray;
  (**************************)
  (*Three-dimensional arrays*)
  (**************************)
  autore, (*Store actor-dependent random effects to evaluate MCMC sample autocorrelations.*)
  autocovre, (*Store (inverse) variance-covariance matrices of actor-dependent random effects to evaluate MCMC sample autocorrelations.*)
  mc_re, (*Used to store actor-dependent random effects sampled from the posterior density.*)
  mc_posre, (*Used to store posterior probability of positive value of actor-dependent random effects sampled from the posterior density.*)
  mc_covre: (*Used to store the (inverse) variance-covariance matrices sampled from the posterior density.*)
    rrrarray;

function NormConstant(var dim: byte): real;

function NumberActors: integer;
(*Gets the number of actors.*)

procedure DefaultRE(var actors: integer; var dim: byte; var re: rrarray);
(*Set default values of random effects.*)

procedure InitialSigma(const dim: byte; var matrix1: pmatrix; const real1: real; var ascii: TFileStream);
(*Initial, crude estimate of variance-covariance matrix of random effects.*)

procedure InitialRE(const actors: integer; const dim: byte; const cholfac: pmatrix; var re: rrarray);
(*Initial values of the random effects: sample from multivariate normal distribution.*)

procedure SaveRE(const actors: integer; const dim: byte; const matrix3: rrarray);
(*Write the ergodic average of the actor-dependent random effects to be used in future runs as starting point.*)

procedure ReadCov(const filename: shortstring; const dim: byte; var matrix1: pmatrix);
(*Exploited in Monte Carlo simulations: read true variance-covariance matrix of the random effects.*)

procedure WriteInitialEstimateFE(var dim: byte; var theta: pvec);
(*Write initial estimates of the fixed effects, obtained by unconditional moment estimation, to file.*)

procedure ReadInitialEstimateFE(var theta: pvec);
(*Read initial estimates of the fixed effects, obtained by unconditional moment estimation, from file.*)

procedure WriteCovTheta(var boolean1, boolean2: boolean; var dim1: byte; var ldim: integer; var omikron, sigma: pmatrix);
(*Writes variance-covariance matrix of parameter estimates to file, as estimated by the Delta method.*)

procedure WriteInitialCovTheta(var dim: byte; var sigma: pmatrix);
(*Write initial estimates of variance-covariance matrix of the parameter vector, obtained by the Delta method, to file.*)

procedure ReadCovTheta(const dim: byte; var sigma: pmatrix; var ascii: TFileStream);
(*Reads variance-covariance matrix of candidate-generating (multivariate) Gaussian density of the parameter vector from file.*)

procedure ReadNetwork(const input: shortstring; const actors: integer; var x: rrarray);

procedure EvaluateObservedDistances(const boolean1: boolean; const actors: integer; const observations: integer; var ascii: TFileStream);
(*Evaluate row-wise distance between two adjacency matrices on some given actor set.*)

procedure IntroduceRE(var ascii: TFileStream; var thetarandom: pbvec; var Description: ppString; var f1: textfile;
const sequences: integer; var ldim1: integer; var f2: textfile; var re, mvn: rrarray; var actors: integer; var dim: byte; var boolean1: boolean);
(*Write specification of random part to the output file.*)

procedure InitialActionsRE(var randomeffects: boolean; var ascii: TFileStream; var thetarandom: pbvec;
var Description: ppstring; var f1: textfile; const sequences: integer; var ldim1: integer;
var f2: textfile; var re, mvn: rrarray; var actors: integer; var dim: byte; var boolean1: boolean);
(*Initial actions concerning random effects.*)

procedure ResetBlackBox(var Bayesian, randomeffects: boolean; var blackbox: blackarray; var scalebeta_i: R_Ar;
var dim1: byte; var scaleRE_i: R_Ar; var actors: integer; var dim2: byte);
(*Reset blackbox to default values. Must be placed before (!) <MCMC_intro>.*)

procedure SetREactor(const dim: byte; const actor: integer; var re: rrarray; const mvn: rrarray);
(*Set <re = mvn> for actor <actor>.*)

procedure RE_positive(var actors: integer; var dim: byte; var re, sumposre: rrarray);
(*Evaluate coordinate-wise the indicator function I(re > 0) and increment the count.*)

function CR(var actors: integer; var detsigma: real; var theta: pvec; const totaltime: real; const totnumber: integer): real;
(*Evaluate the remaining terms of the augmented-data log-likelihood.
Observe that it is assumed that in the rate function there is the intercept, but no other fixed effects.*)

procedure SetProposalRE(const condition: boolean; const actors: integer; const dim: byte; const re: rrarray;
var eb, lastmvn0, reflectionre: rrarray; var sumsigma: pmatrix; const ldim: integer);
(*Prepare some things to be exploited by the candidate-generating density of the actor-dependent random effects.*)

procedure CandidateRE(const actors: integer; const dim: byte; const cholfac: pmatrix;
const scaleRE_i: R_Ar; var mvn: rrarray; const re: rrarray);
(*Generate candidates for actor-dependent random effects.*)

procedure MH_RE_Step(const actors: integer; const dim: byte; const isigma: pmatrix; const multi_block: boolean;
var re, mvn: rrarray; var propmn, presmn: R_Ar);
(*Metropolis-Hastings step to sample actor-dependent random effects.
Given: candidate <mvn> sampled from
(1) random walk density.
(2) autoregressive density.
(3) independence sampler.
To be carried out:
- multi-block Metropolis-Hastings step.
- one-block Metropolis-Hastings step.
*)

function DetScaleSigma(var dim: byte; var sigma: pmatrix; var dim1: byte): real;
(*Evaluates some function of the scale matrix of the prior probability density corresponding to the variance-covariance matrix of the random effects.*)

function EvaluateESCovRE(var dim1, dim2: byte): real;
(*Evaluate the Exponent of the Sample Covariance matrix of the random effects for the Inverse-Wishart density.*)

procedure EvaluateEB(var dim: byte; var ldim1: integer; var matrix3: rrarray; var ldim2: integer; var matrix4: rrarray);
(*Evaluate Empirical Bayes mean and variance of the actor-dependent random effects.*)

function DivergingRE(var actors: integer; var dim: byte; var re: rrarray; var threshold: real; var ascii: textfile; var sigma: pmatrix): boolean;
(*Checks whether some actor-dependent random effects are diverging to infinity.*)

procedure EstimateCovI(var dim: byte; var actors: integer; var re: rrarray; var sigma: pmatrix);
(*Unbiased estimate of variance-covariance matrix.*)

procedure EstimateCovII(var dim: byte; var actors: integer; var re: rrarray; var sigma: pmatrix);
(*Maximum likelihood estimate of variance-covariance matrix: unbiased if mean vector is not estimated, otherwise biased.*)

procedure EstimateCovIII(var dim: byte; var actors: integer; var re: rrarray; var sigma: pmatrix);
(*Compute <sum re * transpose(re)>.*)

procedure Get_Ksi_Li_Cholesky(const dim: byte; const c_cov: pmatrix; var c_icov: pmatrix; var li, ksi: rrarray);
(*Obtain, from the Cholesky factor <c_cov> of the variance-covariance matrix,
- the Cholesky factor <c_icov> of the inverse variance-covariance matrix
- and the matrices <li> and <ksi>.*)

procedure Get_Ksi_Li_Cov(const dim: byte; var c_cov: pmatrix; const sigma: pmatrix;
var li, ksi: rrarray; var count: byte; var theta: pvec);
(*Obtain, from the variance-covariance matrix,
- the Cholesky factor
- the vectors <v(Li)> and <v(Ksi)>, and insert them into the parameter vector <theta>.*)

procedure Get_Cov_Theta(const dim_re, dim_theta: byte; const theta: pvec; var li, ksi: rrarray; var c_icov, c_cov, i_sigma, sigma: pmatrix);
(*Obtain, from the parameter vector, the Cholesky factor of the variance-covariance matrix of the actor-dependent random effects,
and the variance-covariance matrix itself.*)

procedure MCMC_Storage(var dim1: byte; var vector1, vector2: pvec;
var randomeffects: boolean; var dim2: byte; var matrix1: pmatrix);
(*Store MCMC sample to compute average value.*)

procedure WriteScaleCDRE(var ascii: TFileStream; var actors: integer; var scaleRE_i: R_Ar);
(*Writes the actor-dependent scale factors corresponding to the candidate-generating density for the random effects M-H steps to file.*)

procedure WriteScaleCDtheta(var ascii: TFileStream; var scale: real);
(*Writes the scale factor corresponding to the multivariate candidate-generating density for the parameter vector M-H steps to log-file.*)

procedure DistributionCovRE(var randomeffects: boolean; const sequences: integer; var dim: byte; var f2: textfile;
var matrix1: pmatrix; var ldim1: integer);
(*Completes Matlab source file to plot sampling distribution of the estimate of the variance-covariance
matrix of the random effects. Needs to be placed after the <for ldim1 := 1 to sequences> loop and before the
<ConvergenceCovRE> procedure, where the file is closed.*)

procedure ConvergenceCovRE(var randomeffects: boolean; var f1: textfile);
(*Complete Matlab source files containing trace plot of variance-covariance matrix of the random effects.*)

procedure CompleteThetaFile(var f1: textfile; var boolean1: boolean; var f2: textfile);
(*Complete Matlab source file to generate trace plot of fixed effects (parameter) estimates, and acceptance rates.*)

procedure CompleteADLLfile(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of augmented-data likelihood (up to proportionality constant).*)

procedure CompleteARPATHFilee(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of acceptance rates of the Metropolis-Hastings
sampler to sample random effects.*)

procedure CompleteARREFile(const boolean1, boolean2: boolean; var f1, f2: textfile);
(*Complete Matlab source file to generate trace plot of acceptance rates of the Metropolis-Hastings
sampler to sample random effects and the variance-covariance matrix of the random effects.*)

procedure CompleteATheta(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of MCMC sample lag <lag> autocorrelation of rate and non-rate parameter vector.*)

procedure CompleteAREFile(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of MCMC sample lag <lag> autocorrelation among actor-dependent random effects.*)

procedure CompleteACovREfile(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of MCMC sample lag <lag> autocorrelation: variance-covariance
matrix of the actor-dependent random effects.*)

procedure CompleteREFile(var boolean2: boolean; var f1, f2: textfile);
(*Complete Matlab source file to generate trace plot of actor-dependent random effects.*)

procedure OutputCovRE(var ascii: TFileStream; const dim1, dim2, covariance_parameters: byte; var sigma, matrix1, matrix2: pmatrix;
const Bayesian: Boolean; const Description: ppstring);
(*Output variance-covariance matrix of random effects.*)

procedure SaveOutput(const maxlike, Bayesian: boolean; const dim1: byte; const cov1: pmatrix; const dim2: byte; const cov2: pmatrix);
(*Save output that may be used in future runs to tune MCMC algorithms.*)

procedure InfoRE(var boolean1: boolean; var ascii: textfile);
(*Initial estimates of fixed effects by the method of moments, no standard errors estimated.*)

procedure PrepareEB(var boolean1: boolean; var rrarray1: rrarray; var ldim: integer; var dim: byte);

procedure OutputEB_RE(var ascii: TFileStream; const actors: integer; const dim: byte; var eb, mcre, posre, mcposre: rrarray);
(*Write (Empirical) Bayes estimates of actor-dependent random effects (including posterior odds and corresponding Monte Carlo s.e.) to output file.*)

procedure CheckFiles(const maxlike: boolean; const sequences: integer; var ascii: TFileStream; const ldim: integer; const randomeffects, Bayesian: boolean);
(*Write information to output file concerning MCMC output files generated by Siena to detect non-convergence / study sampling distributions.*)

(*--------------------------------------------------------------------------------------------------*)

implementation

uses s_mcmc_II;

(*--------------------------------------------------------------------------------------------------*)

function NormConstant(var dim: byte): real;
var constant: real;
begin
constant := - ((dim / 2) * ln(2 * pi));
NormConstant := exp(constant);
end;

(*--------------------------------------------------------------------------------------------------*)

function NumberActors: integer;
(*Gets the number of actors.*)
var
  i: integer;
  s: shortstring;
  f: textfile;
begin
s := CurrentProjectName + '.mo';
ReadFile(f,s);
i := 0;
repeat
i := i + 1;
readln(f);
until (i = 7);
read(f,i);
closefile(f);
NumberActors := i;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure DefaultRE(var actors: integer; var dim: byte; var re: rrarray);
(*Set default values of random effects.*)
var i, h: integer;
begin
for i := 1 to actors do
  begin
  for h := 1 to dim do re[i,h] := 0;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InitialSigma(const dim: byte; var matrix1: pmatrix; const real1: real; var ascii: TFileStream);
(*Initial, crude estimate of variance-covariance matrix of random effects.*)
var s: shortstring;
x: string;
begin
s := 'CovRE.in';
if FileExists(s) then ReadCov(s,dim,matrix1) (*Initial variance-covariance matrix specified by user.*)
else SetDiagMatrix(dim,matrix1,real1); (*Default: diagonal matrix with value <real1> on main diagonal.*)
x := 'Initial variance-covariance matrix of actor-dependent random effects:'+eol;
ascii.Write(pchar(x)^,length(x));
MatrixWriteLight(ascii,dim,dim,matrix1);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InitialRE(const actors: integer; const dim: byte; const cholfac: pmatrix; var re: rrarray);
(*Initial values of the random effects: sample from multivariate normal distribution.*)
var
  i: integer;
  scale, densities: R_Ar;
  cholfactor: pmatrix;
begin
SetArray(scale,actors);
for i := 1 to actors do scale[i] := 1;
SampleMVN(densities,actors,dim,cholfac,scale,re);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SaveRE(const actors: integer; const dim: byte; const matrix3: rrarray);
(*Write the ergodic average of the actor-dependent random effects to be used in future runs as starting point.*)
var
  i, j: integer;
  s: shortstring;
  f: textfile;
begin
s := 'RE.in';
OpenFile(f,s);
i := 11;
j := 6;
MatrixWrite(actors,dim,f,matrix3,i,j); (*Write average of actor-dependent random effects to file to be used in future runs.*)
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ReadCov(const filename: shortstring; const dim: byte; var matrix1: pmatrix);
(*Exploited in Monte Carlo simulations: read true variance-covariance matrix of the random effects.*)
var
  i, j: byte;
  u: real;
  file1: textfile;
begin
ReadFile(file1,filename); (*Open and reset file.*)
for i := 1 to dim do
  begin
  for j := 1 to i do
    begin
    read(file1,u);
    matrix1[i,j] := u;
    matrix1[j,i] := matrix1[i,j]; (*Complete the symmetric matrix.*)
    end;
  readln(file1);
  end;
closefile(file1);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure WriteInitialEstimateFE(var dim: byte; var theta: pvec);
(*Write initial estimates of the fixed effects, obtained by unconditional moment estimation, to file.*)
var
  i: byte;
  s: shortstring;
  f: textfile;
begin
s := 'FEE.dat';
OpenFile(f,s);
for i := 1 to dim do
  begin
  if (i > 1) then writeln(f);
  write(f,theta[i]:10:10);
  end;
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ReadInitialEstimateFE(var theta: pvec);
(*Read initial estimates of the fixed effects, obtained by unconditional moment estimation, from file.*)
var
  i: byte;
  u: real;
  s: shortstring;
  f: textfile;
begin
s := 'FEE.dat';
if FileExists(s) then
  begin
  ReadFile(f,s);
  i := 0;
  while not(eof(f)) do
    begin
    Inc(i);
    readln(f,u);
    theta[i] := u;
    end;
  closefile(f);
  end;
(*Otherwise: input from <mo> file used.*)
end;

{
Can be called by using the following code:
(*Redundant: admits to input initial parameter value as an alternative to <.mo> file.*)
for i := 1 to pmax do initial_theta[i] := 0; (*If <lprob> has been initialised already, then it was initialised based on <theta>, which must be stored because <theta> will be updated and hence <lprob> must be updated.*)
for i := 1 to p do initial_theta[i] := theta[i];
ReadInitialEstimateFE(theta); (*Initial estimates of fixed effects, obtained by unconditional moment estimation.*)
SetAlpha(outf,theta); (*Sets <alpa> to the new parameter vector.*)
EvaluateMultinomial(1,initial_theta,theta,re,re,mnlactor1,mnlactor2,outf,lf); (*Joint multinomial probability mass, evaluated (!) given the present and (2) given the proposed random effects.*)
UpdateLProbs(true); (*Update the change probabilities <lprob>.*)
}

(*--------------------------------------------------------------------------------------------------*)

procedure WriteCovTheta(var boolean1, boolean2: boolean; var dim1: byte; var ldim: integer; var omikron, sigma: pmatrix);
(*Writes variance-covariance matrix of parameter estimates to file, as estimated by the Delta method.*)
var i, j, k:  byte;
begin
if not(boolean1 or boolean2) then
  begin
  omikron := sigma; (*Variance-covariance matrix of parameter estimates, estimated by the Delta method.*)
  for j := 2 to dim1 do (*Necessary because below diagonal elements are correlations in the original matrix.*)
    begin
    for k := 1 to (j - 1) do omikron[j,k] := omikron[k,j]; (*Necessary because below diagonal elements are correlations in the original matrix.*)
    end;
  k := ldim; (*Computes the number of parameter coordinates of the rate function parameter.*)
  for i := 1 to (dim1 - k)  do (*Discard the elements corresponding to the (vector valued) rate parameter.*)
    begin
    for j := 1 to i do
      begin
      omikron[i,j] := omikron[i+k,j+k];
      omikron[j,i] := omikron[i,j];
      end;
    end;
  k := dim1 - k;
  WriteInitialCovTheta(k,omikron); (*Write initial estimates of variance-covariance matrix of the parameter vector, obtained by the Delta method, to file.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure WriteInitialCovTheta(var dim: byte; var sigma: pmatrix);
(*Write initial estimates of variance-covariance matrix of the parameter vector, obtained by the Delta method, to file.*)
var
  i, j: byte;
  s: shortstring;
  f: textfile;
begin
s := 'CovTheta.in';
OpenFile(f,s);
for i := 1 to dim do
  begin
  if (i > 1) then writeln(f);
  for j := 1 to dim do write(f,' ',sigma[i,j]:10:10);
  end;
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ReadCovTheta(const dim: byte; var sigma: pmatrix; var ascii: TFileStream);
(*Reads variance-covariance matrix of candidate-generating (multivariate) Gaussian density of the parameter vector from file.*)
var
  iterations: integer;
  i, j: byte;
  u: real;
  user_input, violation: boolean;
  s: shortstring;
  f: textfile;
  eigen_values: pvec;
  eigen_vectors, sigma_copy: pmatrix;
  x: string;
begin
user_input := false;
SetMatrixValue(dim,sigma,0);
s := 'CovTheta.in';
if FileExists(s) then (*If the file exists, control whether it contains a suitable matrix.*)
  begin
  ReadFile(f,s);
  i := 0;
  while not(eof(f)) do
    begin
    Inc(i);
    j := 0;
    while not(eoln(f) or (i = j)) do (*Read elements below and on the main diagonal.*)
      begin
      Inc(j);
      read(f,u);
      sigma[i,j] := u;
      end;
    readln(f);
    end;
  closefile(f);
  if (i = dim) then (*Control whether the order of the inputted matrix is correct.*)
    begin
    for i := 2 to dim do (*Symmetrize inputted matrix.*)
      begin
      for j := 1 to (i - 1) do sigma[j,i] := sigma[i,j];
      end;
    SetMatrix(dim,dim,sigma_copy,sigma);
    Jacobi(dim,sigma_copy,eigen_values,eigen_vectors,iterations); (*Compute eigen values of inputted matrix.*)
    violation := false;
    for i := 1 to dim do (*Control whether the inputted matrix is positive definite.*)
      begin
      if (eigen_values[i] <= 0) then violation := true;
      end;
    if not(violation) then user_input := true;
    end;
  end;
if not(user_input) then SetIdentityMatrix(dim,sigma); (*No user input because (1) the file fails to exist, or (2) the matrix in the file is not suitable.*)
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Variance-covariance matrix of candidate-generating (multivariate) Gaussian density of the entire parameter vector:'+eol;
ascii.Write(pchar(x)^,length(x));
MatrixWrite(ascii,dim,dim,sigma);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ReadNetwork(const input: shortstring; const actors: integer; var x: rrarray);
var
  i, j, k: integer;
  s: string;
  f: textfile;
begin
SetArray(x,actors,actors);
ReadFile(f,input);
readln(f,s);
if (s = 'Siena data file: adjacency matrix') then (*Adjacency matrix format.*)
  begin
  for i := 1 to actors do
    begin
    for j := 1 to actors do
      begin
      read(f,k);
      x[i,j] := k;
      end;
    readln(f);
    end;
  end
else (*Edge list format.*)
  begin
  while not(eof(f)) do
    begin
    read(f,i);
    read(f,j);
    readln(f,k);
    x[i,j] := k;
    end;
  end;
closefile(f);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EvaluateObservedDistances(const boolean1: boolean; const actors: integer; const observations: integer; var ascii: TFileStream);
(*Evaluate row-wise distance between two adjacency matrices on some given actor set.*)
var                                       
  h: byte;
  i, j, sum1, sum2, sum3, u: integer;
  s: shortstring;
  x1, x2: rrarray;
  x: string;
begin
if boolean1 then
  begin
  h := 0;
  while (h < (observations - 1)) do
    begin
    Inc(h);
    s := CurrentProjectName + '.d0' + inttostr(h);
    ReadNetwork(s,actors,x1);
    s := CurrentProjectName + '.d0' + inttostr(h+1);
    ReadNetwork(s,actors,x2);
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    x := format('%s%d%s',['Period ',h,':'])+eol;
    ascii.Write(pchar(x)^,length(x));
    x := '---------'+eol;
    ascii.Write(pchar(x)^,length(x));
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    x := 'Observed distance (actor-wise):'+eol;
    ascii.Write(pchar(x)^,length(x));
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    x := format('%20s%20s%20s%20s',['actor','change','change: outdegree','proportion created ties'])+eol;
    ascii.Write(pchar(x)^,length(x));
    ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
    for i := 1 to actors do
      begin
      sum1 := 0;
      sum2 := 0;
      sum3 := 0;
      for j := 1 to actors do
        begin
        u := trunc(x2[i,j] - x1[i,j]);
        sum1 := sum1 + abs(u);
        sum2 := sum2 + u;
        if (u = 1) then Inc(sum3);
        end;
      if (sum1 <> 0) then sum3 := sum3 div sum1;
      x := format('%20d%20d%20d%20d',[i,sum1,sum2,sum3])+eol;
      ascii.Write(pchar(x)^,length(x));
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure IntroduceRE(var ascii: TFileStream; var thetarandom: pbvec; var Description: ppString; var f1: textfile;
const sequences: integer; var ldim1: integer; var f2: textfile; var re, mvn: rrarray; var actors: integer; var dim: byte; var boolean1: boolean);
(*Write specification of random part to the output file.*)
var
  i: byte;
  s: shortstring;
  x : string;
begin
Heading(2,ascii,'Model specification: random part');
x := 'The following coefficients are modeled as actor-dependent random coefficients:'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to pmax do
  begin
  if thetarandom[i] then begin
    x := Description[i]+eol;
    ascii.Write(pchar(x)^,length(x));
  end;
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
if (sequences > 1) then
  begin
  s := CurrentProjectName+'CovRESample.m';
  if (ldim1 = 1) then OpenFile(f2,s)
  else AppendToFile(f2,s); (*Assign file name to textfile variable and rewrite file.*)
  end;
re := nil; (*Cleaning up...*)
mvn := nil;
Setlength(mvn,actors+1,dim+1); (*The second dimsion depends on the number of random coefficients.*)
Setlength(re,actors+1,dim+1); (*The second dimension depends on the number of fixed coefficients.*)
if boolean1 then (*Actors sets with changing composition.*)
  begin
  x := 'N O T E:  actor-dependent random coefficients can be used in the presence of'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := '- missing values'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := '- structural values'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := '- actor sets with changing composition.'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := 'But in the last case, estimation problems may arise when some actors are absent most of the time.'+eol;
  ascii.Write(pchar(x)^,length(x));
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure InitialActionsRE(var randomeffects: boolean; var ascii: TFileStream; var thetarandom: pbvec;
var Description: ppstring; var f1: textfile; const sequences: integer; var ldim1: integer;
var f2: textfile; var re, mvn: rrarray; var actors: integer; var dim: byte; var boolean1: boolean);
(*Initial actions concerning random effects.*)
var
  s: shortstring;
  f: textfile;
begin
if randomeffects then
  begin
  IntroduceRE(ascii,thetarandom,Description,f1,sequences,ldim1,f2,re,mvn,actors,dim,boolean1); (*Write specification of random part to the output file.*)
  s := 'ScaleCDRE.dat';
  assignfile(f,s);
  if FileExists(s) then erase(f);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ResetBlackBox(var Bayesian, randomeffects: boolean; var blackbox: blackarray; var scalebeta_i: R_Ar;
var dim1: byte; var scaleRE_i: R_Ar; var actors: integer; var dim2: byte);
(*Reset blackbox to default values. Must be placed before (!) <MCMC_intro>.*)
var i: integer;
begin
(*Explanation:
0: last proposal accepted: yes = 1, no = 0 (actor-dependent random effects)
1: acceptance rate (actor-dependent random effects)
2: average acceptance probability (actor-dependent random effects)
3: average augmented-data log-likelihood (up to proportionality constant): not functional at the moment,
since in <MH_RE_Step> the statements involving <remainder> have been disabled (actor-dependent random effects).
4: acceptance rate (path)
5: acceptance rate (variance-covariance matrix of actor-dependent random effects)
6: acceptance rate (fixed effects: rate parameters)
7: acceptance rate (fixed effects: objective function parameter)
...
11: scale factor (actor-dependent random effects: candidate-generating density)
12: scale factor (fixed effects: objective function parameter: candidate-generating density)
...
20: temperature (actor-dependent random effects).
*)
if (Bayesian or randomeffects) then
  begin
  for i := 1 to 10 do blackbox[i] := 0;
  blackbox[12] := 1; (*Scale factor: parameter vector.*)
  blackbox[20] := 1; (*Temperature: modifying the stationary distribution of the Markov chain.*)
  if Bayesian then
    begin
    SetLength(scalebeta_i,dim1+1);
    for i := 1 to dim1 do scalebeta_i[i] := 2.54 / sqrt(dim1); (*Coordinate-dependent scale factors corresponding to the multivariate candidate-generating density for the parameter vector M-H steps.*)
    end;
  SetLength(scaleRE_i,actors+1);
  if randomeffects then
    begin
    blackbox[11] := 2.54 / sqrt(dim2); (*Scale factor: random effects.*)
    for i := 1 to actors do scaleRE_i[i] := 2.54 / sqrt(dim2); (*Actor-dependent scale factors corresponding to the candidate-generating densities for the random effects M-H steps.*)
    end;
  end;  
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetREactor(const dim: byte; const actor: integer; var re: rrarray; const mvn: rrarray);
(*Set <re = mvn> for actor <actor>.*)
var h: byte;
begin
for h := 1 to dim do re[actor,h] := mvn[actor,h];
end;

(*--------------------------------------------------------------------------------------------------*)

procedure RE_positive(var actors: integer; var dim: byte; var re, sumposre: rrarray);
(*Evaluate coordinate-wise the indicator function I(re > 0) and increment the count.*)
var i, h: integer;
begin
for i := 1 to actors do
  begin
  for h := 1 to dim do
    begin
    if (re[i,h] > 0) then sumposre[i,h] := sumposre[i,h] + 1;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

function CR(var actors: integer; var detsigma: real; var theta: pvec; const totaltime: real; const totnumber: integer): real;
(*Evaluate the remaining terms of the augmented-data log-likelihood.
Observe that it is assumed that in the rate function there is the intercept, but no other fixed effects.*)
var u: real;
begin
u := - (actors / 2) * log(detsigma); (*Corresponding to the joint normal probability density of the random effects.*)
u := u + EvaluatePoisson(actors,theta[1],totaltime,totnumber); (*Corresponding to the Poisson probability mass of the number of transitions in the given period.*)
CR := u;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SetProposalRE(const condition: boolean; const actors: integer; const dim: byte; const re: rrarray;
var eb, lastmvn0, reflectionre: rrarray; var sumsigma: pmatrix; const ldim: integer);
(*Prepare some things to be exploited by the candidate-generating density of the actor-dependent random effects.*)
begin
if condition then (*Initialise: Phase 1 or Phase 3 (in case Phase 1, 2 are skipped).*)
  begin
  SetArray(lastmvn0,actors,dim);
  SetArray(reflectionre,actors,dim); (*First create <reflectionre>...*)
  if (cdensity2 < 3) then SetMatrix(actors,dim,reflectionre,re); (*...then set <reflectionre = re>.*)
  end
else (*If Phase 1 and Phase 2 are not skipped: start of Phase 2 or Phase 3.*)
  begin
  ScalarDiv(actors,dim,eb,ldim); (*First average <eb>...*)
  if (cdensity2 < 3) then SetMatrix(actors,dim,reflectionre,eb); (*...then set <reflectionre = eb>.*)
  end;
SetArray(eb,actors,dim); (*(Re)set (empirical) Bayes estimates of actor-dependent random effects.*)
SetMatrixValue(dim,sumsigma,0); (*(Re)set sum of variance-covariance matrices of actor-dependent random effects.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CandidateRE(const actors: integer; const dim: byte; const cholfac: pmatrix;
const scaleRE_i: R_Ar; var mvn: rrarray; const re: rrarray);
(*Generate candidates for actor-dependent random effects.*)
var
  mvndensities: R_Ar;
  mvn0: rrarray;
begin
(*Remark:
- <cdensity2> refers to the candidate-generating density.
- concerning <reflectionre>: in case <cdensity2 = 2>, the Markov chain is reflected about <reflectionre>.*)
CandidateMH(mvn0,actors,dim,mvndensities,cholfac,scaleRE_i,cdensity2,mvn,re,reflectionre);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MH_RE_Step(const actors: integer; const dim: byte; const isigma: pmatrix; const multi_block: boolean;
var re, mvn: rrarray; var propmn, presmn: R_Ar);
(*Metropolis-Hastings step to sample actor-dependent random effects.
Given: candidate <mvn> sampled from
(1) random walk density.
(2) autoregressive density.
(3) independence sampler.
To be carried out:
- multi-block Metropolis-Hastings step.
- one-block Metropolis-Hastings step.
*)
var
  i: integer;
  mnratio, mvnratio, pr: real;
  accept: boolean;
  mvndensities, redensities, propMVN, presMVN, sum: R_Ar;
begin
(*-------------------------*)
(*Evaluate Gaussian density*)
(*-------------------------*)
EvaluateMVN(redensities,actors,dim,re,isigma); (*Logarithm of Gaussian kernel: present random effects.*)
EvaluateMVN(mvndensities,actors,dim,mvn,isigma); (*Logarithm of Gaussian kernel: proposed random effects.*)
(*--------------------------------------*)
(*Evaluate candidate-generating density*)
(*-------------------------------------*)
(*Remark: if <cdensity2 = 1, 2>, the ratio of candidate-generating densities vanishes.*)
if (cdensity2 = 3) then (*Independence sampler.*)
  begin
  EvaluateFunctionMVN2(presMVN,actors,dim,re,isigma,scaleRE_i); (*Logarithm of Gaussian kernel, rescaled: present random effects.*)
  for i := 1 to actors do redensities[i] := redensities[i] - presMVN[i]; (*M-H acceptance ratio (independence sampler): <f(mvn) / f(re)>, where <f(re) = p(re) / q(re)>.*)
  EvaluateFunctionMVN2(propMVN,actors,dim,mvn,isigma,scaleRE_i); (*Logarithm of Gaussian kernel, rescaled: proposed random effects.*)
  for i := 1 to actors do mvndensities[i] := mvndensities[i] - propMVN[i]; (*M-H acceptance ratio (independence sampler): <f(mvn) / f(re)>, where <f(mvn) = p(mvn) / q(mvn)>.*)
  presMVN := nil;
  propMVN := nil;
  end;
(*------------------------*)
(*Metropolis-Hastings step*)
(*------------------------*)
if multi_block then (*Multi-block Metropolis-Hastings.*)
  begin
  SetArray(sum,2);
  SetArray(accepted,actors);
  for i := 1 to actors do (*Cycle algorithm.*)
    begin
    mnratio := propmn[i] - presmn[i]; (*Log ratio of multinomial choice probabilities.*)
    mvnratio := mvndensities[i] - redensities[i]; (*Log ratio of Gaussian densities.*)
    accept := DecisionMH(pr,mnratio,mvnratio,blackbox); (*Decide whether or not the random effects candidate is to be accepted.*)
    if accept then (*Candidate accepted.*)
      begin
      accepted[i] := true;
      sum[1] := sum[1] + 1; (*Acceptance rate.*)
      sum[2] := sum[2] + pr; (*Average acceptance probability.*)
      presmn[i] := propmn[i]; (*Update multinomial choice probability.*)
      SetREactor(dim,i,re,mvn); (*Set <re = mvn> for actor <i>.*)
      end;
    end;
  blackbox[0] := sum[1] / actors; (*Acceptance rate.*)
  blackbox[1] := blackbox[1] + (sum[1] / actors); (*Acceptance rate.*)
  blackbox[2] := blackbox[2] + (sum[2] / actors); (*Average acceptance probability.*)
  sum := nil;
  end
else (*one-block Metropolis-Hastings.*)
  begin
  mnratio := propmn[0] - presmn[0]; (*Log ratio of joint multinomial choice probabilities.*)
  mvnratio := mvndensities[0] - redensities[0]; (*Log ratio of joint Gaussian densities.*)
  accept := DecisionMH(pr,mnratio,mvnratio,blackbox); (*Decide whether or not the random effects candidate is to be accepted.*)
  if accept then
    begin
    blackbox[0] := 1; (*Candidate accepted.*)
    blackbox[1] := blackbox[1] + 1; (*Acceptance rate.*)
    blackbox[2] := blackbox[2] + pr; (*Average acceptance probability.*)
    presmn[0] := propmn[0]; (*Update joint multinomial choice probability.*)
    SetMatrix(actors,dim,re,mvn); (*Set <re = mvn>.*)
    end
  else blackbox[0] := 0; (*Candidate not accepted.*)
  end;
redensities := nil;
mvndensities := nil;
end;

(*--------------------------------------------------------------------------------------------------*)

function DetScaleSigma(var dim: byte; var sigma: pmatrix; var dim1: byte): real;
(*Evaluates some function of the scale matrix of the prior probability density (Wishart) corresponding to the variance-covariance matrix of the random effects.*)
var
  u: real;
  cholfactor: pmatrix;
begin
Cholesky(dim,cholfactor,sigma,u);
u := - (dim1 / 2) * ln(u);
DetScaleSigma := exp(u);
end;

(*--------------------------------------------------------------------------------------------------*)

function EvaluateESCovRE(var dim1, dim2: byte): real;
(*Evaluate the Exponent of the Sample Covariance matrix of the random effects for the Wishart density.*)
var u: real;
begin
u := dim1 - dim2 - 1;
EvaluateESCovRE := u / 2;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EvaluateEB(var dim: byte; var ldim1: integer; var matrix3: rrarray; var ldim2: integer; var matrix4: rrarray);
(*Evaluate Empirical Bayes mean and variance of the actor-dependent random effects.*)
var
  i: integer;
  h: byte;
  sum: real;
begin
for h := 1 to dim do
  begin
  sum := 0;
  for i := 1 to ldim1 do
    begin
    matrix3[i,h] := matrix3[i,h] / ldim2;
    sum := sum + matrix3[i,h];
    end;
  matrix3[h,0] := sum / ldim1;
  end;
for i := 1 to ldim1 do
  begin
  for h := 1 to dim do matrix4[i,h] := matrix3[i,h];
  end;
{
setlength(sums,ldim1+1);
s := CurrentProjectName + 'RE.m';
ReadFile(f,s);
for l := 1 to ldim2 do
  begin
  for h := 1 to dim do
    begin
    for i := 1 to ldim1 do
      begin
      read(f,u)
      sums[i] := sqr(u - matrix4[i,h]);
      end;
    end;
  readln(f);
  end;
for l := 1 to ldim do
  begin
  end;
closefile(f);
}
end;

(*--------------------------------------------------------------------------------------------------*)

function DivergingRE(var actors: integer; var dim: byte; var re: rrarray; var threshold: real; var ascii: textfile; var sigma: pmatrix): boolean;
(*Checks whether some actor-dependent random effects are diverging to infinity.*)
var
  i, j: integer;
  h: byte;
  condition: boolean;
begin
condition := false;
DivergingRE := false;
for i := 1 to actors do
  begin
  for h := 1 to dim do
    begin
    if (abs(re[i,h]) > threshold) then
      begin
      condition := true;
      DivergingRE := true;
      writeln(ascii,'W A R N I N G: some random effects seem to diverge (threshold value: ',threshold:4:4,')');
      writeln(ascii);
      writeln(ascii,'...maybe too little change in the target statistic...');
      writeln(ascii);
      end;
    end;
  end;
if condition then (*Rescale proposal distribution.*)
  begin
  for i := 1 to dim do
    begin
    for j := 1 to dim do
      begin
      if (sigma[i,j] > threshold) then
        begin
        if (i = j) then sigma[i,j] := 1 (*Variance.*)
        else sigma[i,j] := 0; (*Covariance.*)
        end;
      end;
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EstimateCovI(var dim: byte; var actors: integer; var re: rrarray; var sigma: pmatrix);
(*Unbiased estimate of variance-covariance matrix.*)
const threshold = 1e-10;
var
  i, j, h: integer;
  sum: real;
begin
(*Matrix procedures, such as in <S_matrices>, could be used equally well, but involve more operations, due to transposing <Sigma>.*)
for i := 1 to dim do
  begin
  for j := 1 to i do (*Lower-diagonal triangle half of variance-covariance matrix plus diagonal.*)
    begin
    sum := 0;
    for h := 1 to actors do
      begin
      sum := sum + (re[h,i] * re[h,j]); (*<re> stores the last accepted proposal = present random effects.*)
      end;
    sigma[i,j] := sum / (actors - 1); (*The m.l.e. is unbiased (plugging-in the m.l.e. of the mean would indeed lead to bias).*)
    if (abs(sigma[i,j]) < tin) then sigma[i,j] := 0;
    sigma[j,i] := sigma[i,j];
    end;
  if (sigma[i,i] < threshold) then sigma[i,i] := threshold; (*To prevent singularity.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EstimateCovII(var dim: byte; var actors: integer; var re: rrarray; var sigma: pmatrix);
(*Maximum likelihood estimate of variance-covariance matrix: unbiased if mean vector is not estimated, otherwise biased.*)
const threshold = 1e-10;
var
  i, j, h: integer;
  sum: real;
begin
(*Matrix procedures, such as in <S_matrices>, could be used equally well, but involve more operations, due to transposing <Sigma>.*)
for i := 1 to dim do
  begin
  for j := 1 to i do (*Lower-diagonal triangle half of variance-covariance matrix plus diagonal.*)
    begin
    sum := 0;
    for h := 1 to actors do
      begin
      sum := sum + (re[h,i] * re[h,j]); (*<re> stores the last accepted proposal = present random effects.*)
      end;
    sigma[i,j] := sum / actors; (*The m.l.e. is unbiased (plugging-in the m.l.e. of the mean would indeed lead to bias).*)
    if (abs(sigma[i,j]) < tin) then sigma[i,j] := 0;
    sigma[j,i] := sigma[i,j];
    end;
  if (sigma[i,i] < threshold) then sigma[i,i] := threshold; (*To prevent singularity.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure EstimateCovIII(var dim: byte; var actors: integer; var re: rrarray; var sigma: pmatrix);
(*Compute <sum re * transpose(re)>.*)
const threshold = 1e-10;
var
  i, j, h: integer;
  sum: real;
begin
(*Matrix procedures, such as in <S_matrices>, could be used equally well, but involve more operations, due to transposing <Sigma>.*)
for i := 1 to dim do
  begin
  for j := 1 to i do (*Lower-diagonal triangle half of variance-covariance matrix plus diagonal.*)
    begin
    sum := 0;
    for h := 1 to actors do
      begin
      sum := sum + (re[h,i] * re[h,j]); (*<re> stores the last accepted proposal = present random effects.*)
      end;
    sigma[i,j] := sum;
    if (abs(sigma[i,j]) < tin) then sigma[i,j] := 0;
    sigma[j,i] := sigma[i,j];
    end;
  if (sigma[i,i] < threshold) then sigma[i,i] := threshold; (*To prevent singularity.*)
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Get_Ksi_Li_Cholesky(const dim: byte; const c_cov: pmatrix; var c_icov: pmatrix; var li, ksi: rrarray);
(*Obtain, from the Cholesky factor <c_cov> of the variance-covariance matrix,
- the Cholesky factor <c_icov> of the inverse variance-covariance matrix
- and the matrices <li> and <ksi>.*)
var
  i, j: byte;
  a: pmatrix;
begin
(*Remark: <inverse(Sigma) = inverse(AA') = inverse(A') * inverse(A) = inverse(A)' * inverse(A).*)
CholeskyI(dim,c_cov,a); (*Inverse of Cholesky factor: <a = inverse(c_cov)>.*)
MatrixT(dim,c_icov,a); (*Evaluate <c_icov = transpose(a) = transpose(inverse(c_cov))>.*)
SetArray(li,dim); (*<li> is, by the nil operator, null.*)
SetArray(ksi,dim); (*<ksi> is, by the nil operator, null.*)
(*Evaluate <ksi>:
<inverse(Sigma) = c_icov * c_icov' = li * sqrt(exp(ksi)) * sqrt(exp(ksi)) * li' = li * exp(ksi / 2) * exp(ksi / 2) * li'>;
thus, to obtain <ksi>, take the logarithm of <c_icov> and scale up by factor <2>.*)
for i := 1 to dim do ksi[i,i] := 2 * ln(c_icov[i,i]);
(*Evaluate <li>:*)
for i := 1 to dim do li[i,i] := 1; (*Diagonal elements.*)
for i := 2 to dim do (*Below-diagonal elements; the above-diagonal elements vanish.*)
  begin
  for j := 1 to (i - 1) do
    begin
    li[i,j] := c_icov[i,j] / c_icov[j,j]; (*Incl. the diagonal element, giving unit values on the main diagonal.*)
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Get_Ksi_Li_Cov(const dim: byte; var c_cov: pmatrix; const sigma: pmatrix;
var li, ksi: rrarray; var count: byte; var theta: pvec);
(*Obtain, from the variance-covariance matrix,
- the Cholesky factor
- the vectors <v(Li)> and <v(Ksi)>, and insert them into the parameter vector <theta>.*)
var
  i, j: byte;
  det: real;
  c_icov: pmatrix;
begin
{-------------------}
{-Evaluate matrices-}
{-------------------}
Cholesky(dim,c_cov,sigma,det); (*Evaluate the Cholesky factor of the variance-covariance matrix.*)
Get_Ksi_Li_Cholesky(dim,c_cov,c_icov,li,ksi); (*Obtain <Li> and <Ksi> satisfying <inverse(Sigma) = Li * exp(Ksi) * transpose(Li)>.*)
{----------------------------------------------------}
{-Stack non-redundant elements into parameter vector-}
{----------------------------------------------------}
(*Elements below the main diagonal of  <li>; other elements are restricted due to linear constraints.*)
for j := 1 to dim do (*For each column do...*)
  begin
  for i := (j + 1) to dim do (*...start with the first below-diagonal element.*)
    begin
    Inc(count); (*The initial value of <count> must be inputted.*)
    theta[count] := li[i,j]; (*That is, <li[j+1,j]> up to, and including, <li[dim,j]>.*)
    end;
  end;
for i := 1 to dim do (*Elements on the main diagonal of <ksi>.*)
  begin
  Inc(count);
  theta[count] := ksi[i,i];
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure Get_Cov_Theta(const dim_re, dim_theta: byte; const theta: pvec; var li, ksi: rrarray; var c_icov, c_cov, i_sigma, sigma: pmatrix);
(*Obtain, from the parameter vector, the Cholesky factor of the variance-covariance matrix of the actor-dependent random effects,
and the variance-covariance matrix itself.*)
var
  i, j, h, dimension: byte;
  t_c_icov: pmatrix;
  sqrt_lambda: rrarray;
begin
{----------------------------------------}
{-Extract matrices from parameter vector-}
{----------------------------------------}
SetArray(li,dim_re); (*<li> is, by the nil operator, null.*)
SetArray(ksi,dim_re); (*<ksi> is, by the nil operator, null.*)
SetArray(sqrt_lambda,dim_re); (*<sqrt_lambda> is, by the nil operator, null; note: as long as <sqrt_lambda> is only used in <Matrix_AdiagB_Prod>, resetting is not strictly necessary.*)
dimension := (dim_re * (dim_re + 1)) div 2; (*<sqr(dim) - dim) / 2> elements below the diagonal of <li> plus <dim> diagonal elements gives <sqr(dim) + dim) / 2>.*)
h := dim_theta - dimension; (*The conventional parameter coordinates are stored up to position <h>; positions <h + 1> up to <dim_theta> are (co)variance parameters.*)
for j := 1 to dim_re do (*For each column do...*)
  begin
  for i := (j + 1) to dim_re do (*...start with first below-diagonal element.*)
    begin
    Inc(h);
    li[i,j] := theta[h]; (*That is, <li[j+1,j]> up to, and including, <li[dim,j]>.*)
    end;
  end;
for i := 1 to dim_re do li[i,i] := 1; (*Diagonal elements: restricted to unity.*)
(*Remark: <inverse(Sigma) = c_icov * c_icov' = li * sqrt(exp(ksi)) * sqrt(exp(ksi)) * li' = li * exp(ksi / 2) * exp(ksi / 2) * li'>.*)
for i := 1 to dim_re do
  begin
  Inc(h);
  ksi[i,i] := theta[h];
  sqrt_lambda[i,i] := expo(ksi[i,i] / 2); (*Corresponds to <sqrt(Gamma)> of the random effects paper.*)
  end;
{-------------------------------------------------------------}
{-Evaluate variance-covariance matrix and its Cholesky factor-}
{-------------------------------------------------------------}
(*Remark: <Sigma = inverse(AA') = inverse(A') * inverse(A)>, where <A = c_icov> is the Cholesky factor of the inverse variance-covariance matrix.*)
Matrix_AdiagB_Prod(dim_re,c_icov,li,sqrt_lambda); (*Evaluate the Cholesky factor of the inverse variance-covariance matrix: <c_icov = li * sqrt_lambda = li * exp(ksi / 2)>.*)
MatrixT(dim_re,t_c_icov,c_icov); (*Transpose the Cholesky factor of the inverse variance-covariance matrix: <t_c_icov = transpose(c_icov)>.*)
CholeskyI(dim_re,t_c_icov,c_cov); (*Evaluate the Cholesky factor of the variance-covariance matrix: <c_cov = inverse(t_c_icov) = inverse(transpose(c_icov))>: needed to sample random effects from multivariate Gaussians.*)
MatrixTProductI(dim_re,c_icov,i_sigma); (*Evaluate the inverse variance-covariance matrix <i_sigma = c_icov * transpose(c_icov)>.*)
MatrixTProductI(dim_re,c_cov,sigma); (*Evaluate the variance-covariance matrix <sigma = c_cov * transpose(c_cov)>: needed for convergence diagnostics.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure WriteScaleCDRE(var ascii: TFileStream; var actors: integer; var scaleRE_i: R_Ar);
(*Writes the actor-dependent scale factors corresponding to the candidate-generating density for the random effects M-H steps to file.*)
var i: byte;
x: string;
begin
x := 'Random effects: variance-covariance matrix of candidate-generating (multivariate) Gaussian density to be rescaled by factor '+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to actors do begin
  x := format('%s%4.10f',[' ',scaleRE_i[i]]);
  ascii.Write(pchar(x)^,length(x));
  end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure WriteScaleCDtheta(var ascii: TFileStream; var scale: real);
(*Writes the scale factor corresponding to the multivariate candidate-generating density for the parameter vector M-H steps to log-file.*)
var x: string;
begin
x := 'Objective function parameter: variance-covariance matrix of candidate-generating (multivariate) Gaussian density to be rescaled by factor '+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := format('%s%4.10f',[' ',scale]);
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
end;

(*--------------------------------------------------------------------------------------------------*)

procedure DistributionCovRE(var randomeffects: boolean; const sequences: integer; var dim: byte; var f2: textfile;
var matrix1: pmatrix; var ldim1: integer);
(*Completes Matlab source file to plot sampling distribution of the estimate of the variance-covariance
matrix of the random effects. Needs to be placed after the <for ldim1 := 1 to sequences> loop and before the
<ConvergenceCovRE> procedure, where the file is closed.*)
begin
if randomeffects then
  begin
  if (sequences > 1) then
    begin
    MatrixWriteCov(dim,f2,matrix1); (*Sampling distribution of estimate of variance-covariance matrix of random effects written to file.*)
    if (ldim1 = sequences) then writeln(f2,'#Sampling distribution of covariance matrix of random effects.')
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure ConvergenceCovRE(var randomeffects: boolean; var f1: textfile);
(*Complete Matlab source files containing trace plot of variance-covariance matrix of the random effects.*)
begin
if randomeffects then
  begin
  writeln(f1,'plot(x)');
  writeln(f1,'title(',char(39),'Trace plot of variance-covariance matrix of random effects',char(39),')');
  writeln(f1,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(f1,'ylabel(',char(39),'value of (co)variance(s)',char(39),')');
  closefile(f1); 
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteThetaFile(var f1: textfile; var boolean1: boolean; var f2: textfile);
(*Complete Matlab source file to generate trace plot of fixed effects (parameter) estimates, and acceptance rates.*)
begin
MatlabPlot(f1); (*Adds some convenient Matlab code that enables columnwise plots.*)
writeln(f1,'title(',char(39),'Trace plot of parameter estimates (rate and non-rate parameter)',char(39),')');
writeln(f1,'xlabel(',char(39),'iteration number',char(39),')');
writeln(f1,'ylabel(',char(39),'estimates',char(39),')');
closefile(f1);
if boolean1 then
  begin
  MatlabPlot(f2); (*Adds some convenient Matlab code that enables columnwise plots.*)
  writeln(f2,'title(',char(39),'Trace plot of parameter acceptance rates (rate and non-rate parameter)',char(39),')');
  writeln(f2,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(f2,'ylabel(',char(39),'acceptance rates',char(39),')');
  closefile(f2);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteADLLfile(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of augmented-data likelihood (up to proportionality constant).*)
begin
if boolean1 then
  begin
  writeln(ascii,']');
  writeln(ascii,'plot(x)');
  MatlabPlotExtra(ascii); (*Some Matlab code.*)
  writeln(ascii,'title(',char(39),'Trace plot of augmented-data log-likelihood (up to proportionality constant)',char(39),')');
  writeln(ascii,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(ascii,'ylabel(',char(39),'value of augmented-data log-likelihood',char(39),')');
  closefile(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteARPATHFilee(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of acceptance rates of the Metropolis-Hastings
sampler to sample random effects.*)
begin
if boolean1 then
  begin
  writeln(ascii,']');
  writeln(ascii,'plot(x)');
  writeln(ascii,'title(',char(39),'Trace plot of acceptance rates of the M-H algorithm to sample paths',char(39),')');
  writeln(ascii,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(ascii,'ylabel(',char(39),'value of acceptance rate',char(39),')');
  closefile(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteARREFile(const boolean1, boolean2: boolean; var f1, f2: textfile);
(*Complete Matlab source file to generate trace plot of acceptance rates of the Metropolis-Hastings
sampler to sample random effects and the variance-covariance matrix of the random effects.*)
begin
if boolean1 then (*Random effects models.*)
  begin
  writeln(f1,']');
  writeln(f1,'plot(x)');
  writeln(f1,'title(',char(39),'Trace plot of acceptance rates of the M-H algorithm to sample random effects',char(39),')');
  writeln(f1,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(f1,'ylabel(',char(39),'value of acceptance rate',char(39),')');
  closefile(f1);
  if boolean2 then (*Bayesian random effects models.*)
    begin
    writeln(f2,']');
    writeln(f2,'plot(x)');
    writeln(f2,'title(',char(39),'Trace plot of acceptance rates of the M-H algorithm to sample variance-covariance matrices',char(39),')');
    writeln(f2,'xlabel(',char(39),'iteration number',char(39),')');
    writeln(f2,'ylabel(',char(39),'value of acceptance rate',char(39),')');
    closefile(f2);
    end;
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteATheta(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of MCMC sample lag <lag> autocorrelation of rate and non-rate parameter vector.*)
begin
if boolean1 then
  begin
  MatlabPlot(ascii); (*Adds some convenient Matlab code that enables columnwise plots.*)
  writeln(ascii,'title(',char(39),'Trace plot of MCMC sample lag 1 autocorrelation: rate and non-rate parameter vector (acceptance rate, see file)',char(39),')');
  writeln(ascii,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(ascii,'ylabel(',char(39),'value of MCMC sample lag 1 autocorrelations',char(39),')');
  closefile(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteAREFile(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of MCMC sample lag <lag> autocorrelation among actor-dependent random effects.*)
begin
if boolean1 then
  begin
  MatlabPlot(ascii); (*Adds some convenient Matlab code that enables columnwise plots.*)
  writeln(ascii,'title(',char(39),'Trace plot of MCMC sample lag 1 autocorrelation: random effects (acceptance rate, see file)',char(39),')');
  writeln(ascii,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(ascii,'ylabel(',char(39),'value of MCMC sample lag 1 autocorrelation',char(39),')');
  closefile(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteACovREfile(var boolean1: boolean; var ascii: textfile);
(*Complete Matlab source file to generate trace plot of MCMC sample lag <lag> autocorrelation: variance-covariance
matrix of the actor-dependent random effects.*)
begin
if boolean1 then
  begin
  MatlabPlot(ascii); (*Adds some convenient Matlab code that enables columnwise plots.*)
  writeln(ascii,'title(',char(39),'Trace plot of MCMC sample lag 1 autocorrelation: variance-covariance matrix of random effects (acceptance rate, see file)',char(39),')');
  writeln(ascii,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(ascii,'ylabel(',char(39),'value of MCMC sample lag 1 autocorrelation',char(39),')');
  closefile(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CompleteREFile(var boolean2: boolean; var f1, f2: textfile);
(*Complete Matlab source file to generate trace plot of actor-dependent random effects.*)
begin
if boolean2 then
  begin
  MatlabPlot(f1); (*Adds some convenient Matlab code that enables columnwise plots.*)
  writeln(f1,'title(',char(39),'Trace plot of actor-dependent random effect(s)',char(39),')');
  writeln(f1,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(f1,'ylabel(',char(39),'value of random effect(s)',char(39),')');
  closefile(f1);
  MatlabPlot(f2); (*Adds some convenient Matlab code that enables columnwise plots.*)
  writeln(f2,'title(',char(39),'Trace plot of smallest latent root of variance-covariance matrix of actor-dependent random effect(s)',char(39),')');
  writeln(f2,'xlabel(',char(39),'iteration number',char(39),')');
  writeln(f2,'ylabel(',char(39),'value of smallest latent root',char(39),')');
  closefile(f2);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure OutputCovRE(var ascii: TFileStream; const dim1, dim2, covariance_parameters: byte; var sigma, matrix1, matrix2: pmatrix;
const Bayesian: Boolean; const Description: ppstring);
(*Output variance-covariance matrix of random effects.*)
var
  i, j, h: integer;
  se, variances: pmatrix;
  x: string;
begin
x := 'Variance-covariance matrix of actor-dependent random effects';
ascii.Write(pchar(x)^,length(x));
if (dim1 = 1) then ascii.write(pchar(string(':'))^,1)
else
  begin
  x := ' (correlations below diagonal):';
  ascii.Write(pchar(x)^,length(x));
  CorrelationMatrix(dim1,sigma); (*Transforms below-diagonal elements of variance-covariance matrix into correlations.*)
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
x := 'Order of rows and columns:'+eol;
ascii.Write(pchar(x)^,length(x));
h := 0;
for i := 1 to dim2 do
  begin
  if thetarandom[i] then
    begin
    Inc(h);
    x := format('%s%d%s%s',['(',h,')',Description[i]])+eol;
    ascii.Write(pchar(x)^,length(x));
    end;
  end;
MatrixWrite(ascii,dim1,dim1,sigma,12,6);
if Bayesian then (*Bayesian estimation.*)
  begin
  x := 'Monte Carlo standard errors of (co)variance parameters:'+eol;
  ascii.Write(pchar(x)^,length(x));
  MatrixWriteTriangle(dim1,ascii,matrix1,12,6);
  end
else (*ML estimation.*)
  begin
  x := 'Standard errors of (co)variance parameters:'+eol;
  ascii.Write(pchar(x)^,length(x));
  h := dim2 - covariance_parameters; (*Start with diagonal element <dim2-covariance_parameters+1> of <matrix2>.*)
  for i := 1 to dim1 do (*Column <i>.*)
    begin
    for j := i to dim1 do (*Non-redundant elements of row <j>.*)
      begin
      (*Remarks:
      - the diagonal elements <dim2-covariance_parameters, ..., dim2> of <matrix2> contain the variances of the (co)variance parameters.
      - the ordering of <matrix2> follows from <v(sigma)>:
      the first column comes first, (1,1), (2,1), ..., (n,1), then follow the non-redundant elements of the second column, (2,2), (3,2), ..., (n, 2), and so on.*)
      Inc(h);
      variances[j,i] := matrix2[h,h];
      if (i <> j) then variances[i,j] := variances[j,i]; (*Symmetrize.*)
      end;
    end;
  for i := 1 to dim1 do
    begin
    for j := 1 to dim1 do se[i,j] := sqrt(variances[i,j]);
    end;
  MatrixWriteTriangle(dim1,ascii,se,12,6);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure SaveOutput(const maxlike, Bayesian: boolean; const dim1: byte; const cov1: pmatrix; const dim2: byte; const cov2: pmatrix);
(*Save output that may be used in future runs to tune MCMC algorithms.*)
var
  i, j: byte;
  s: shortstring;
  f: textfile;
  covariance: pmatrix;
begin
if (maxlike or Bayesian) then (*Save variance-covariance matrix of point estimates.*)
  begin
  for i := 1 to dim1 do covariance[i,i] := cov1[i,i];
  for i := 2 to dim1 do (*Precautionary measure: the lower-triangle half may contain correlations: overwrite them.*)
    begin
    for j := 1 to (i - 1) do
      begin
      covariance[i,j] := cov1[j,i];
      covariance[j,i] := covariance[i,j];
      end;
    end;
  s := 'CovTheta.in';
  OpenFile(f,s);
  MatrixWrite(f,dim1,dim1,covariance);
  closefile(f);
  end;
if randomeffects then (*Save variance-covariance matrix of actor-dependent random effects.*)
  begin
  if simulate then exit; (*In simulations, do not overwrite the data-generating variance-covariance matrix.*)
  for i := 1 to dim2 do covariance[i,i] := cov2[i,i];
  for i := 2 to dim2 do (*Precautionary measure: the lower-triangle half may contain correlations: overwrite them.*)
    begin
    for j := 1 to (i - 1) do
      begin
      covariance[i,j] := cov2[j,i];
      covariance[j,i] := covariance[i,j];
      end;
    end;
  s := 'CovRE.in';
  OpenFile(f,s);
  MatrixWriteLight(f,dim2,dim2,covariance);
  closefile(f);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure PrepareEB(var boolean1: boolean; var rrarray1: rrarray; var ldim: integer; var dim: byte);
begin
if boolean1 then
  begin
  rrarray1 := nil;
  Setlength(rrarray1,ldim+1,dim+1);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure MCMC_Storage(var dim1: byte; var vector1, vector2: pvec;
var randomeffects: boolean; var dim2: byte; var matrix1: pmatrix);
(*Store MCMC sample to compute average value.*)
var i, j, h: byte;
begin
for i := 1 to dim1 do vector1[i] := vector1[i] + vector2[i]; (*Parameter vector.*)
if randomeffects then (*Actor-dependent random effects in the network objective / gratification function.*)
  begin
  j := i;
  for i := 1 to dim2 do
    begin
    for h := 1 to i do (*Below-diagonal triangle including diagonal.*)
      begin
      inc(j);
      vector1[j] := vector1[j] + matrix1[i,h]; (*Variance-covariance of the random effects.*)
      end;
    end;
  end;
end;

 (*--------------------------------------------------------------------------------------------------*)

procedure InfoRE(var boolean1: boolean; var ascii: textfile);
(*Initial estimates of fixed effects by the method of moments, no standard errors estimated.*)
begin
if boolean1 then
  begin
  writeln(ascii,'W A R N I N G:  standard errors not estimated...values within braces are meaningless!');
  writeln(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

procedure OutputEB_RE(var ascii: TFileStream; const actors: integer; const dim: byte; var eb, mcre, posre, mcposre: rrarray);
(*Write (Empirical) Bayes estimates of actor-dependent random effects (including posterior odds and corresponding Monte Carlo s.e.) to output file.*)
var
  i, j: integer;
  u: real;
  x: string;
begin
Heading(3,ascii,'(Empirical) Bayes (EB) estimates of actor-dependent random effects');
x := 'Columns: actor, EB, posterior probability <positive>, posterior odds <positive>, prediction <positive>.'+eol;
ascii.Write(pchar(x)^,length(x));
x := 'Monte Carlo standard errors within braces.'+eol;
ascii.Write(pchar(x)^,length(x));
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
for i := 1 to actors do
  begin
  x := format('%7d',[i]);
  ascii.Write(pchar(x)^,length(x));
  for j := 1 to dim do
    begin
    x := format('%10.6f%s%7.6f%s',[eb[i,j],'  (',mcre[i,j],')']); (*(Empirical) Bayes estimates of random effects.*)
    ascii.Write(pchar(x)^,length(x));
    x := format('%10.6f%s%7.6f%s',[posre[i,j],'  (',mcposre[i,j],')']); (*(Conditional posterior) probability of positive value of random effect.*)
    ascii.Write(pchar(x)^,length(x));
    if (posre[i,j] < 1) then (*Odds defined.*)
      begin
      u := posre[i,j] / (1 - posre[i,j]);
      x := format('%10.6f',[u]); (*(Conditional) posterior probability of positive value of random effect.*)
      ascii.Write(pchar(x)^,length(x));
      end
    else begin
      x := '  infinite'; (*Odds undefined.*)
      ascii.Write(pchar(x)^,length(x));
    end;
    if (posre[i,j] > 0.5) then begin
      x := '     yes  '; (*Prediction of positive value of random effects, based on the prediction rule <if p > 0.5, then +, otherwise ->.*)
      ascii.Write(pchar(x)^,length(x));
    end
    else begin
      x := '      no  ';
      ascii.Write(pchar(x)^,length(x));
    end;
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
EvaluateCMarginal(dim,actors,eb);
x := 'Column average EB:'+eol;
ascii.Write(pchar(x)^,length(x));
for j := 1 to dim do begin
  x := format('%10.6f',[eb[0,j]]);
  ascii.Write(pchar(x)^,length(x));
  end;
ascii.Write(pchar(eol)^,length(eol));ascii.Write(pchar(eol)^,length(eol));//writeln(ascii); writeln(ascii);
SaveRE(actors,dim,eb); (*Used for future estimation of random effects models from the same data.*)
end;

(*--------------------------------------------------------------------------------------------------*)

procedure CheckFiles (const maxlike: boolean; const sequences: integer; var ascii: TFileStream; const ldim: integer; const randomeffects, Bayesian: boolean);
(*Write information to output file concerning MCMC output files generated by Siena to detect non-convergence / study sampling distributions.*)
var x: string;
begin
if (maxlike or Bayesian or randomeffects or (sequences > 1)) then
  begin
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  x := 'PLEASE NOTE: Siena generated files:'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
  end;
if ((sequences > 1) and (ldim = sequences)) then
  begin
  x := CurrentProjectName+'Theta.m'+'     histogram: parameter estimates'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := CurrentProjectName+'T.m'+'    histogram: test statistic <estimate/s.e.>'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := CurrentProjectName+'Score.m'+ '   histogram: test statistic <c>'+eol;
  ascii.Write(pchar(x)^,length(x));
  if randomeffects then
    begin
    x := CurrentProjectName+'CovRESample.m'+'   histogram: variance-covariance matrix of random effects'+eol;
    ascii.Write(pchar(x)^,length(x));
    ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
    x := '...and with respect to the last run:'+eol;
    ascii.Write(pchar(x)^,length(x));
    end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
if randomeffects then
  begin
  x := CurrentProjectName+'RE.m'+ '   trace plot: actor-dependent random effects'+eol;
  ascii.Write(pchar(x)^,length(x));
  x := CurrentProjectName+'CovRE.m'+ '   trace plot: variance-covariance matrix of actor-dependent random effects'+eol;
  ascii.Write(pchar(x)^,length(x));
  if Bayesian then begin
    x := CurrentProjectName+'SLRCovRE.m'+ '   trace plot: smallest latent root of variance-covariance matrix of actor-dependent random effects'+eol;
   ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
  end;
if (maxlike or Bayesian or randomeffects) then begin
  x := '...numerous files containing acceptance rates, lag t autocorrelations etc.'+eol;
  ascii.Write(pchar(x)^,length(x));
  end;
  ascii.Write(pchar(eol)^,length(eol));//writeln(ascii);
if (maxlike or Bayesian or randomeffects or (sequences > 1)) then
  begin
  x := 'All files stored in the current directory.'+eol;
  ascii.Write(pchar(x)^,length(x));
  ascii.Write(pchar(eol)^,length(eol));// writeln(ascii);
  end;
end;

(*--------------------------------------------------------------------------------------------------*)

end.
